@@ -1,23 +1,20 @@
 /* Copyright (C) 2001-2002 artofcode LLC.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: gdevijs.c,v 1.1.4.2 2002/02/05 21:03:41 giles Exp $ */
+/* $Id: gdevijs.c,v 1.10 2005/03/10 14:57:23 alexcher Exp $ */
 /*
  * IJS device for Ghostscript.
  * Intended to work with any IJS compliant inkjet driver, including
@@ -53,11 +50,54 @@ private dev_proc_close_device(gsijs_close);
 private dev_proc_output_page(gsijs_output_page);
 private dev_proc_get_params(gsijs_get_params);
 private dev_proc_put_params(gsijs_put_params);
+private dev_proc_finish_copydevice(gsijs_finish_copydevice);
 
-private const gx_device_procs gsijs_procs =
-prn_color_params_procs(gsijs_open, gsijs_output_page, gsijs_close,
-		   gx_default_rgb_map_rgb_color, gx_default_rgb_map_color_rgb,
-		   gsijs_get_params, gsijs_put_params);
+private const gx_device_procs gsijs_procs = {
+	gsijs_open,
+	NULL,	/* get_initial_matrix */
+	NULL,	/* sync_output */
+	gsijs_output_page,
+	gsijs_close,
+	gx_default_rgb_map_rgb_color,
+	gx_default_rgb_map_color_rgb,
+	NULL,	/* fill_rectangle */
+	NULL,	/* tile_rectangle */
+	NULL,	/* copy_mono */
+	NULL,	/* copy_color */
+	NULL,	/* draw_line */
+	NULL,	/* get_bits */
+	gsijs_get_params,
+	gsijs_put_params,
+	NULL,	/* map_cmyk_color */
+	NULL,	/* get_xfont_procs */
+	NULL,	/* get_xfont_device */
+	NULL,	/* map_rgb_alpha_color */
+	gx_page_device_get_page_device,
+	NULL,	/* get_alpha_bits */
+	NULL,	/* copy_alpha */
+	NULL,	/* get_band */
+	NULL,	/* copy_rop */
+	NULL,	/* fill_path */
+	NULL,	/* stroke_path */
+	NULL,	/* fill_mask */
+	NULL,	/* fill_trapezoid */
+	NULL,	/* fill_parallelogram */
+	NULL,	/* fill_triangle */
+	NULL,	/* draw_thin_line */
+	NULL,	/* begin_image */
+	NULL,	/* image_data */
+	NULL,	/* end_image */
+	NULL,	/* strip_tile_rectangle */
+	NULL,	/* strip_copy_rop, */
+	NULL,	/* get_clipping_box */
+	NULL,	/* begin_typed_image */
+	NULL,	/* get_bits_rectangle */
+	NULL,	/* map_color_rgb_alpha */
+	NULL,	/* create_compositor */
+	NULL,	/* get_hardware_params */
+	NULL,	/* text_begin */
+	gsijs_finish_copydevice
+};
 
 typedef struct gx_device_ijs_s gx_device_ijs;
 
@@ -401,10 +441,10 @@ gsijs_set_margin_params(gx_device_ijs *ijsdev)
 
 	if (code == 0) {
 	    m[0] = printable_left;
-	    m[1] = printable_top;
+	    m[3] = printable_top;
 	    m[2] = ijsdev->MediaSize[0] * (1.0 / 72) -
 		printable_left - printable_width;
-	    m[3] = ijsdev->MediaSize[1] * (1.0 / 72) -
+	    m[1] = ijsdev->MediaSize[1] * (1.0 / 72) -
 		printable_top - printable_height;
 	    gx_device_set_margins((gx_device *)ijsdev, m, true);
 	    sprintf (buf, "%gx%g", printable_left, printable_top);
@@ -574,7 +614,30 @@ gsijs_open(gx_device *dev)
 	code = gsijs_set_margin_params(ijsdev);
 
     return code;
-};
+}
+
+/* Finish device initialization. */
+private int
+gsijs_finish_copydevice(gx_device *dev, const gx_device *from_dev)
+{
+    int code;
+    static const char rgb[] = "DeviceRGB";
+    gx_device_ijs *ijsdev = (gx_device_ijs *)dev;
+
+    code = gx_default_finish_copydevice(dev, from_dev);
+    if(code < 0)
+        return code;
+    
+    if (!ijsdev->ColorSpace) {
+	ijsdev->ColorSpace = gs_malloc(ijsdev->memory, sizeof(rgb), 1, 
+		"gsijs_finish_copydevice");
+        if (!ijsdev->ColorSpace)
+ 	    return gs_note_error(gs_error_VMerror);
+        ijsdev->ColorSpace_size = sizeof(rgb);
+        memcpy(ijsdev->ColorSpace, rgb, sizeof(rgb));
+    }
+    return code;
+}
 
 /* Close the gsijs driver */
 private int
@@ -591,18 +654,18 @@ gsijs_close(gx_device *dev)
 
     code = gdev_prn_close(dev);
     if (ijsdev->IjsParams)
-	gs_free(ijsdev->IjsParams, ijsdev->IjsParams_size, 1, 
+	gs_free(dev->memory, ijsdev->IjsParams, ijsdev->IjsParams_size, 1, 
 	    "gsijs_read_string_malloc");
     if (ijsdev->ColorSpace)
-	gs_free(ijsdev->ColorSpace,
+	gs_free(dev->memory, ijsdev->ColorSpace,
 		ijsdev->ColorSpace_size, 1, 
 		"gsijs_read_string_malloc");
     if (ijsdev->DeviceManufacturer)
-	gs_free(ijsdev->DeviceManufacturer,
+	gs_free(dev->memory, ijsdev->DeviceManufacturer,
 		ijsdev->DeviceManufacturer_size, 1, 
 		"gsijs_read_string_malloc");
     if (ijsdev->DeviceModel)
-	gs_free(ijsdev->DeviceModel, ijsdev->DeviceModel_size, 1, 
+	gs_free(dev->memory, ijsdev->DeviceModel, ijsdev->DeviceModel_size, 1, 
 		"gsijs_read_string_malloc");
     ijsdev->IjsParams = NULL;
     ijsdev->IjsParams_size = 0;
@@ -919,11 +982,11 @@ gsijs_read_string_malloc(gs_param_list *plist, gs_param_name pname, char **str,
 	    }
 	    if (new_value.size >= *size) {
 	        if (*str)
-		    gs_free(str, *size, 1, "gsijs_read_string_malloc");
+		    gs_free(plist->memory, str, *size, 1, "gsijs_read_string_malloc");
 		*str = NULL;
 		*size = 0;
 	    }
-	    *str = gs_malloc(new_value.size + 1, 1, 
+	    *str = gs_malloc(plist->memory, new_value.size + 1, 1, 
 		"gsijs_read_string_malloc");
 	    if (*str == NULL) {
                 code = gs_note_error(gs_error_VMerror);
@@ -1053,19 +1116,32 @@ gsijs_set_color_format(gx_device_ijs *ijsdev)
 	    ijsdev->procs.map_rgb_color = gx_default_gray_map_rgb_color;
 	    ijsdev->procs.map_color_rgb = gx_default_gray_map_color_rgb;
 	}
+	ijsdev->procs.encode_color = gx_default_gray_fast_encode;
+	ijsdev->procs.decode_color = gx_default_decode_color;
+	dci.polarity = GX_CINFO_POLARITY_ADDITIVE;
+	dci.gray_index = 0;
     } else if (!strcmp (ColorSpace, "DeviceRGB")) {
 	components = 3;
 	ijsdev->procs.map_rgb_color = gx_default_rgb_map_rgb_color;
 	ijsdev->procs.map_color_rgb = gx_default_rgb_map_color_rgb;
+	ijsdev->procs.encode_color = gx_default_rgb_map_rgb_color;
+	ijsdev->procs.decode_color = gx_default_rgb_map_color_rgb;
+	dci.polarity = GX_CINFO_POLARITY_ADDITIVE;
+	dci.gray_index = GX_CINFO_COMP_NO_INDEX;
     } else if (!strcmp (ColorSpace, "DeviceCMYK")) {
 	components = 4;
 	ijsdev->procs.map_cmyk_color = cmyk_8bit_map_cmyk_color;
 	ijsdev->procs.map_color_rgb = cmyk_8bit_map_color_rgb;
+	ijsdev->procs.encode_color = cmyk_8bit_map_cmyk_color;
+	ijsdev->procs.decode_color = gx_default_decode_color;
+	dci.polarity = GX_CINFO_POLARITY_SUBTRACTIVE;
+	dci.gray_index = 3;
     } else {
 	return -1;
     }
 
     maxvalue = (1 << bpc) - 1;
+    dci.max_components = components;
     dci.num_components = components;
     dci.depth = bpc * components;
     dci.max_gray = maxvalue;
@@ -1073,10 +1149,12 @@ gsijs_set_color_format(gx_device_ijs *ijsdev)
     dci.dither_grays = maxvalue+1;
     dci.dither_colors = components > 1 ? maxvalue+1 : 0;
 
-    /* restore old anti_alias info */
-    dci.anti_alias = ijsdev->color_info.anti_alias;
-
+    dci.separable_and_linear = GX_CINFO_SEP_LIN;
+    dci.cm_name = ColorSpace;
+    
     ijsdev->color_info = dci;
 
+    set_linear_color_bits_mask_shift((gx_device *)ijsdev);
+
     return 0;
 }
