@@ -1,23 +1,20 @@
 /* Copyright (C) 1989, 1995, 1996, 1997, 1998, 1999 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: gdevpcfb.c,v $ $Revision: 1.2.2.1 $ */
+/* $Id: gdevpcfb.c,v 1.6 2004/09/20 22:14:59 dan Exp $ */
 /* IBM PC frame buffer (EGA/VGA) drivers */
 #include "memory_.h"
 #include "gconfigv.h"		/* for USE_ASM */
@@ -183,17 +180,15 @@ svga16_put_params(gx_device * dev, gs_param_list * plist)
 /* Map a r-g-b color to an EGA color code. */
 #define Nb gx_color_value_bits
 private gx_color_index
-ega0_map_rgb_color(gx_device * dev, gx_color_value r, gx_color_value g,
-		   gx_color_value b)
+ega0_map_rgb_color(gx_device * dev, const gx_color_value cv[])
 {
-    return pc_4bit_map_rgb_color(dev, r, r, r);
+    return pc_4bit_map_rgb_color(dev, cv);
 }
 private gx_color_index
-ega1_map_rgb_color(gx_device * dev, gx_color_value r, gx_color_value g,
-		   gx_color_value b)
+ega1_map_rgb_color(gx_device * dev, const gx_color_value cv[])
 {
 #define cvtop (gx_color_value)(1 << (Nb - 1))
-    return pc_4bit_map_rgb_color(dev, r & cvtop, g & cvtop, b & cvtop);
+    return pc_4bit_map_rgb_color(dev, cv[0] & cvtop, cv[1] & cvtop, cv[2] & cvtop);
 }
 #undef Nb
 
@@ -233,7 +228,7 @@ typedef rop_params _ss *rop_ptr;
 /* Assembly language routines */
 
 #if USE_ASM
-void memsetcol(P1(rop_ptr));	/* dest, draster, height, data */
+void memsetcol(rop_ptr);	/* dest, draster, height, data */
 #else
 #define memsetcol cmemsetcol
 private void
@@ -253,7 +248,7 @@ cmemsetcol(rop_ptr rop)
 #endif
 
 #if USE_ASM
-void memsetrect(P1(rop_ptr));	/* dest, draster, width, height, data */
+void memsetrect(rop_ptr);	/* dest, draster, width, height, data */
 #else
 #define memsetrect cmemsetrect
 private void
@@ -294,7 +289,7 @@ cmemsetrect(rop_ptr rop)
 #endif
 
 #if USE_ASM
-void memrwcol(P1(rop_ptr));	/* dest, draster, src, sraster, height, shift, invert */
+void memrwcol(rop_ptr);	/* dest, draster, src, sraster, height, shift, invert */
 #  define memrwcol0(rop) memrwcol(rop)	/* same except shift = 0 */
 #else
 #  define memrwcol cmemrwcol
@@ -335,7 +330,7 @@ cmemrwcol0(rop_ptr rop)
 #endif
 
 #if USE_ASM
-void memrwcol2(P1(rop_ptr));	/* dest, draster, src, sraster, height, shift, invert */
+void memrwcol2(rop_ptr);	/* dest, draster, src, sraster, height, shift, invert */
 #else
 #define memrwcol2 cmemrwcol2
 private void
@@ -357,9 +352,9 @@ cmemrwcol2(rop_ptr rop)
 #endif
 
 /* Forward definitions */
-int ega_write_dot(P4(gx_device *, int, int, gx_color_index));
-private void fill_rectangle(P4(rop_ptr, int, int, int));
-private void fill_row_only(P4(byte *, int, int, int));
+int ega_write_dot(gx_device *, int, int, gx_color_index);
+private void fill_rectangle(rop_ptr, int, int, int);
+private void fill_row_only(byte *, int, int, int);
 
 /* Clean up after writing */
 #define dot_end()\
