@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: gdevprn.c,v 1.5.6.2 2002/02/22 19:45:56 ray Exp $ */
+/* $Id: gdevprn.c,v 1.19 2005/03/14 18:08:36 dan Exp $ */
 /* Generic printer driver support */
 #include "ctype_.h"
 #include "gdevprn.h"
@@ -25,6 +25,7 @@
 #include "gxclio.h"
 #include "gxgetbit.h"
 #include "gdevplnx.h"
+#include "gstrans.h"
 
 /*#define DEBUGGING_HACKS*/
 
@@ -54,9 +55,9 @@ const gx_device_procs prn_std_procs =
     prn_procs(gdev_prn_open, gdev_prn_output_page, gdev_prn_close);
 
 /* Forward references */
-int gdev_prn_maybe_realloc_memory(P4(gx_device_printer *pdev, 
-				     gdev_prn_space_params *old_space,
-				     int old_width, int old_height));
+int gdev_prn_maybe_realloc_memory(gx_device_printer *pdev, 
+				  gdev_prn_space_params *old_space,
+				  int old_width, int old_height);
 
 /* ------ Open/close ------ */
 
@@ -141,10 +142,11 @@ open_c:
     clist_init_params(pclist_dev, base, space, pdev,
 		      ppdev->printer_procs.buf_procs,
 		      space_params->band, ppdev->is_async_renderer,
-		      (ppdev->bandlist_memory == 0 ? &gs_memory_default :
+		      (ppdev->bandlist_memory == 0 ? pdev->memory->non_gc_memory:
 		       ppdev->bandlist_memory),
 		      ppdev->free_up_bandlist_memory,
-		      ppdev->clist_disable_mask);
+		      ppdev->clist_disable_mask,
+		      ppdev->page_uses_transparency);
     code = (*gs_clist_device_procs.open_device)( (gx_device *)pcldev );
     if (code < 0) {
 	/* If there wasn't enough room, and we haven't */
@@ -220,13 +222,14 @@ gdev_prn_allocate(gx_device *pdev, gdev_prn_space_params *new_space_params,
     gx_device_memory * const pmemdev = (gx_device_memory *)pdev;
     byte *the_memory = 0;
     gdev_prn_space_params save_params;
-    int save_width, save_height;
-    bool is_command_list;
-    bool save_is_command_list;
+    int save_width = 0x0badf00d; /* Quiet compiler */
+    int save_height = 0x0badf00d; /* Quiet compiler */
+    bool is_command_list = false; /* Quiet compiler */
+    bool save_is_command_list = false; /* Quiet compiler */
     int ecode = 0;
     int pass;
     gs_memory_t *buffer_memory =
-	(ppdev->buffer_memory == 0 ? &gs_memory_default :
+	(ppdev->buffer_memory == 0 ? pdev->memory->non_gc_memory :
 	 ppdev->buffer_memory);
 
     /* If reallocate, find allocated memory & tear down buffer device */
@@ -237,6 +240,7 @@ gdev_prn_allocate(gx_device *pdev, gdev_prn_space_params *new_space_params,
     ppdev->orig_procs = pdev->procs;
     for ( pass = 1; pass <= (reallocate ? 2 : 1); ++pass ) {
 	ulong mem_space;
+	ulong pdf14_trans_buffer_size = 0;
 	byte *base = 0;
 	bool bufferSpace_is_default = false;
 	gdev_prn_space_params space_params;
@@ -266,7 +270,11 @@ gdev_prn_allocate(gx_device *pdev, gdev_prn_space_params *new_space_params,
 	memset(ppdev->skip, 0, sizeof(ppdev->skip));
 	ppdev->printer_procs.buf_procs.size_buf_device
 	    (&buf_space, pdev, NULL, pdev->height, false);
-	mem_space = buf_space.bits + buf_space.line_ptrs;
+	if (ppdev->page_uses_transparency)
+	    pdf14_trans_buffer_size = new_height
+		* (ESTIMATED_PDF14_ROW_SPACE(new_width) >> 3);
+	mem_space = buf_space.bits + buf_space.line_ptrs
+		    + pdf14_trans_buffer_size;
 
 	/* Compute desired space params: never use the space_params as-is. */
 	/* Rather, give the dev-specific driver a chance to adjust them. */
@@ -306,7 +314,7 @@ gdev_prn_allocate(gx_device *pdev, gdev_prn_space_params *new_space_params,
 		the_memory = base;
 	}
 	if (!is_command_list && pass == 1 && PRN_MIN_MEMORY_LEFT != 0
-	    && buffer_memory == &gs_memory_default) {
+	    && buffer_memory == pdev->memory->non_gc_memory) {
 	    /* before using full memory buffer, ensure enough working mem left */
 	    byte * left = gs_alloc_bytes( buffer_memory,
 					  PRN_MIN_MEMORY_LEFT, "printer mem left");
@@ -385,6 +393,16 @@ gdev_prn_allocate(gx_device *pdev, gdev_prn_space_params *new_space_params,
 	COPY_PROC(get_clipping_box);
 	COPY_PROC(map_color_rgb_alpha);
 	COPY_PROC(get_hardware_params);
+	COPY_PROC(get_color_mapping_procs);
+	COPY_PROC(get_color_comp_index);
+	COPY_PROC(encode_color);
+	COPY_PROC(decode_color);
+	COPY_PROC(begin_image);
+	COPY_PROC(text_begin);
+	COPY_PROC(fill_path);
+	COPY_PROC(stroke_path);
+	COPY_PROC(fill_rectangle_hl_color);
+	COPY_PROC(update_spot_equivalent_colors);
 #undef COPY_PROC
 	/* If using a command list, already opened the device. */
 	if (is_command_list)
@@ -422,7 +440,7 @@ gdev_prn_free_memory(gx_device *pdev)
     gx_device_printer * const ppdev = (gx_device_printer *)pdev;
     byte *the_memory = 0;
     gs_memory_t *buffer_memory =
-	(ppdev->buffer_memory == 0 ? &gs_memory_default :
+	(ppdev->buffer_memory == 0 ? pdev->memory->non_gc_memory :
 	 ppdev->buffer_memory);
 
     gdev_prn_tear_down(pdev, &the_memory);
@@ -472,6 +490,8 @@ gdev_prn_get_params(gx_device * pdev, gs_param_list * plist)
 	(code = param_write_long(plist, "BandBufferSpace", &ppdev->space_params.band.BandBufferSpace)) < 0 ||
 	(code = param_write_bool(plist, "OpenOutputFile", &ppdev->OpenOutputFile)) < 0 ||
 	(code = param_write_bool(plist, "ReopenPerPage", &ppdev->ReopenPerPage)) < 0 ||
+	(code = param_write_bool(plist, "PageUsesTransparency",
+			 	&ppdev->page_uses_transparency)) < 0 ||
 	(ppdev->Duplex_set >= 0 &&
 	 (code = (ppdev->Duplex_set ?
 		  param_write_bool(plist, "Duplex", &ppdev->Duplex) :
@@ -507,6 +527,7 @@ gdev_prn_put_params(gx_device * pdev, gs_param_list * plist)
     bool is_open = pdev->is_open;
     bool oof = ppdev->OpenOutputFile;
     bool rpp = ppdev->ReopenPerPage;
+    bool page_uses_transparency = ppdev->page_uses_transparency;
     bool duplex;
     int duplex_set = -1;
     int width = pdev->width;
@@ -536,6 +557,16 @@ gdev_prn_put_params(gx_device * pdev, gs_param_list * plist)
 	    break;
     }
 
+    switch (code = param_read_bool(plist, (param_name = "PageUsesTransparency"),
+			    				&page_uses_transparency)) {
+	default:
+	    ecode = code;
+	    param_signal_error(plist, param_name, ecode);
+	case 0:
+	case 1:
+	    break;
+    }
+
     if (ppdev->Duplex_set >= 0)	/* i.e., Duplex is supported */
 	switch (code = param_read_bool(plist, (param_name = "Duplex"),
 				       &duplex)) {
@@ -588,8 +619,11 @@ label:\
 
     switch (code = param_read_string(plist, (param_name = "OutputFile"), &ofs)) {
 	case 0:
-	    if (pdev->LockSafetyParams)
+	    if (pdev->LockSafetyParams &&
+		    bytes_compare(ofs.data, ofs.size,
+			(const byte *)ppdev->fname, strlen(ppdev->fname))) {
 	        code = gs_error_invalidaccess;
+	    }
 	    else
 		code = validate_output_file(&ofs);
 	    if (code >= 0)
@@ -632,6 +666,7 @@ label:\
 
     ppdev->OpenOutputFile = oof;
     ppdev->ReopenPerPage = rpp;
+    ppdev->page_uses_transparency = page_uses_transparency;
     if (duplex_set >= 0) {
 	ppdev->Duplex = duplex;
 	ppdev->Duplex_set = duplex_set;
@@ -975,6 +1010,8 @@ gx_default_create_buf_device(gx_device **pbdev, gx_device *target,
     if (target == (gx_device *)mdev) {
 	/* The following is a special hack for setting up printer devices. */
 	assign_dev_procs(mdev, mdproto);
+        check_device_separable((gx_device *)mdev);
+	gx_device_fill_in_procs((gx_device *)mdev);
     } else
 	gs_make_mem_device(mdev, mdproto, mem, (for_band ? 1 : 0),
 			   (target == (gx_device *)mdev ? NULL : target));
