@@ -1,23 +1,20 @@
 /* Copyright (C) 1997, 2000 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: gdevvec.c,v $ $Revision: 1.7.2.3 $ */
+/* $Id: gdevvec.c,v 1.25 2004/09/27 04:55:00 ray Exp $ */
 /* Utilities for "vector" devices */
 #include "math_.h"
 #include "memory_.h"
@@ -133,7 +130,7 @@ gdev_vector_dopath(gx_device_vector *vdev, const gx_path * ppath,
 		}
 	    }
 	    if (incomplete_line) {
-		if (need_moveto) {      /* see gs_pe_moveto case */
+		if (need_moveto) {	/* see gs_pe_moveto case */
 		    code = gdev_vector_dopath_segment(&state, gs_pe_moveto,
 						      &line_start);
 		    if (code < 0)
@@ -255,27 +252,40 @@ gdev_vector_reset(gx_device_vector * vdev)
     {gs_imager_state_initial(1)};
 
     vdev->state = state_initial;
-    color_unset(&vdev->fill_color);
-    color_unset(&vdev->stroke_color);
+    gx_hld_saved_color_init(&vdev->saved_fill_color);
+    gx_hld_saved_color_init(&vdev->saved_stroke_color);
     vdev->clip_path_id =
-	vdev->no_clip_path_id = gs_next_ids(1);
+	vdev->no_clip_path_id = gs_next_ids(vdev->memory, 1);
 }
 
 /* Open the output file and stream. */
 int
-gdev_vector_open_file_bbox(gx_device_vector * vdev, uint strmbuf_size,
-			   bool bbox)
-{				/* Open the file as positionable if possible. */
-    int code = gx_device_open_output_file((gx_device *) vdev, vdev->fname,
-					  true, true, &vdev->file);
+gdev_vector_open_file_options(gx_device_vector * vdev, uint strmbuf_size,
+			      int open_options)
+{
+    bool binary = !(open_options & VECTOR_OPEN_FILE_ASCII);
+    int code = -1;		/* (only for testing, never returned) */
 
+    /* Open the file as seekable or sequential, as requested. */
+    if (!(open_options & VECTOR_OPEN_FILE_SEQUENTIAL)) {
+	/* Try to open as seekable. */
+	code =
+	    gx_device_open_output_file((gx_device *)vdev, vdev->fname,
+				       binary, true, &vdev->file);
+    }
+    if (code < 0 && (open_options & (VECTOR_OPEN_FILE_SEQUENTIAL |
+				     VECTOR_OPEN_FILE_SEQUENTIAL_OK))) {
+	/* Try to open as sequential. */
+	code = gx_device_open_output_file((gx_device *)vdev, vdev->fname,
+					  binary, false, &vdev->file);
+    }
     if (code < 0)
 	return code;
     if ((vdev->strmbuf = gs_alloc_bytes(vdev->v_memory, strmbuf_size,
 					"vector_open(strmbuf)")) == 0 ||
 	(vdev->strm = s_alloc(vdev->v_memory,
 			      "vector_open(strm)")) == 0 ||
-	(bbox &&
+	((open_options & VECTOR_OPEN_FILE_BBOX) &&
 	 (vdev->bbox_device =
 	  gs_alloc_struct_immovable(vdev->v_memory,
 				    gx_device_bbox, &st_device_bbox,
@@ -299,13 +309,15 @@ gdev_vector_open_file_bbox(gx_device_vector * vdev, uint strmbuf_size,
     }
     vdev->strmbuf_size = strmbuf_size;
     swrite_file(vdev->strm, vdev->file, vdev->strmbuf, strmbuf_size);
+    vdev->open_options = open_options;
     /*
      * We don't want finalization to close the file, but we do want it
      * to flush the stream buffer.
      */
     vdev->strm->procs.close = vdev->strm->procs.flush;
     if (vdev->bbox_device) {
-	gx_device_bbox_init(vdev->bbox_device, NULL);
+	gx_device_bbox_init(vdev->bbox_device, NULL, vdev->v_memory);
+        rc_increment(vdev->bbox_device);
 	gx_device_set_resolution((gx_device *) vdev->bbox_device,
 				 vdev->HWResolution[0],
 				 vdev->HWResolution[1]);
@@ -331,19 +343,6 @@ gdev_vector_stream(gx_device_vector * vdev)
     return vdev->strm;
 }
 
-/* Compare two drawing colors. */
-/* Right now we don't attempt to handle non-pure colors. */
-private bool
-drawing_color_eq(const gx_drawing_color * pdc1, const gx_drawing_color * pdc2)
-{
-    return (gx_dc_is_pure(pdc1) ?
-	    gx_dc_is_pure(pdc2) &&
-	    gx_dc_pure_color(pdc1) == gx_dc_pure_color(pdc2) :
-	    gx_dc_is_null(pdc1) ?
-	    gx_dc_is_null(pdc2) :
-	    false);
-}
-
 /* Update the logical operation. */
 int
 gdev_vector_update_log_op(gx_device_vector * vdev, gs_logical_operation_t lop)
@@ -360,27 +359,47 @@ gdev_vector_update_log_op(gx_device_vector * vdev, gs_logical_operation_t lop)
     return 0;
 }
 
+/* Update color (fill or stroke). */
+private int
+gdev_vector_update_color(gx_device_vector * vdev,
+			      const gs_imager_state * pis,
+			      const gx_drawing_color * pdcolor,
+			      gx_hl_saved_color *sc,
+			      int (*setcolor) (gx_device_vector * vdev, 
+			                       const gs_imager_state * pis, 
+					       const gx_drawing_color * pdc))
+{
+    gx_hl_saved_color temp;
+    int code;
+    bool hl_color = (*vdev_proc(vdev, can_handle_hl_color)) (vdev, pis, pdcolor);
+    const gs_imager_state *pis_for_hl_color = (hl_color ? pis : NULL);
+    
+    gx_hld_save_color(pis_for_hl_color, pdcolor, &temp);
+    if (gx_hld_saved_color_equal(&temp, sc))
+	return 0;
+    code = (*setcolor) (vdev, pis_for_hl_color, pdcolor);
+    if (code < 0)
+	return code;
+    *sc = temp;
+    return 0;
+}
+
 /* Update the fill color. */
 int
 gdev_vector_update_fill_color(gx_device_vector * vdev,
+			      const gs_imager_state * pis,
 			      const gx_drawing_color * pdcolor)
 {
-    if (!drawing_color_eq(pdcolor, &vdev->fill_color)) {
-	int code = (*vdev_proc(vdev, setfillcolor)) (vdev, pdcolor);
-
-	if (code < 0)
-	    return code;
-	vdev->fill_color = *pdcolor;
-    }
-    return 0;
+    return gdev_vector_update_color(vdev, pis, pdcolor, &vdev->saved_fill_color, 
+                                    vdev_proc(vdev, setfillcolor));
 }
 
 /* Update the state for filling a region. */
 private int
-update_fill(gx_device_vector * vdev, const gx_drawing_color * pdcolor,
-	    gs_logical_operation_t lop)
+update_fill(gx_device_vector * vdev, const gs_imager_state * pis, 
+	    const gx_drawing_color * pdcolor, gs_logical_operation_t lop)
 {
-    int code = gdev_vector_update_fill_color(vdev, pdcolor);
+    int code = gdev_vector_update_fill_color(vdev, pis, pdcolor);
 
     if (code < 0)
 	return code;
@@ -399,7 +418,7 @@ gdev_vector_prepare_fill(gx_device_vector * vdev, const gs_imager_state * pis,
 	    return code;
 	vdev->state.flatness = params->flatness;
     }
-    return update_fill(vdev, pdcolor, pis->log_op);
+    return update_fill(vdev, pis, pdcolor, pis->log_op);
 }
 
 /* Compare two dash patterns. */
@@ -416,95 +435,102 @@ dash_pattern_eq(const float *stored, const gx_dash_params * set, floatp scale)
 
 /* Bring state up to date for stroking. */
 int
-gdev_vector_prepare_stroke(gx_device_vector * vdev, const gs_imager_state * pis,
-	  const gx_stroke_params * params, const gx_drawing_color * pdcolor,
+gdev_vector_prepare_stroke(gx_device_vector * vdev,
+			   const gs_imager_state * pis,	/* may be NULL */
+			   const gx_stroke_params * params, /* may be NULL */
+			   const gx_drawing_color * pdcolor, /* may be NULL */
 			   floatp scale)
 {
-    int pattern_size = pis->line_params.dash.pattern_size;
-    float dash_offset = pis->line_params.dash.offset * scale;
-    float half_width = pis->line_params.half_width * scale;
+    if (pis) {
+	int pattern_size = pis->line_params.dash.pattern_size;
+	float dash_offset = pis->line_params.dash.offset * scale;
+	float half_width = pis->line_params.half_width * scale;
 
-    if (pattern_size > max_dash)
-	return_error(gs_error_limitcheck);
-    if (dash_offset != vdev->state.line_params.dash.offset ||
-	pattern_size != vdev->state.line_params.dash.pattern_size ||
-	(pattern_size != 0 &&
-	 !dash_pattern_eq(vdev->dash_pattern, &pis->line_params.dash,
-			  scale))
-	) {
-	float pattern[max_dash];
-	int i, code;
+	if (pattern_size > max_dash)
+	    return_error(gs_error_limitcheck);
+	if (dash_offset != vdev->state.line_params.dash.offset ||
+	    pattern_size != vdev->state.line_params.dash.pattern_size ||
+	    (pattern_size != 0 &&
+	     !dash_pattern_eq(vdev->dash_pattern, &pis->line_params.dash,
+			      scale))
+	    ) {
+	    float pattern[max_dash];
+	    int i, code;
 
-	for (i = 0; i < pattern_size; ++i)
-	    pattern[i] = pis->line_params.dash.pattern[i] * scale;
-	code = (*vdev_proc(vdev, setdash))
-	    (vdev, pattern, pattern_size, dash_offset);
-	if (code < 0)
-	    return code;
-	memcpy(vdev->dash_pattern, pattern, pattern_size * sizeof(float));
+	    for (i = 0; i < pattern_size; ++i)
+		pattern[i] = pis->line_params.dash.pattern[i] * scale;
+	    code = (*vdev_proc(vdev, setdash))
+		(vdev, pattern, pattern_size, dash_offset);
+	    if (code < 0)
+		return code;
+	    memcpy(vdev->dash_pattern, pattern, pattern_size * sizeof(float));
 
-	vdev->state.line_params.dash.pattern_size = pattern_size;
-	vdev->state.line_params.dash.offset = dash_offset;
+	    vdev->state.line_params.dash.pattern_size = pattern_size;
+	    vdev->state.line_params.dash.offset = dash_offset;
+	}
+	if (half_width != vdev->state.line_params.half_width) {
+	    int code = (*vdev_proc(vdev, setlinewidth))
+		(vdev, half_width * 2);
+
+	    if (code < 0)
+		return code;
+	    vdev->state.line_params.half_width = half_width;
+	}
+	if (pis->line_params.miter_limit != vdev->state.line_params.miter_limit) {
+	    int code = (*vdev_proc(vdev, setmiterlimit))
+		(vdev, pis->line_params.miter_limit);
+
+	    if (code < 0)
+		return code;
+	    gx_set_miter_limit(&vdev->state.line_params,
+			       pis->line_params.miter_limit);
+	}
+	if (pis->line_params.cap != vdev->state.line_params.cap) {
+	    int code = (*vdev_proc(vdev, setlinecap))
+		(vdev, pis->line_params.cap);
+
+	    if (code < 0)
+		return code;
+	    vdev->state.line_params.cap = pis->line_params.cap;
+	}
+	if (pis->line_params.join != vdev->state.line_params.join) {
+	    int code = (*vdev_proc(vdev, setlinejoin))
+		(vdev, pis->line_params.join);
+
+	    if (code < 0)
+		return code;
+	    vdev->state.line_params.join = pis->line_params.join;
+	} {
+	    int code = gdev_vector_update_log_op(vdev, pis->log_op);
+
+	    if (code < 0)
+		return code;
+	}
     }
-    if (params->flatness != vdev->state.flatness) {
-	int code = (*vdev_proc(vdev, setflat)) (vdev, params->flatness);
+    if (params) {
+	if (params->flatness != vdev->state.flatness) {
+	    int code = (*vdev_proc(vdev, setflat)) (vdev, params->flatness);
 
-	if (code < 0)
-	    return code;
-	vdev->state.flatness = params->flatness;
+	    if (code < 0)
+		return code;
+	    vdev->state.flatness = params->flatness;
+	}
     }
-    if (half_width != vdev->state.line_params.half_width) {
-	int code = (*vdev_proc(vdev, setlinewidth))
-	    (vdev, half_width * 2);
+    if (pdcolor) {
+	int code = gdev_vector_update_color(vdev, pis, pdcolor, 
+		    &vdev->saved_stroke_color, vdev_proc(vdev, setstrokecolor));
 
 	if (code < 0)
 	    return code;
-	vdev->state.line_params.half_width = half_width;
-    }
-    if (pis->line_params.miter_limit != vdev->state.line_params.miter_limit) {
-	int code = (*vdev_proc(vdev, setmiterlimit))
-	    (vdev, pis->line_params.miter_limit);
-
-	if (code < 0)
-	    return code;
-	gx_set_miter_limit(&vdev->state.line_params,
-			   pis->line_params.miter_limit);
-    }
-    if (pis->line_params.cap != vdev->state.line_params.cap) {
-	int code = (*vdev_proc(vdev, setlinecap))
-	    (vdev, pis->line_params.cap);
-
-	if (code < 0)
-	    return code;
-	vdev->state.line_params.cap = pis->line_params.cap;
-    }
-    if (pis->line_params.join != vdev->state.line_params.join) {
-	int code = (*vdev_proc(vdev, setlinejoin))
-	    (vdev, pis->line_params.join);
-
-	if (code < 0)
-	    return code;
-	vdev->state.line_params.join = pis->line_params.join;
-    } {
-	int code = gdev_vector_update_log_op(vdev, pis->log_op);
-
-	if (code < 0)
-	    return code;
-    }
-    if (!drawing_color_eq(pdcolor, &vdev->stroke_color)) {
-	int code = (*vdev_proc(vdev, setstrokecolor)) (vdev, pdcolor);
-
-	if (code < 0)
-	    return code;
-	vdev->stroke_color = *pdcolor;
     }
     return 0;
 }
 
 /*
- * Compute the scale or transformation matrix for transforming the line
- * width and dash pattern for a stroke operation.  Return 0 if scaling,
- * 1 if a full matrix is needed.
+ * Compute the scale for transforming the line width and dash pattern for a
+ * stroke operation, and, if necessary to handle anisotropic scaling, a full
+ * transformation matrix to be inverse-applied to the path elements as well.
+ * Return 0 if only scaling, 1 if a full matrix is needed.
  */
 int
 gdev_vector_stroke_scaling(const gx_device_vector *vdev,
@@ -538,10 +564,11 @@ gdev_vector_stroke_scaling(const gx_device_vector *vdev,
     }
     if (set_ctm) {
 	/*
-	 * Adobe Acrobat Reader can't handle user coordinates larger than
-	 * 32K.  If we scale the matrix down too far, the coordinates will
-	 * get too big: don't allow this to happen.  (This does no harm
-	 * for other output formats.)
+	 * Adobe Acrobat Reader has limitations on the maximum user
+	 * coordinate value.  If we scale the matrix down too far, the
+	 * coordinates will get too big: limit the scale factor to prevent
+	 * this from happening.  (This does no harm for other output
+	 * formats.)
 	 */
 	double
 	    mxx = pis->ctm.xx / vdev->scale.x,
@@ -600,7 +627,7 @@ gdev_vector_dopath_segment(gdev_vector_dopath_state_t *state, int pe_op,
 	    if (state->first)
 		state->start = vp[0], state->first = false;
 	    code = vdev_proc(vdev, moveto)
-		(vdev, state->prev.x, state->prev.y, vp[0].x, vp[0].y,
+		(vdev, 0/*unused*/, 0/*unused*/, vp[0].x, vp[0].y,
 		 state->type);
 	    state->prev = vp[0];
 	    break;
@@ -779,7 +806,8 @@ gdev_vector_close_file(gx_device_vector * vdev)
     vdev->file = 0;
     err = ferror(f);
     /* We prevented sclose from closing the file. */
-    if (fclose(f) != 0 || err != 0)
+    if (gx_device_close_output_file((gx_device *)vdev, vdev->fname, f) != 0 
+	|| err != 0)
 	return_error(gs_error_ioerror);
     return 0;
 }
@@ -819,7 +847,7 @@ gdev_vector_begin_image(gx_device_vector * vdev,
 	(code = gdev_vector_update_clip_path(vdev, pcpath)) < 0 ||
 	((pim->ImageMask ||
 	  (pim->CombineWithColor && rop3_uses_T(pis->log_op))) &&
-	 (code = gdev_vector_update_fill_color(vdev, pdcolor)) < 0) ||
+	 (code = gdev_vector_update_fill_color(vdev, pis, pdcolor)) < 0) ||
 	(vdev->bbox_device &&
 	 (code = (*dev_proc(vdev->bbox_device, begin_image))
 	  ((gx_device *) vdev->bbox_device, pis, pim, format, prect,
@@ -926,10 +954,12 @@ gdev_vector_put_params(gx_device * dev, gs_param_list * plist)
 		/* The new name is the same as the old name.  Do nothing. */
 		ofns.data = 0;
 		break;
-	    } else if (dev->is_open && vdev->strm != 0 &&
-		       stell(vdev->strm) != 0
+	    } else if (dev->LockSafetyParams ||
+	    		(dev->is_open && vdev->strm != 0 &&
+		       stell(vdev->strm) != 0)
 		       )
-		ecode = gs_error_rangecheck;
+		ecode = (dev->LockSafetyParams) ? gs_error_invalidaccess : 
+				gs_error_rangecheck;
 	    else
 		break;
 	    goto ofe;
@@ -965,11 +995,10 @@ gdev_vector_put_params(gx_device * dev, gs_param_list * plist)
 	    vdev->bbox_device = bbdev;
 	    if (code < 0)
 		return code;
-	    return gdev_vector_open_file_bbox(vdev, vdev->strmbuf_size,
-					      bbdev != 0);
+	    return gdev_vector_open_file_options(vdev, vdev->strmbuf_size,
+						 vdev->open_options);
 	}
     }
-    gdev_vector_load_cache(vdev);	/* in case color mapping changed */
     return 0;
 }
 
@@ -984,15 +1013,18 @@ gdev_vector_fill_rectangle(gx_device * dev, int x, int y, int w, int h,
     /* Ignore the initial fill with white. */
     if (!vdev->in_page && color == vdev->white)
 	return 0;
-    color_set_pure(&dcolor, color);
+    /*
+     * The original colorspace and client color are unknown so use
+     * set_nonclient_dev_color instead of color_set_pure.
+     */
+    set_nonclient_dev_color(&dcolor, color);
     {
-	int code = update_fill(vdev, &dcolor, rop3_T);
+	/* Make sure we aren't being clipped. */
+	int code = gdev_vector_update_clip_path(vdev, NULL);
 
 	if (code < 0)
 	    return code;
-	/* Make sure we aren't being clipped. */
-	code = gdev_vector_update_clip_path(vdev, NULL);
-	if (code < 0)
+	if ((code = update_fill(vdev, NULL, &dcolor, rop3_T)) < 0)
 	    return code;
     }
     if (vdev->bbox_device) {
@@ -1075,7 +1107,7 @@ gdev_vector_fill_trapezoid(gx_device * dev, const gs_fixed_edge * left,
 
 #define y0 ybot
 #define y1 ytop
-    int code = update_fill(vdev, pdevc, lop);
+    int code = update_fill(vdev, NULL, pdevc, lop);
     gs_fixed_point points[4];
 
     if (code < 0)
@@ -1115,7 +1147,7 @@ gdev_vector_fill_parallelogram(gx_device * dev,
 		  const gx_device_color * pdevc, gs_logical_operation_t lop)
 {
     fixed pax = px + ax, pay = py + ay;
-    int code = update_fill(vdev, pdevc, lop);
+    int code = update_fill(vdev, NULL, pdevc, lop);
     gs_fixed_point points[4];
 
     if (code < 0)
@@ -1145,7 +1177,7 @@ gdev_vector_fill_triangle(gx_device * dev,
 		 fixed px, fixed py, fixed ax, fixed ay, fixed bx, fixed by,
 		  const gx_device_color * pdevc, gs_logical_operation_t lop)
 {
-    int code = update_fill(vdev, pdevc, lop);
+    int code = update_fill(vdev, NULL, pdevc, lop);
     gs_fixed_point points[3];
 
     if (code < 0)
