@@ -1,23 +1,20 @@
 /* Copyright (C) 1995, 2000 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: gsalloc.c,v $ $Revision: 1.8.2.2 $ */
+/* $Id: gsalloc.c,v 1.22 2004/08/19 19:33:09 stefan Exp $ */
 /* Standard memory allocator */
 #include "gx.h"
 #include "memory_.h"
@@ -28,6 +25,15 @@
 #include "gxalloc.h"
 #include "stream.h"		/* for clearing stream list */
 
+/*
+ * Define whether to try consolidating space before adding a new chunk.
+ * The default is not to do this, because it is computationally
+ * expensive and doesn't seem to help much.  However, this is done for
+ * "controlled" spaces whether or not the #define is in effect.
+ */
+/*#define CONSOLIDATE_BEFORE_ADDING_CHUNK */
+
+
 /*
  * This allocator produces tracing messages of the form
  *      [aNMOTS]...
@@ -94,7 +100,7 @@ private RELOC_PTRS_WITH(ref_memory_reloc_ptrs, gs_ref_memory_t *mptr)
 RELOC_PTRS_END
 
 /*
- * Define flags for the alloc_obj, which implements all but the fastest
+ * Define the flags for alloc_obj, which implements all but the fastest
  * case of allocation.
  */
 typedef enum {
@@ -103,16 +109,16 @@ typedef enum {
 } alloc_flags_t;
 
 /* Forward references */
-private void remove_range_from_freelist(P3(gs_ref_memory_t *mem, void* bottom, void* top));
-private obj_header_t *large_freelist_alloc(P2(gs_ref_memory_t *mem, uint size));
-private obj_header_t *scavenge_low_free(P2(gs_ref_memory_t *mem, unsigned request_size));
-private ulong compute_free_objects(P1(gs_ref_memory_t *));
-private obj_header_t *alloc_obj(P5(gs_ref_memory_t *, ulong, gs_memory_type_ptr_t, alloc_flags_t, client_name_t));
-private void consolidate_chunk_free(P2(chunk_t *cp, gs_ref_memory_t *mem));
-private void trim_obj(P4(gs_ref_memory_t *mem, obj_header_t *obj, uint size, chunk_t *cp));
-private chunk_t *alloc_acquire_chunk(P4(gs_ref_memory_t *, ulong, bool, client_name_t));
-private chunk_t *alloc_add_chunk(P3(gs_ref_memory_t *, ulong, client_name_t));
-void alloc_close_chunk(P1(gs_ref_memory_t *));
+private void remove_range_from_freelist(gs_ref_memory_t *mem, void* bottom, void* top);
+private obj_header_t *large_freelist_alloc(gs_ref_memory_t *mem, uint size);
+private obj_header_t *scavenge_low_free(gs_ref_memory_t *mem, unsigned request_size);
+private ulong compute_free_objects(gs_ref_memory_t *);
+private obj_header_t *alloc_obj(gs_ref_memory_t *, ulong, gs_memory_type_ptr_t, alloc_flags_t, client_name_t);
+private void consolidate_chunk_free(chunk_t *cp, gs_ref_memory_t *mem);
+private void trim_obj(gs_ref_memory_t *mem, obj_header_t *obj, uint size, chunk_t *cp);
+private chunk_t *alloc_acquire_chunk(gs_ref_memory_t *, ulong, bool, client_name_t);
+private chunk_t *alloc_add_chunk(gs_ref_memory_t *, ulong, client_name_t);
+void alloc_close_chunk(gs_ref_memory_t *);
 
 /*
  * Define the standard implementation (with garbage collection)
@@ -179,10 +185,10 @@ const gs_memory_procs_t gs_ref_memory_procs =
  * Allocate and mostly initialize the state of an allocator (system, global,
  * or local).  Does not initialize global or space.
  */
-private void *ialloc_solo(P3(gs_raw_memory_t *, gs_memory_type_ptr_t,
-			     chunk_t **));
+private void *ialloc_solo(gs_memory_t *, gs_memory_type_ptr_t,
+			  chunk_t **);
 gs_ref_memory_t *
-ialloc_alloc_state(gs_raw_memory_t * parent, uint chunk_size)
+ialloc_alloc_state(gs_memory_t * parent, uint chunk_size)
 {
     chunk_t *cp;
     gs_ref_memory_t *iimem = ialloc_solo(parent, &st_ref_memory, &cp);
@@ -191,14 +197,18 @@ ialloc_alloc_state(gs_raw_memory_t * parent, uint chunk_size)
 	return 0;
     iimem->stable_memory = (gs_memory_t *)iimem;
     iimem->procs = gs_ref_memory_procs;
-    iimem->parent = parent;
+    iimem->gs_lib_ctx = parent->gs_lib_ctx;
+    iimem->non_gc_memory = parent;
     iimem->chunk_size = chunk_size;
     iimem->large_size = ((chunk_size / 4) & -obj_align_mod) + 1;
     iimem->is_controlled = false;
     iimem->gc_status.vm_threshold = chunk_size * 3L;
     iimem->gc_status.max_vm = max_long;
     iimem->gc_status.psignal = NULL;
+    iimem->gc_status.signal_value = 0;
     iimem->gc_status.enabled = false;
+    iimem->gc_status.requested = 0;
+    iimem->gc_allocated = 0;
     iimem->previous_status.allocated = 0;
     iimem->previous_status.used = 0;
     ialloc_reset(iimem);
@@ -219,7 +229,7 @@ ialloc_alloc_state(gs_raw_memory_t * parent, uint chunk_size)
 
 /* Allocate a 'solo' object with its own chunk. */
 private void *
-ialloc_solo(gs_raw_memory_t * parent, gs_memory_type_ptr_t pstype,
+ialloc_solo(gs_memory_t * parent, gs_memory_type_ptr_t pstype,
 	    chunk_t ** pcp)
 {	/*
 	 * We can't assume that the parent uses the same object header
@@ -334,6 +344,14 @@ ialloc_reset_free(gs_ref_memory_t * mem)
     mem->largest_free_size = 0;
 }
 
+/*
+ * Set an arbitrary limit so that the amount of allocated VM does not grow
+ * indefinitely even when GC is disabled.  Benchmarks have shown that
+ * the resulting GC's are infrequent enough not to degrade performance
+ * significantly.
+ */
+#define FORCE_GC_LIMIT 8000000
+
 /* Set the allocation limit after a change in one or more of */
 /* vm_threshold, max_vm, or enabled, or after a GC. */
 void
@@ -359,7 +377,7 @@ ialloc_set_limit(register gs_ref_memory_t * mem)
 	    mem->limit = min(limit, max_allocated);
 	}
     } else
-	mem->limit = max_allocated;
+	mem->limit = min(max_allocated, mem->gc_allocated + FORCE_GC_LIMIT);
     if_debug7('0', "[0]space=%d, max_vm=%ld, prev.alloc=%ld, enabled=%d,\n\
       gc_alloc=%ld, threshold=%ld => limit=%ld\n",
 	      mem->space, (long)mem->gc_status.max_vm,
@@ -433,6 +451,36 @@ gs_memory_set_gc_status(gs_ref_memory_t * mem, const gs_memory_gc_status_t * pst
     ialloc_set_limit(mem);
 }
 
+/* Set VM threshold. */
+void
+gs_memory_set_vm_threshold(gs_ref_memory_t * mem, long val)
+{
+    gs_memory_gc_status_t stat;
+    gs_ref_memory_t * stable = (gs_ref_memory_t *)mem->stable_memory;
+
+    gs_memory_gc_status(mem, &stat);
+    stat.vm_threshold = val;
+    gs_memory_set_gc_status(mem, &stat);
+    gs_memory_gc_status(stable, &stat);
+    stat.vm_threshold = val;
+    gs_memory_set_gc_status(stable, &stat);
+}
+
+/* Set VM reclaim. */
+void
+gs_memory_set_vm_reclaim(gs_ref_memory_t * mem, bool enabled)
+{
+    gs_memory_gc_status_t stat;
+    gs_ref_memory_t * stable = (gs_ref_memory_t *)mem->stable_memory;
+
+    gs_memory_gc_status(mem, &stat);
+    stat.enabled = enabled;
+    gs_memory_set_gc_status(mem, &stat);
+    gs_memory_gc_status(stable, &stat);
+    stat.enabled = enabled;
+    gs_memory_set_gc_status(stable, &stat);
+}
+
 /* ================ Objects ================ */
 
 /* Allocate a small object quickly if possible. */
@@ -585,6 +633,13 @@ i_alloc_struct_array(gs_memory_t * mem, uint num_elements,
     obj_header_t *obj;
 
     ALLOC_CHECK_SIZE(pstype);
+#ifdef DEBUG
+    if (pstype->enum_ptrs == basic_enum_ptrs) {
+	dprintf2("  i_alloc_struct_array: called with incorrect structure type (not element), struct='%s', client='%s'\n",
+		pstype->sname, cname);
+	return NULL;		/* fail */
+    }
+#endif
     obj = alloc_obj(imem,
 		    (ulong) num_elements * pstype->ssize,
 		    pstype, ALLOC_DIRECT, cname);
@@ -630,16 +685,16 @@ i_resize_object(gs_memory_t * mem, void *obj, uint new_num_elements,
 	pp->o_size = new_size;
 	new_obj = obj;
     } else
-    if ((byte *)obj + old_size_rounded == imem->cc.cbot &&
-	imem->cc.ctop - (byte *)obj >= new_size_rounded ) {
-	imem->cc.cbot = (byte *)obj + new_size_rounded;
-	pp->o_size = new_size;
-	new_obj = obj;
-    } else /* try and trim the object -- but only if room for a dummy header */
-	if (new_size_rounded + sizeof(obj_header_t) <= old_size_rounded) {
-		trim_obj(imem, obj, new_size, (chunk_t *)0);
+	if ((byte *)obj + old_size_rounded == imem->cc.cbot &&
+	    imem->cc.ctop - (byte *)obj >= new_size_rounded ) {
+	    imem->cc.cbot = (byte *)obj + new_size_rounded;
+	    pp->o_size = new_size;
 	    new_obj = obj;
-	}
+	} else /* try and trim the object -- but only if room for a dummy header */
+	    if (new_size_rounded + sizeof(obj_header_t) <= old_size_rounded) {
+		trim_obj(imem, obj, new_size, (chunk_t *)0);
+		new_obj = obj;
+	    }
     if (new_obj) {
 	if_debug8('A', "[a%d:%c%c ]%s %s(%lu=>%lu) 0x%lx\n",
 		  alloc_trace_space(imem),
@@ -802,8 +857,19 @@ i_alloc_string(gs_memory_t * mem, uint nbytes, client_name_t cname)
 {
     gs_ref_memory_t * const imem = (gs_ref_memory_t *)mem;
     byte *str;
+    /*
+     * Cycle through the chunks at the current save level, starting
+     * with the currently open one.
+     */
+    chunk_t *cp_orig = imem->pcc;
 
-top:if (imem->cc.ctop - imem->cc.cbot > nbytes) {
+    if (cp_orig == 0) {
+	/* Open an arbitrary chunk. */
+	cp_orig = imem->pcc = imem->cfirst;
+	alloc_open_chunk(imem);
+    }
+top:
+    if (imem->cc.ctop - imem->cc.cbot > nbytes) {
 	if_debug4('A', "[a%d:+> ]%s(%u) = 0x%lx\n",
 		  alloc_trace_space(imem), client_name_string(cname), nbytes,
 		  (ulong) (imem->cc.ctop - nbytes));
@@ -811,6 +877,18 @@ top:if (imem->cc.ctop - imem->cc.cbot > nbytes) {
 	gs_alloc_fill(str, gs_alloc_fill_alloc, nbytes);
 	return str;
     }
+    /* Try the next chunk. */
+    {
+	chunk_t *cp = imem->cc.cnext;
+
+	alloc_close_chunk(imem);
+	if (cp == 0)
+	    cp = imem->cfirst;
+	imem->pcc = cp;
+	alloc_open_chunk(imem);
+	if (cp != cp_orig)
+	    goto top;
+    }
     if (nbytes > string_space_quanta(max_uint - sizeof(chunk_head_t)) *
 	string_data_quantum
 	) {			/* Can't represent the size in a uint! */
@@ -995,7 +1073,7 @@ large_freelist_alloc(gs_ref_memory_t *mem, uint size)
     uint aligned_max_size =
 	aligned_min_size + obj_align_round(aligned_min_size / 8);
     obj_header_t *best_fit = 0;
-    obj_header_t **best_fit_prev;
+    obj_header_t **best_fit_prev = NULL; /* Initialize against indeteminizm. */
     uint best_fit_size = max_uint;
     obj_header_t *pfree;
     obj_header_t **ppfprev = &mem->freelists[LARGE_FREELIST_INDEX];
@@ -1073,40 +1151,102 @@ alloc_obj(gs_ref_memory_t *mem, ulong lsize, gs_memory_type_ptr_t pstype,
 	cp->cbot += asize;
 	ptr->o_alone = 1;
 	ptr->o_size = lsize;
-    } else if (lsize > max_freelist_size && (flags & ALLOC_DIRECT) &&
-	      (ptr = large_freelist_alloc(mem, lsize)) != 0) {
-	/* We hadn't checked the large block freelist yet. */
-	--ptr;			/* must point to header */
-	goto done;
     } else {
+	/*
+	 * Cycle through the chunks at the current save level, starting
+	 * with the currently open one.
+	 */
+	chunk_t *cp_orig = mem->pcc;
 	uint asize = obj_size_round((uint) lsize);
-	bool consolidate = mem->is_controlled;
-	bool allocate_success = true;
+	bool allocate_success = false;
 
-	while (mem->cc.ctop -
-	       (byte *) (ptr = (obj_header_t *) mem->cc.cbot)
-	       <= asize + sizeof(obj_header_t)) {
-	    if (consolidate) {
+	if (lsize > max_freelist_size && (flags & ALLOC_DIRECT)) {
+	    /* We haven't checked the large block freelist yet. */
+	    if ((ptr = large_freelist_alloc(mem, lsize)) != 0) {
+		--ptr;			/* must point to header */
+		goto done;
+	    }
+	}
+
+	if (cp_orig == 0) {
+	    /* Open an arbitrary chunk. */
+	    cp_orig = mem->pcc = mem->cfirst;
+	    alloc_open_chunk(mem);
+	}
+
+#define CAN_ALLOC_AT_END(cp)\
+  ((cp)->ctop - (byte *) (ptr = (obj_header_t *) (cp)->cbot)\
+   > asize + sizeof(obj_header_t))
+
+	do {
+	    if (CAN_ALLOC_AT_END(&mem->cc)) {
+		allocate_success = true;
+		break;
+	    } else if (mem->is_controlled) {
 		/* Try consolidating free space. */
 		gs_consolidate_free((gs_memory_t *)mem);
-		consolidate = false;
-		continue;
-	    } else {
-		/* Add another chunk. */
-		chunk_t *cp =
-		    alloc_add_chunk(mem, (ulong)mem->chunk_size, "chunk");
-
-		if (cp == 0) {
-		    allocate_success = false;
+		if (CAN_ALLOC_AT_END(&mem->cc)) {
+		    allocate_success = true;
 		    break;
 		}
 	    }
+	    /* No luck, go on to the next chunk. */
+	    {
+		chunk_t *cp = mem->cc.cnext;
+
+		alloc_close_chunk(mem);
+		if (cp == 0)
+		    cp = mem->cfirst;
+		mem->pcc = cp;
+		alloc_open_chunk(mem);
+	    }
+	} while (mem->pcc != cp_orig);
+
+#ifdef CONSOLIDATE_BEFORE_ADDING_CHUNK
+	if (!allocate_success) {
+	    /*
+	     * Try consolidating free space before giving up.
+	     * It's not clear this is a good idea, since it requires quite
+	     * a lot of computation and doesn't seem to improve things much.
+	     */
+	    if (!mem->is_controlled) { /* already did this if controlled */
+		chunk_t *cp = cp_orig;
+
+		alloc_close_chunk(mem);
+		do {
+		    consolidate_chunk_free(cp, mem);
+		    if (CAN_ALLOC_AT_END(cp)) {
+			mem->pcc = cp;
+			alloc_open_chunk(mem);
+			allocate_success = true;
+			break;
+		    }
+		    if ((cp = cp->cnext) == 0)
+			cp = mem->cfirst;
+		} while (cp != cp_orig);
+	    }
 	}
+#endif
+
+#undef CAN_ALLOC_AT_END
+
+	if (!allocate_success) {
+	    /* Add another chunk. */
+	    chunk_t *cp =
+		alloc_add_chunk(mem, (ulong)mem->chunk_size, "chunk");
+			
+	    if (cp) {
+		/* mem->pcc == cp, mem->cc == *mem->pcc. */
+		ptr = (obj_header_t *)cp->cbot;
+		allocate_success = true;
+	    }
+	}
+
 	/*
-	 * If no success, try to scavenge from low free memory. This is only
-	 * enabled for controlled memory (currently only async renderer)
-	 * because it's too much work to prevent it from examining outer
-	 * save levels in the general case.
+	 * If no success, try to scavenge from low free memory. This is
+	 * only enabled for controlled memory (currently only async
+	 * renderer) because it's too much work to prevent it from
+	 * examining outer save levels in the general case.
 	 */
 	if (allocate_success)
 	    mem->cc.cbot = (byte *) ptr + asize;
@@ -1221,7 +1361,7 @@ scavenge_low_free(gs_ref_memory_t *mem, unsigned request_size)
 	    	    break;
 	    } else
 	    	begin_free = 0;
-	END_OBJECTS_SCAN_INCOMPLETE
+	END_OBJECTS_SCAN_NO_ABORT
 
 	/* Found sufficient range of empty memory */
 	if (begin_free != 0 && found_free >= need_free) {
@@ -1400,7 +1540,7 @@ i_register_root(gs_memory_t * mem, gs_gc_root_t * rp, gs_ptr_type_t ptype,
     gs_ref_memory_t * const imem = (gs_ref_memory_t *)mem;
 
     if (rp == NULL) {
-	rp = gs_raw_alloc_struct_immovable(imem->parent, &st_gc_root_t,
+	rp = gs_raw_alloc_struct_immovable(imem->non_gc_memory, &st_gc_root_t,
 					   "i_register_root");
 	if (rp == 0)
 	    return_error(gs_error_VMerror);
@@ -1429,7 +1569,7 @@ i_unregister_root(gs_memory_t * mem, gs_gc_root_t * rp, client_name_t cname)
 	rpp = &(*rpp)->next;
     *rpp = (*rpp)->next;
     if (rp->free_on_unregister)
-	gs_free_object(imem->parent, rp, "i_unregister_root");
+	gs_free_object(imem->non_gc_memory, rp, "i_unregister_root");
 }
 
 /* ================ Chunks ================ */
@@ -1499,7 +1639,7 @@ private chunk_t *
 alloc_acquire_chunk(gs_ref_memory_t * mem, ulong csize, bool has_strings,
 		    client_name_t cname)
 {
-    gs_raw_memory_t *parent = mem->parent;
+    gs_memory_t *parent = mem->non_gc_memory;
     chunk_t *cp;
     byte *cdata;
 
@@ -1509,16 +1649,20 @@ alloc_acquire_chunk(gs_ref_memory_t * mem, ulong csize, bool has_strings,
 	return 0;
 #endif
     cp = gs_raw_alloc_struct_immovable(parent, &st_chunk, cname);
-    if ((ulong) (mem->allocated + mem->inherited) >= mem->limit) {
-	mem->gc_status.requested += csize;
-	if (mem->limit >= mem->gc_status.max_vm ||
-	    mem->gc_status.psignal == 0
-	    )
-	    return 0;
-	if_debug4('0', "[0]signaling space=%d, allocated=%ld, limit=%ld, requested=%ld\n",
-		  mem->space, (long)mem->allocated,
-		  (long)mem->limit, (long)mem->gc_status.requested);
-	*mem->gc_status.psignal = mem->gc_status.signal_value;
+
+    if( mem->gc_status.psignal != 0) {  
+	/* we have a garbage collector */
+	if ((ulong) (mem->allocated + mem->inherited) >= mem->limit) {
+	    mem->gc_status.requested += csize;
+	    if (mem->limit >= mem->gc_status.max_vm) {
+		gs_free_object(parent, cp, cname);
+		return 0;
+	    }
+	    if_debug4('0', "[0]signaling space=%d, allocated=%ld, limit=%ld, requested=%ld\n",
+		      mem->space, (long)mem->allocated,
+		      (long)mem->limit, (long)mem->gc_status.requested);
+	    *mem->gc_status.psignal = mem->gc_status.signal_value;
+	}
     }
     cdata = gs_alloc_bytes_immovable(parent, csize, cname);
     if (cp == 0 || cdata == 0) {
@@ -1663,7 +1807,7 @@ alloc_unlink_chunk(chunk_t * cp, gs_ref_memory_t * mem)
 void
 alloc_free_chunk(chunk_t * cp, gs_ref_memory_t * mem)
 {
-    gs_raw_memory_t *parent = mem->parent;
+    gs_memory_t *parent = mem->non_gc_memory;
     byte *cdata = (byte *)cp->chead;
     ulong csize = (byte *)cp->cend - cdata;
 
@@ -1825,7 +1969,7 @@ debug_dump_contents(const byte * bot, const byte * top, int indent,
 /* Relevant options: type_addresses, no_types, pointers, pointed_strings, */
 /* contents. */
 void
-debug_print_object(const void *obj, const dump_control_t * control)
+debug_print_object(const gs_memory_t *mem, const void *obj, const dump_control_t * control)
 {
     const obj_header_t *pre = ((const obj_header_t *)obj) - 1;
     ulong size = pre_obj_contents_size(pre);
@@ -1860,7 +2004,7 @@ debug_print_object(const void *obj, const dump_control_t * control)
 	gs_ptr_type_t ptype;
 
 	if (proc != gs_no_struct_enum_ptrs)
-	    for (; (ptype = (*proc)(pre + 1, size, index, &eptr, type, NULL)) != 0;
+	    for (; (ptype = (*proc)(mem, pre + 1, size, index, &eptr, type, NULL)) != 0;
 		 ++index
 		) {
 		const void *ptr = eptr.ptr;
@@ -1892,7 +2036,7 @@ debug_print_object(const void *obj, const dump_control_t * control)
 /* Print the contents of a chunk with the given options. */
 /* Relevant options: all. */
 void
-debug_dump_chunk(const chunk_t * cp, const dump_control_t * control)
+debug_dump_chunk(const gs_memory_t *mem, const chunk_t * cp, const dump_control_t * control)
 {
     dprintf1("chunk at 0x%lx:\n", (ulong) cp);
     dprintf3("   chead=0x%lx  cbase=0x%lx sbase=0x%lx\n",
@@ -1923,20 +2067,16 @@ debug_dump_chunk(const chunk_t * cp, const dump_control_t * control)
 				  (const byte *)(pre + 1) + size,
 				  control)
 	)
-	debug_print_object(pre + 1, control);
-/* Temporarily redefine gs_exit so a chunk parsing error */
-/* won't actually exit. */
-#define gs_exit(n) DO_NOTHING
-    END_OBJECTS_SCAN
-#undef gs_exit
+	debug_print_object(mem, pre + 1, control);
+    END_OBJECTS_SCAN_NO_ABORT
 }
 void 
-debug_print_chunk(const chunk_t * cp)
+debug_print_chunk(const gs_memory_t *mem, const chunk_t * cp)
 {
     dump_control_t control;
 
     control = dump_control_default;
-    debug_dump_chunk(cp, &control);
+    debug_dump_chunk(mem, cp, &control);
 }
 
 /* Print the contents of all chunks managed by an allocator. */
@@ -1950,7 +2090,7 @@ debug_dump_memory(const gs_ref_memory_t * mem, const dump_control_t * control)
 	const chunk_t *cp = (mcp == mem->pcc ? &mem->cc : mcp);
 
 	if (obj_in_control_region(cp->cbase, cp->cend, control))
-	    debug_dump_chunk(cp, control);
+	    debug_dump_chunk((const gs_memory_t *)mem, cp, control);
     }
 }
 
@@ -1972,16 +2112,14 @@ debug_find_pointers(const gs_ref_memory_t *mem, const void *target)
 	    enum_ptr_t eptr;
 
 	    if (proc)		/* doesn't trace refs */
-		for (; (*proc)(pre + 1, size, index, &eptr, pre->o_type, NULL); ++index)
+		for (; (*proc)((const gs_memory_t *)mem, pre + 1, size, index, 
+			       &eptr, pre->o_type, NULL); 
+		     ++index)
 		    if (eptr.ptr == target) {
 			dprintf1("Index %d in", index);
-			debug_print_object(pre + 1, &control);
+			debug_print_object((const gs_memory_t *)mem, pre + 1, &control);
 		    }
-/* Temporarily redefine gs_exit so a chunk parsing error */
-/* won't actually exit. */
-#define gs_exit(n) DO_NOTHING
-	END_OBJECTS_SCAN
-#undef gs_exit
+	END_OBJECTS_SCAN_NO_ABORT
     }
 }
 
