@@ -1,28 +1,25 @@
-/* Copyright (C) 1996-2002 artofcode LLC.  All rights reserved.
+/* Copyright (C) 1996-2005 artofcode LLC. All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/* $Id: gscdef.c,v 1.9.2.6 2002/02/11 20:47:56 giles Exp $ */
+/* $Id: gscdef.c,v 1.53 2005/04/12 22:05:40 giles Exp $ */
 /* Configuration scalars */
 
 #include "std.h"
 #include "gscdefs.h"		/* interface */
-#include "gconf.h"		/* for #defines */
+#include "gconfigd.h"		/* for #defines */
 
 /* ---------------- Miscellaneous system parameters ---------------- */
 
@@ -37,19 +34,19 @@ CONFIG_CONST long gs_buildtime = GS_BUILDTIME;
 
 #ifndef GS_COPYRIGHT
 #  define GS_COPYRIGHT\
-	"Copyright (C) 2002 artofcode LLC, Benicia, CA. All rights reserved."
+	"Copyright (C) 2005 artofcode LLC, Benicia, CA.  All rights reserved."
 #endif
 const char *CONFIG_CONST gs_copyright = GS_COPYRIGHT;
 
 #ifndef GS_PRODUCTFAMILY
 #  define GS_PRODUCTFAMILY\
-	"GNU Ghostscript"
+	"AFPL Ghostscript"
 #endif
 const char *CONFIG_CONST gs_productfamily = GS_PRODUCTFAMILY;
 
 #ifndef GS_PRODUCT
 #  define GS_PRODUCT\
-	GS_PRODUCTFAMILY ""
+	GS_PRODUCTFAMILY
 #endif
 const char *CONFIG_CONST gs_product = GS_PRODUCT;
 
