@@ -1,23 +1,20 @@
 /* Copyright (C) 1989, 2000 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: gsdevice.c,v $ $Revision: 1.7.2.1 $ */
+/* $Id: gsdevice.c,v 1.24 2004/11/19 01:44:52 raph Exp $ */
 /* Device operators for Ghostscript library */
 #include "ctype_.h"
 #include "memory_.h"		/* for memchr, memcpy */
@@ -37,6 +34,7 @@
 #include "gxdevice.h"
 #include "gxdevmem.h"
 #include "gxiodev.h"
+#include "gxcspace.h"
 
 /* Include the extern for the device list. */
 extern_gs_lib_device_list();
@@ -55,7 +53,7 @@ gx_device_finalize(void *vptr)
 	dev->finalize(dev);
     discard(gs_closedevice(dev));
     if (dev->stype_is_dynamic)
-	gs_free_const_object(&gs_memory_default, dev->stype,
+	gs_free_const_object(dev->memory->non_gc_memory, dev->stype,
 			     "gx_device_finalize");
 }
 
@@ -253,7 +251,7 @@ gs_copydevice2(gx_device ** pnew_dev, const gx_device * dev, bool keep_open,
 	 * Just allocate a new stype and copy the old one into it.
 	 */
 	a_std = (gs_memory_struct_type_t *)
-	    gs_alloc_bytes_immovable(&gs_memory_default, sizeof(*std),
+	    gs_alloc_bytes_immovable(mem->non_gc_memory, sizeof(*std),
 				     "gs_copydevice(stype)");
 	if (!a_std)
 	    return_error(gs_error_VMerror);
@@ -265,7 +263,7 @@ gs_copydevice2(gx_device ** pnew_dev, const gx_device * dev, bool keep_open,
     } else {
 	/* We need to figure out or adjust the stype. */
 	a_std = (gs_memory_struct_type_t *)
-	    gs_alloc_bytes_immovable(&gs_memory_default, sizeof(*std),
+	    gs_alloc_bytes_immovable(mem->non_gc_memory, sizeof(*std),
 				     "gs_copydevice(stype)");
 	if (!a_std)
 	    return_error(gs_error_VMerror);
@@ -281,6 +279,7 @@ gs_copydevice2(gx_device ** pnew_dev, const gx_device * dev, bool keep_open,
     if (new_dev == 0)
 	return_error(gs_error_VMerror);
     gx_device_init(new_dev, dev, mem, false);
+    gx_device_set_procs(new_dev);
     new_dev->stype = new_std;
     new_dev->stype_is_dynamic = new_std != std;
     /*
@@ -296,7 +295,7 @@ gs_copydevice2(gx_device ** pnew_dev, const gx_device * dev, bool keep_open,
     if (code < 0) {
 	gs_free_object(mem, new_dev, "gs_copydevice(device)");
 	if (a_std)
-	    gs_free_object(&gs_memory_default, a_std, "gs_copydevice(stype)");
+	    gs_free_object(dev->memory->non_gc_memory, a_std, "gs_copydevice(stype)");
 	return code;
     }
     *pnew_dev = new_dev;
@@ -315,6 +314,7 @@ gs_opendevice(gx_device *dev)
 {
     if (dev->is_open)
 	return 0;
+    check_device_separable(dev);
     gx_device_fill_in_procs(dev);
     {
 	int code = (*dev_proc(dev, open_device))(dev);
@@ -401,10 +401,25 @@ gs_setdevice_no_init(gs_state * pgs, gx_device * dev)
     /*
      * Just set the device, possibly changing color space but no other
      * device parameters.
+     * 
+     * Make sure we don't close the device if dev == pgs->device
+     * This could be done by allowing the rc_assign to close the
+     * old 'dev' if the rc goes to 0 (via the device structure's
+     * finalization procedure), but then the 'code' from the dev
+     * closedevice would not be propagated up. We want to allow
+     * the code to be handled, particularly for the pdfwrite
+     * device.
      */
+    if (pgs->device != NULL && pgs->device->rc.ref_count == 1 &&
+	pgs->device != dev) {
+	int code = gs_closedevice(pgs->device);
+
+	if (code < 0)
+	    return code;
+    }
     rc_assign(pgs->device, dev, "gs_setdevice_no_init");
     gs_state_update_device(pgs);
-    return 0;
+    return pgs->overprint ? gs_do_set_overprint(pgs) : 0;
 }
 
 /* Initialize a just-allocated device. */
@@ -426,8 +441,20 @@ gs_make_null_device(gx_device_null *dev_null, gx_device *dev,
     gx_device_init((gx_device *)dev_null, (const gx_device *)&gs_null_device,
 		   mem, true);
     gx_device_set_target((gx_device_forward *)dev_null, dev);
-    if (dev)
-	gx_device_copy_color_params((gx_device *)dev_null, dev);
+    if (dev) {
+	/* The gx_device_copy_color_params() call below should
+	   probably copy over these new-style color mapping procs, as
+	   well as the old-style (map_rgb_color and friends). However,
+	   the change was made here instead, to minimize the potential
+	   impact of the patch.
+	*/
+	gx_device *dn = (gx_device *)dev_null;
+	set_dev_proc(dn, get_color_mapping_procs, gx_forward_get_color_mapping_procs);
+	set_dev_proc(dn, get_color_comp_index, gx_forward_get_color_comp_index);
+	set_dev_proc(dn, encode_color, gx_forward_encode_color);
+	set_dev_proc(dn, decode_color, gx_forward_decode_color);
+	gx_device_copy_color_params(dn, dev);
+    }
 }
 
 /* Mark a device as retained or not retained. */
@@ -472,9 +499,9 @@ gs_closedevice(gx_device * dev)
 
     if (dev->is_open) {
 	code = (*dev_proc(dev, close_device))(dev);
+	dev->is_open = false;
 	if (code < 0)
 	    return_error(code);
-	dev->is_open = false;
     }
     return code;
 }
@@ -536,6 +563,24 @@ gx_device_set_margins(gx_device * dev, const float *margins /*[4] */ ,
     }
 }
 
+
+/* Handle 90 and 270 degree rotation of the Tray
+ * Device must support TrayOrientation in its InitialMatrix and get/put params
+ */
+private void
+gx_device_TrayOrientationRotate(gx_device *dev)
+{
+  if ( dev->TrayOrientation == 90 || dev->TrayOrientation == 270) {
+    /* page sizes don't rotate, height and width do rotate 
+     * HWResolution, HWSize, and MediaSize parameters interact, 
+     * and must be set before TrayOrientation
+     */
+    int tmp = dev->height;
+    dev->height = dev->width;
+    dev->width = tmp;
+  }
+}
+
 /* Set the width and height, updating MediaSize to remain consistent. */
 void
 gx_device_set_width_height(gx_device * dev, int width, int height)
@@ -544,6 +589,7 @@ gx_device_set_width_height(gx_device * dev, int width, int height)
     dev->height = height;
     dev->MediaSize[0] = width * 72.0 / dev->HWResolution[0];
     dev->MediaSize[1] = height * 72.0 / dev->HWResolution[1];
+    gx_device_TrayOrientationRotate(dev);
 }
 
 /* Set the resolution, updating width and height to remain consistent. */
@@ -552,8 +598,9 @@ gx_device_set_resolution(gx_device * dev, floatp x_dpi, floatp y_dpi)
 {
     dev->HWResolution[0] = x_dpi;
     dev->HWResolution[1] = y_dpi;
-    dev->width = dev->MediaSize[0] * x_dpi / 72.0 + 0.5;
-    dev->height = dev->MediaSize[1] * y_dpi / 72.0 + 0.5;
+    dev->width = (int)(dev->MediaSize[0] * x_dpi / 72.0 + 0.5);
+    dev->height = (int)(dev->MediaSize[1] * y_dpi / 72.0 + 0.5);
+    gx_device_TrayOrientationRotate(dev);
 }
 
 /* Set the MediaSize, updating width and height to remain consistent. */
@@ -562,8 +609,9 @@ gx_device_set_media_size(gx_device * dev, floatp media_width, floatp media_heigh
 {
     dev->MediaSize[0] = media_width;
     dev->MediaSize[1] = media_height;
-    dev->width = media_width * dev->HWResolution[0] / 72.0 + 0.499;
-    dev->height = media_height * dev->HWResolution[1] / 72.0 + 0.499;
+    dev->width = (int)(media_width * dev->HWResolution[0] / 72.0 + 0.499);
+    dev->height = (int)(media_height * dev->HWResolution[1] / 72.0 + 0.499);
+    gx_device_TrayOrientationRotate(dev);
 }
 
 /*
@@ -580,6 +628,11 @@ gx_device_copy_color_procs(gx_device *dev, const gx_device *target)
     dev_proc_map_color_rgb((*to_rgb)) =
 	dev_proc(dev, map_color_rgb);
 
+    /* The logic in this function seems a bit stale; it sets the
+       old-style color procs, but not the new ones
+       (get_color_mapping_procs, get_color_comp_index, encode_color,
+       and decode_color). It should probably copy those as well.
+    */
     if (from_cmyk == gx_forward_map_cmyk_color ||
 	from_cmyk == cmyk_1bit_map_cmyk_color ||
 	from_cmyk == cmyk_8bit_map_cmyk_color) {
@@ -646,6 +699,63 @@ gx_device_copy_params(gx_device *dev, const gx_device *target)
 
 #undef COPY_PARAM
 
+/*
+ * Parse the output file name detecting and validating any %nnd format
+ * for inserting the page count.  If a format is present, store a pointer
+ * to its last character in *pfmt, otherwise store 0 there.
+ * Note that we assume devices have already been scanned, and any % must
+ * precede a valid format character.
+ *
+ * If there was a format, then return the max_width
+ */
+private int
+gx_parse_output_format(gs_parsed_file_name_t *pfn, const char **pfmt)
+{
+    bool have_format = false, field = 0;
+    int width[2], int_width = sizeof(int) * 3, w = 0;
+    uint i;
+
+    /* Scan the file name for a format string, and validate it if present. */
+    width[0] = width[1] = 0;
+    for (i = 0; i < pfn->len; ++i)
+	if (pfn->fname[i] == '%') {
+	    if (i + 1 < pfn->len && pfn->fname[i + 1] == '%')
+		continue;
+	    if (have_format)	/* more than one % */
+		return_error(gs_error_undefinedfilename);
+	    have_format = true;
+	sw:
+	    if (++i == pfn->len)
+		return_error(gs_error_undefinedfilename);
+	    switch (pfn->fname[i]) {
+		case 'l':
+		    int_width = sizeof(long) * 3;
+		case ' ': case '#': case '+': case '-':
+		    goto sw;
+		case '.':
+		    if (field)
+			return_error(gs_error_undefinedfilename);
+		    field = 1;
+		    continue;
+		case '0': case '1': case '2': case '3': case '4':
+		case '5': case '6': case '7': case '8': case '9':
+		    width[field] = width[field] * 10 + pfn->fname[i] - '0';
+		    goto sw;
+		case 'd': case 'i': case 'u': case 'o': case 'x': case 'X':
+		    *pfmt = &pfn->fname[i];
+		    continue;
+		default:
+		    return_error(gs_error_undefinedfilename);
+	    }
+	}
+    if (have_format) {
+	/* Calculate a conservative maximum width. */
+	w = max(width[0], width[1]);
+	w = max(w, int_width) + 5;
+    }
+    return w;
+}
+
 /*
  * Parse the output file name for a device, recognizing "-" and "|command",
  * and also detecting and validating any %nnd format for inserting the
@@ -658,35 +768,31 @@ gx_parse_output_file_name(gs_parsed_file_name_t *pfn, const char **pfmt,
 			  const char *fname, uint fnlen)
 {
     int code;
-    bool have_format = false, field = 0;
-    int width[2], int_width = sizeof(int) * 3, w = 0;
-    uint i;
 
     *pfmt = 0;
-    if (fnlen == 0) {		/* allow null name */
-	pfn->memory = 0;
-	pfn->iodev = NULL;
-	pfn->fname = 0;		/* irrelevant since length = 0 */
-	pfn->len = 0;
+    pfn->memory = 0;
+    pfn->iodev = NULL;
+    pfn->fname = NULL;		/* irrelevant since length = 0 */
+    pfn->len = 0;
+    if (fnlen == 0)  		/* allow null name */
 	return 0;
-    }
     /*
      * If the file name begins with a %, it might be either an IODevice
-     * or a %nnd format.  Distinguish the two by searching for a second %.
+     * or a %nnd format.  Check (carefully) for this case.
      */
-    if (fname[0] == '%' && !memchr(fname + 1, '%', fnlen - 1)) {
-	/* This is a file name starting with a format. */
-	pfn->memory = 0;
-	pfn->iodev = NULL;
-	pfn->fname = fname;
-	pfn->len = fnlen;
-    } else {
-	code = gs_parse_file_name(pfn, fname, fnlen);
+    code = gs_parse_file_name(pfn, fname, fnlen);
+    if (code < 0) {
+	if (fname[0] == '%') {
+	    /* not a recognized iodev -- may be a leading format descriptor */
+	    pfn->len = fnlen;
+	    pfn->fname = fname;
+	    code = gx_parse_output_format(pfn, pfmt);
+	} 
 	if (code < 0)
 	    return code;
     }
     if (!pfn->iodev) {
-	if (!strcmp(pfn->fname, "-")) {
+	if ( (pfn->len == 1) && (pfn->fname[0] == '-') ) {
 	    pfn->iodev = gs_findiodevice((const byte *)"%stdout", 7);
 	    pfn->fname = NULL;
 	} else if (pfn->fname[0] == '|') {
@@ -699,46 +805,11 @@ gx_parse_output_file_name(gs_parsed_file_name_t *pfn, const char **pfmt,
     }
     if (!pfn->fname)
 	return 0;
-    /* Scan the file name for a format string, and validate it if present. */
-    width[0] = width[1] = 0;
-    for (i = 0; i < pfn->len; ++i)
-	if (pfn->fname[i] == '%') {
-	    if (i + 1 < pfn->len && pfn->fname[i + 1] == '%')
-		continue;
-	    if (have_format)	/* more than one % */
-		return_error(gs_error_rangecheck);
-	    have_format = true;
-	sw:
-	    if (++i == pfn->len)
-		return_error(gs_error_rangecheck);
-	    switch (pfn->fname[i]) {
-		case 'l':
-		    int_width = sizeof(long) * 3;
-		case ' ': case '#': case '+': case '-':
-		    goto sw;
-		case '.':
-		    if (field)
-			return_error(gs_error_rangecheck);
-		    field = 1;
-		    continue;
-		case '0': case '1': case '2': case '3': case '4':
-		case '5': case '6': case '7': case '8': case '9':
-		    width[field] = width[field] * 10 + pfn->fname[i] - '0';
-		    goto sw;
-		case 'd': case 'i': case 'u': case 'o': case 'x': case 'X':
-		    *pfmt = &pfn->fname[i];
-		    continue;
-		default:
-		    return_error(gs_error_rangecheck);
-	    }
-	}
-    if (have_format) {
-	/* Calculate a conservative maximum width. */
-	w = max(width[0], width[1]);
-	w = max(w, int_width) + 5;
-    }
-    if (strlen(pfn->iodev->dname) + pfn->len + w >= gp_file_name_sizeof)
-	return_error(gs_error_rangecheck);
+    code = gx_parse_output_format(pfn, pfmt);
+    if (code < 0)
+        return code;
+    if (strlen(pfn->iodev->dname) + pfn->len + code >= gp_file_name_sizeof)
+	return_error(gs_error_undefinedfilename);
     return 0;
 }
 
@@ -757,7 +828,7 @@ gx_device_open_output_file(const gx_device * dev, char *fname,
     if (parsed.iodev && !strcmp(parsed.iodev->dname, "%stdout%")) {
 	if (parsed.fname)
 	    return_error(gs_error_undefinedfilename);
-	*pfile = gs_stdout;
+	*pfile = dev->memory->gs_lib_ctx->fstdout;
 	/* Force stdout to binary. */
 	return gp_setmode_binary(*pfile, true);
     }
@@ -779,16 +850,18 @@ gx_device_open_output_file(const gx_device * dev, char *fname,
 	if (!parsed.fname)
 	    return_error(gs_error_undefinedfilename);
 	strcpy(fmode, gp_fmode_wb);
-	if (positionable)
-	    strcat(fmode, "+");
-	code = parsed.iodev->procs.fopen(parsed.iodev, parsed.fname, fmode,
-					 pfile, NULL, 0);
-	if (*pfile)
-	    return 0;
+  	if (positionable)
+  	    strcat(fmode, "+");
+ 	code = parsed.iodev->procs.fopen(parsed.iodev, parsed.fname, fmode,
+  					 pfile, NULL, 0);
+ 	if (code)
+     	    eprintf1("**** Could not open the file %s .\n", parsed.fname);
+ 	return code;
     }
     *pfile = gp_open_printer((fmt ? pfname : fname), binary);
     if (*pfile)
-	return 0;
+  	return 0;
+    eprintf1("**** Could not open the file %s .\n", (fmt ? pfname : fname));
     return_error(gs_error_invalidfileaccess);
 }
 
