@@ -1,4 +1,4 @@
-/* Copyright (C) 1997, 2000 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1997, 2000, 2002 Aladdin Enterprises.  All rights reserved.
   
   This software is provided AS-IS with no warranty, either express or
   implied.
@@ -14,12 +14,19 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: gsfunc.h,v 1.4.6.1 2002/02/22 19:45:57 ray Exp $ */
+/* $Id: gsfunc.h,v 1.11 2004/03/07 12:06:11 igor Exp $ */
 /* Generic definitions for Functions */
 
 #ifndef gsfunc_INCLUDED
 #  define gsfunc_INCLUDED
 
+#include "gstypes.h"		/* for gs_range_t */
+
+#ifndef stream_DEFINED
+#  define stream_DEFINED
+typedef struct stream_s stream;
+#endif
+
 /* ---------------- Types and structures ---------------- */
 
 /*
@@ -40,13 +47,6 @@ typedef int gs_function_type_t;
     int n;			/* # of outputs */\
     const float *Range		/* 2 x n, optional except for type 0 */
 
-/* Define calculation effort values (currently only used for monotonicity). */
-typedef enum {
-    EFFORT_EASY = 0,
-    EFFORT_MODERATE = 1,
-    EFFORT_ESSENTIAL = 2
-} gs_function_effort_t;
-
 /* Define abstract types. */
 #ifndef gs_data_source_DEFINED
 #  define gs_data_source_DEFINED
@@ -74,43 +74,65 @@ typedef struct gs_function_info_s {
 
 /* Evaluate a function. */
 #define FN_EVALUATE_PROC(proc)\
-  int proc(P3(const gs_function_t * pfn, const float *in, float *out))
+  int proc(const gs_function_t * pfn, const float *in, float *out)
 typedef FN_EVALUATE_PROC((*fn_evaluate_proc_t));
 
 /* Test whether a function is monotonic. */
 #define FN_IS_MONOTONIC_PROC(proc)\
-  int proc(P4(const gs_function_t * pfn, const float *lower,\
-	      const float *upper, gs_function_effort_t effort))
+  int proc(const gs_function_t * pfn, const float *lower,\
+	   const float *upper)
 typedef FN_IS_MONOTONIC_PROC((*fn_is_monotonic_proc_t));
 
 /* Get function information. */
 #define FN_GET_INFO_PROC(proc)\
-  void proc(P2(const gs_function_t *pfn, gs_function_info_t *pfi))
+  void proc(const gs_function_t *pfn, gs_function_info_t *pfi)
 typedef FN_GET_INFO_PROC((*fn_get_info_proc_t));
 
 /* Put function parameters on a parameter list. */
 #define FN_GET_PARAMS_PROC(proc)\
-  int proc(P2(const gs_function_t *pfn, gs_param_list *plist))
+  int proc(const gs_function_t *pfn, gs_param_list *plist)
 typedef FN_GET_PARAMS_PROC((*fn_get_params_proc_t));
 
+/*
+ * Create a new function with scaled output.  The i'th output value is
+ * transformed linearly so that [0 .. 1] are mapped to [pranges[i].rmin ..
+ * pranges[i].rmax].  Any necessary parameters or subfunctions of the
+ * original function are copied, not shared, even if their values aren't
+ * changed, so that the new function can be freed without having to worry
+ * about freeing data that should be kept.  Note that if there is a "data
+ * source", it is shared, not copied: this should not be a problem, since
+ * gs_function_free does not free the data source.
+ */
+#define FN_MAKE_SCALED_PROC(proc)\
+  int proc(const gs_function_t *pfn, gs_function_t **ppsfn,\
+	   const gs_range_t *pranges, gs_memory_t *mem)
+typedef FN_MAKE_SCALED_PROC((*fn_make_scaled_proc_t));
+
 /* Free function parameters. */
 #define FN_FREE_PARAMS_PROC(proc)\
-  void proc(P2(gs_function_params_t * params, gs_memory_t * mem))
+  void proc(gs_function_params_t * params, gs_memory_t * mem)
 typedef FN_FREE_PARAMS_PROC((*fn_free_params_proc_t));
 
 /* Free a function. */
 #define FN_FREE_PROC(proc)\
-  void proc(P3(gs_function_t * pfn, bool free_params, gs_memory_t * mem))
+  void proc(gs_function_t * pfn, bool free_params, gs_memory_t * mem)
 typedef FN_FREE_PROC((*fn_free_proc_t));
 
+/* Serialize a function. */
+#define FN_SERIALIZE_PROC(proc)\
+  int proc(const gs_function_t * pfn, stream *s)
+typedef FN_SERIALIZE_PROC((*fn_serialize_proc_t));
+
 /* Define the generic function structures. */
 typedef struct gs_function_procs_s {
     fn_evaluate_proc_t evaluate;
     fn_is_monotonic_proc_t is_monotonic;
     fn_get_info_proc_t get_info;
     fn_get_params_proc_t get_params;
+    fn_make_scaled_proc_t make_scaled;
     fn_free_params_proc_t free_params;
     fn_free_proc_t free;
+    fn_serialize_proc_t serialize;
 } gs_function_procs_t;
 typedef struct gs_function_head_s {
     gs_function_type_t type;
@@ -149,35 +171,32 @@ typedef struct gs_function_XxYy_params_s {
  * header file, one to allocate and initialize an instance of that type,
  * and one to free the parameters of that type.
 
-int gs_function_XxYy_init(P3(gs_function_t **ppfn,
-			     const gs_function_XxYy_params_t *params,
-			     gs_memory_t *mem));
+int gs_function_XxYy_init(gs_function_t **ppfn,
+			  const gs_function_XxYy_params_t *params,
+			  gs_memory_t *mem));
 
-void gs_function_XxYy_free_params(P2(gs_function_XxYy_params_t *params,
-				     gs_memory_t *mem));
+void gs_function_XxYy_free_params(gs_function_XxYy_params_t *params,
+				  gs_memory_t *mem);
 
  */
 
+/* Allocate an array of function pointers. */
+int alloc_function_array(uint count, gs_function_t *** pFunctions,
+			 gs_memory_t *mem);
+
 /* Evaluate a function. */
 #define gs_function_evaluate(pfn, in, out)\
   ((pfn)->head.procs.evaluate)(pfn, in, out)
 
 /*
- * Test whether a function is monotonic on a given (closed) interval.  If
- * the test requires too much effort, the procedure may return
- * gs_error_undefined; normally, it returns 0 for false, >0 for true,
- * gs_error_rangecheck if any part of the interval is outside the function's
- * domain.  If lower[i] > upper[i], the result is not defined.
+ * Test whether a function is monotonic on a given (closed) interval.
+ * return 1 if true, 0 if don't know, <0 on error.
+ * If lower[i] > upper[i], the result is not defined.
+ * NOTE : currently it is underimplemented for cubic interpolation functions.
+ *        Need to find the cubic surface extremes.
  */
-#define gs_function_is_monotonic(pfn, lower, upper, effort)\
-  ((pfn)->head.procs.is_monotonic)(pfn, lower, upper, effort)
-/*
- * If the function is monotonic, is_monotonic returns the direction of
- * monotonicity for output value N in bits 2N and 2N+1.  (Functions with
- * more than sizeof(int) * 4 - 1 outputs are never identified as monotonic.)
- */
-#define FN_MONOTONIC_INCREASING 1
-#define FN_MONOTONIC_DECREASING 2
+#define gs_function_is_monotonic(pfn, lower, upper)\
+  ((pfn)->head.procs.is_monotonic)(pfn, lower, upper)
 
 /* Get function information. */
 #define gs_function_get_info(pfn, pfi)\
@@ -187,6 +206,10 @@ void gs_function_XxYy_free_params(P2(gs_function_XxYy_params_t *params,
 #define gs_function_get_params(pfn, plist)\
   ((pfn)->head.procs.get_params(pfn, plist))
 
+/* Create a scaled function. */
+#define gs_function_make_scaled(pfn, ppsfn, pranges, mem)\
+  ((pfn)->head.procs.make_scaled(pfn, ppsfn, pranges, mem))
+
 /* Free function parameters. */
 #define gs_function_free_params(pfn, mem)\
   ((pfn)->head.procs.free_params(&(pfn)->params, mem))
@@ -195,4 +218,8 @@ void gs_function_XxYy_free_params(P2(gs_function_XxYy_params_t *params,
 #define gs_function_free(pfn, free_params, mem)\
   ((pfn)->head.procs.free(pfn, free_params, mem))
 
+/* Serialize a function. */
+#define gs_function_serialize(pfn, s)\
+  ((pfn)->head.procs.serialize(pfn, s))
+
 #endif /* gsfunc_INCLUDED */
