@@ -1,23 +1,20 @@
-/* Copyright (C) 2000 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 2000, 2001 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: gsfunc4.c,v $ $Revision: 1.3.2.2 $ */
+/* $Id: gsfunc4.c,v 1.16 2004/10/07 05:18:34 ray Exp $ */
 /* Implementation of FunctionType 4 (PostScript Calculator) Functions */
 #include "math_.h"
 #include "memory_.h"
@@ -31,6 +28,7 @@
 #include "strimpl.h"
 #include "sfilter.h"		/* for SubFileDecode */
 #include "spprint.h"
+#include "stream.h"
 
 typedef struct gs_function_PtCr_s {
     gs_function_head_t head;
@@ -61,7 +59,7 @@ typedef struct calc_value_s {
 } calc_value_t;
 
 /* Store a float. */
-private void
+private inline void
 store_float(calc_value_t *vsp, floatp f)
 {
     vsp->value.f = f;
@@ -107,7 +105,8 @@ private int
 fn_PtCr_evaluate(const gs_function_t *pfn_common, const float *in, float *out)
 {
     const gs_function_PtCr_t *pfn = (const gs_function_PtCr_t *)pfn_common;
-    calc_value_t vstack[1 + MAX_VSTACK + 1];
+    calc_value_t vstack_buf[2 + MAX_VSTACK + 1];
+    calc_value_t *vstack = &vstack_buf[1];
     calc_value_t *vsp = vstack + pfn->params.m;
     const byte *p = pfn->params.ops.data;
     int i;
@@ -216,6 +215,7 @@ fn_PtCr_evaluate(const gs_function_t *pfn_common, const float *in, float *out)
 
     };
 
+    vstack[-1].type = CVT_NONE;  /* for type dispatch in empty stack case */
     vstack[0].type = CVT_NONE;	/* catch underflow */
     for (i = 0; i < pfn->params.m; ++i)
 	store_float(&vstack[i + 1], in[i]);
@@ -292,7 +292,7 @@ fn_PtCr_evaluate(const gs_function_t *pfn_common, const float *in, float *out)
 	    vsp->value.f = ceil(vsp->value.f);
 	    continue;
 	case PtCr_cos:
-	    vsp->value.f = cos(vsp->value.f);
+	    vsp->value.f = gs_cos_degrees(vsp->value.f);
 	    continue;
 	case PtCr_cvi:
 	    vsp->value.i = (int)(vsp->value.f);
@@ -365,7 +365,7 @@ fn_PtCr_evaluate(const gs_function_t *pfn_common, const float *in, float *out)
 	    vsp->value.f = floor(vsp->value.f + 0.5);
 	    continue;
 	case PtCr_sin:
-	    vsp->value.f = sin(vsp->value.f);
+	    vsp->value.f = gs_sin_degrees(vsp->value.f);
 	    continue;
 	case PtCr_sqrt:
 	    vsp->value.f = sqrt(vsp->value.f);
@@ -530,7 +530,7 @@ fn_PtCr_evaluate(const gs_function_t *pfn_common, const float *in, float *out)
     for (i = 0; i < pfn->params.n; ++i) {
 	switch (vstack[i + 1].type) {
 	case CVT_INT:
-	    out[i] = vstack[i + 1].value.i;
+	    out[i] = (float)vstack[i + 1].value.i;
 	    break;
 	case CVT_FLOAT:
 	    out[i] = vstack[i + 1].value.f;
@@ -545,8 +545,7 @@ fn_PtCr_evaluate(const gs_function_t *pfn_common, const float *in, float *out)
 /* Test whether a PostScript Calculator function is monotonic. */
 private int
 fn_PtCr_is_monotonic(const gs_function_t * pfn_common,
-		     const float *lower, const float *upper,
-		     gs_function_effort_t effort)
+		     const float *lower, const float *upper)
 {
     /*
      * No reasonable way to tell.  Eventually we should check for
@@ -608,10 +607,12 @@ calc_put_ops(stream *s, const byte *ops, uint size)
 		stream_puts(s, " ifelse ");
 	    } else
 		stream_puts(s, " if ");
+	    break;
 	}
 	case PtCr_else:
 	    if (p != ops + size - 2)
 		return_error(gs_error_rangecheck);
+	    spputc(s, '}');
 	    return 1;
 	/*case PtCr_return:*/	/* not possible */
 	default: {		/* must be < PtCr_NUM_OPS */
@@ -647,18 +648,30 @@ calc_access(const gs_data_source_t *psrc, ulong start, uint length,
     const gs_function_PtCr_t *const pfn =
 	(const gs_function_PtCr_t *)
 	  ((const char *)psrc - offset_of(gs_function_PtCr_t, data_source));
+    /*
+     * The caller wants a specific substring of the symbolic definition.
+     * Generate the entire definition, using a SubFileDecode filter (in an
+     * output pipeline!) to extract the substring.  This is very
+     * inefficient, but this code is rarely used, and almost never actually
+     * has to break up the definition into pieces to fit in the caller's
+     * buffer.
+     */
     stream_SFD_state st;
-    stream s;
+    stream ds, bs;
+    byte dbuf[200];		/* arbitrary */
     const stream_template *const template = &s_SFD_template;
 
+    /* Set up the stream that writes into the buffer. */
+    s_init(&bs, NULL);
+    swrite_string(&bs, buf, length);
+    /* Set up the SubFileDecode stream. */
+    s_init(&ds, NULL);
+    s_init_state((stream_state *)&st, template, NULL);
     template->set_defaults((stream_state *)&st);
     st.skip_count = start;
-    swrite_string(&s, buf, length);
-    s.procs.process = template->process;
-    s.state = (stream_state *)&st;
-    if (template->init)
-	template->init((stream_state *)&st);
-    calc_put(&s, pfn);
+    s_init_filter(&ds, (stream_state *)&st, dbuf, sizeof(dbuf), &bs);
+    calc_put(&ds, pfn);
+    sclose(&ds);
     if (ptr)
 	*ptr = buf;
     return 0;
@@ -676,12 +689,76 @@ fn_PtCr_get_info(const gs_function_t *pfn_common, gs_function_info_t *pfi)
     {
 	stream s;
 
+	s_init(&s, NULL);
 	swrite_position_only(&s);
 	calc_put(&s, pfn);
 	pfi->data_size = stell(&s);
     }
 }
 
+/* Make a scaled copy of a PostScript Calculator function. */
+private int
+fn_PtCr_make_scaled(const gs_function_PtCr_t *pfn, gs_function_PtCr_t **ppsfn,
+		    const gs_range_t *pranges, gs_memory_t *mem)
+{
+    gs_function_PtCr_t *psfn =
+	gs_alloc_struct(mem, gs_function_PtCr_t, &st_function_PtCr,
+			"fn_PtCr_make_scaled");
+    /* We are adding {<int> 1 roll <float> mul <float> add} for each output. */
+    int n = pfn->params.n;
+    uint opsize = pfn->params.ops.size + (9 + 2 * sizeof(float)) * n;
+    byte *ops = gs_alloc_string(mem, opsize, "fn_PtCr_make_scaled(ops)");
+    byte *p;
+    int code, i;
+
+    if (psfn == 0 || ops == 0) {
+	gs_free_string(mem, ops, opsize, "fn_PtCr_make_scaled(ops)");
+	gs_free_object(mem, psfn, "fn_PtCr_make_scaled");
+	return_error(gs_error_VMerror);
+    }
+    psfn->params = pfn->params;
+    psfn->params.ops.data = ops;
+    psfn->params.ops.size = opsize;
+    psfn->data_source = pfn->data_source;
+    code = fn_common_scale((gs_function_t *)psfn, (const gs_function_t *)pfn,
+			   pranges, mem);
+    if (code < 0) {
+	gs_function_free((gs_function_t *)psfn, true, mem);
+	return code;
+    }
+    memcpy(ops, pfn->params.ops.data, pfn->params.ops.size - 1); /* minus return */
+    p = ops + pfn->params.ops.size - 1;
+    for (i = n; --i >= 0; ) {
+	float base = pranges[i].rmin;
+	float factor = pranges[i].rmax - base;
+
+	if (factor != 1) {
+	    p[0] = PtCr_float; memcpy(p + 1, &factor, sizeof(float));
+	    p += 1 + sizeof(float);
+	    *p++ = PtCr_mul;
+	}
+	if (base != 0) {
+	    p[0] = PtCr_float; memcpy(p + 1, &base, sizeof(float));
+	    p += 1 + sizeof(float);
+	    *p++ = PtCr_add;
+	}
+	if (n != 1) {
+	    p[0] = PtCr_byte; p[1] = (byte)n;
+	    p[2] = PtCr_byte; p[3] = 1;
+	    p[4] = PtCr_roll;
+	    p += 5;
+	}
+    }
+    *p++ = PtCr_return;
+    psfn->params.ops.size = p - ops;
+    psfn->params.ops.data =
+	gs_resize_string(mem, ops, opsize, psfn->params.ops.size,
+			 "fn_PtCr_make_scaled");
+    *ppsfn = psfn;
+    return 0;
+}
+
+
 /* Free the parameters of a PostScript Calculator function. */
 void
 gs_function_PtCr_free_params(gs_function_PtCr_params_t * params, gs_memory_t * mem)
@@ -690,6 +767,22 @@ gs_function_PtCr_free_params(gs_function_PtCr_params_t * params, gs_memory_t * m
     fn_common_free_params((gs_function_params_t *) params, mem);
 }
 
+/* Serialize. */
+private int
+gs_function_PtCr_serialize(const gs_function_t * pfn, stream *s)
+{
+    uint n;
+    const gs_function_PtCr_params_t * p = (const gs_function_PtCr_params_t *)&pfn->params;
+    int code = fn_common_serialize(pfn, s);
+
+    if (code < 0)
+	return code;
+    code = sputs(s, (const byte *)&p->ops.size, sizeof(p->ops.size), &n);
+    if (code < 0)
+	return code;
+    return sputs(s, p->ops.data, p->ops.size, &n);
+}
+
 /* Allocate and initialize a PostScript Calculator function. */
 int
 gs_function_PtCr_init(gs_function_t ** ppfn,
@@ -701,9 +794,11 @@ gs_function_PtCr_init(gs_function_t ** ppfn,
 	    (fn_evaluate_proc_t) fn_PtCr_evaluate,
 	    (fn_is_monotonic_proc_t) fn_PtCr_is_monotonic,
 	    (fn_get_info_proc_t) fn_PtCr_get_info,
-	    (fn_get_params_proc_t) fn_common_get_params,
+	    fn_common_get_params,
+	    (fn_make_scaled_proc_t) fn_PtCr_make_scaled,
 	    (fn_free_params_proc_t) gs_function_PtCr_free_params,
-	    fn_common_free
+	    fn_common_free,
+	    (fn_serialize_proc_t) gs_function_PtCr_serialize,
 	}
     };
     int code;
@@ -759,7 +854,7 @@ gs_function_PtCr_init(gs_function_t ** ppfn,
 	pfn->data_source.access = calc_access;
 	pfn->head = function_PtCr_head;
 	pfn->head.is_monotonic =
-	    fn_domain_is_monotonic((gs_function_t *)pfn, EFFORT_MODERATE);
+	    fn_domain_is_monotonic((gs_function_t *)pfn);
 	*ppfn = (gs_function_t *) pfn;
     }
     return 0;
