@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: gsmemret.c,v 1.2.6.1 2002/02/22 19:45:58 ray Exp $ */
+/* $Id: gsmemret.c,v 1.5 2004/08/04 19:36:12 stefan Exp $ */
 /* Retrying memory allocator */
 
 #include "gx.h"
@@ -94,6 +94,8 @@ gs_memory_retrying_init(
     rmem->stable_memory = 0;
     rmem->procs = retrying_procs;
     rmem->target = target;
+    rmem->gs_lib_ctx = target->gs_lib_ctx;
+    rmem->non_gc_memory = (gs_memory_t *)rmem;
     gs_memory_retrying_set_recover(rmem, no_recover_proc, NULL);
     return 0;
 }
