@@ -1,23 +1,20 @@
 /* Copyright (C) 1995, 2000 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: gxalloc.h,v $ $Revision: 1.5.2.1 $ */
+/* $Id: gxalloc.h,v 1.11 2004/08/04 19:36:12 stefan Exp $ */
 /* Structure definitions for standard allocator */
 /* Requires gsmemory.h, gsstruct.h */
 
@@ -202,6 +199,8 @@ extern_st(st_chunk);
  *      DO_ALL
  *              << code for all objects -- size is set >>
  *      END_OBJECTS_SCAN
+ *
+ * NB on error END_OBJECTS_SCAN calls gs_abort in debug systems.
  */
 #define SCAN_CHUNK_OBJECTS(cp)\
 	{	obj_header_t *pre = (obj_header_t *)((cp)->cbase);\
@@ -215,7 +214,7 @@ extern_st(st_chunk);
 #define DO_ALL\
 			size = pre_obj_contents_size(pre);\
 			{
-#define END_OBJECTS_SCAN_INCOMPLETE\
+#define END_OBJECTS_SCAN_NO_ABORT\
 			}\
 		}\
 	}
@@ -226,19 +225,19 @@ extern_st(st_chunk);
 		if ( pre != end )\
 		{	lprintf2("Chunk parsing error, 0x%lx != 0x%lx\n",\
 				 (ulong)pre, (ulong)end);\
-			gs_exit(1);\
+		    /*gs_abort((const gs_memory_t *)NULL);*/	\
 		}\
 	}
 #else
-#  define END_OBJECTS_SCAN END_OBJECTS_SCAN_INCOMPLETE
+#  define END_OBJECTS_SCAN END_OBJECTS_SCAN_NO_ABORT
 #endif
 
 /* Initialize a chunk. */
 /* This is exported for save/restore. */
-void alloc_init_chunk(P5(chunk_t *, byte *, byte *, bool, chunk_t *));
+void alloc_init_chunk(chunk_t *, byte *, byte *, bool, chunk_t *);
 
 /* Initialize the string freelists in a chunk. */
-void alloc_init_free_strings(P1(chunk_t *));
+void alloc_init_free_strings(chunk_t *);
 
 /* Find the chunk for a pointer. */
 /* Note that ptr_is_within_chunk returns true even if the pointer */
@@ -254,7 +253,7 @@ typedef struct chunk_locator_s {
     const gs_ref_memory_t *memory;	/* for head & tail of chain */
     chunk_t *cp;		/* one-element cache */
 } chunk_locator_t;
-bool chunk_locate_ptr(P2(const void *, chunk_locator_t *));
+bool chunk_locate_ptr(const void *, chunk_locator_t *);
 
 #define chunk_locate(ptr, clp)\
   (((clp)->cp != 0 && ptr_is_in_chunk(ptr, (clp)->cp)) ||\
@@ -262,18 +261,18 @@ bool chunk_locate_ptr(P2(const void *, chunk_locator_t *));
 
 /* Close up the current chunk. */
 /* This is exported for save/restore and for the GC. */
-void alloc_close_chunk(P1(gs_ref_memory_t * mem));
+void alloc_close_chunk(gs_ref_memory_t * mem);
 
 /* Reopen the current chunk after a GC. */
-void alloc_open_chunk(P1(gs_ref_memory_t * mem));
+void alloc_open_chunk(gs_ref_memory_t * mem);
 
 /* Insert or remove a chunk in the address-ordered chain. */
 /* These are exported for the GC. */
-void alloc_link_chunk(P2(chunk_t *, gs_ref_memory_t *));
-void alloc_unlink_chunk(P2(chunk_t *, gs_ref_memory_t *));
+void alloc_link_chunk(chunk_t *, gs_ref_memory_t *);
+void alloc_unlink_chunk(chunk_t *, gs_ref_memory_t *);
 
 /* Free a chunk.  This is exported for save/restore and for the GC. */
-void alloc_free_chunk(P2(chunk_t *, gs_ref_memory_t *));
+void alloc_free_chunk(chunk_t *, gs_ref_memory_t *);
 
 /* Print a chunk debugging message. */
 /* Unfortunately, the ANSI C preprocessor doesn't allow us to */
@@ -334,7 +333,6 @@ typedef struct ref_s ref;
 struct gs_ref_memory_s {
     /* The following are set at initialization time. */
     gs_memory_common;
-    gs_raw_memory_t *parent;	/* for allocating chunks */
     uint chunk_size;
     uint large_size;		/* min size to give large object */
 				/* its own chunk: must be */
@@ -454,20 +452,20 @@ extern const dump_control_t dump_control_all;
 /* Print one object with the given options. */
 /* Relevant options: type_addresses, no_types, pointers, pointed_strings, */
 /* contents. */
-void debug_print_object(P2(const void *obj, const dump_control_t * control));
+void debug_print_object(const gs_memory_t *mem, const void *obj, const dump_control_t * control);
 
 /* Print the contents of a chunk with the given options. */
 /* Relevant options: all. */
-void debug_dump_chunk(P2(const chunk_t * cp, const dump_control_t * control));
-void debug_print_chunk(P1(const chunk_t * cp));	/* default options */
+void debug_dump_chunk(const gs_memory_t *mem, const chunk_t * cp, const dump_control_t * control);
+void debug_print_chunk(const gs_memory_t *mem, const chunk_t * cp);	/* default options */
 
 /* Print the contents of all chunks managed by an allocator. */
 /* Relevant options: all. */
-void debug_dump_memory(P2(const gs_ref_memory_t * mem,
-			  const dump_control_t * control));
+void debug_dump_memory(const gs_ref_memory_t *mem,
+		       const dump_control_t *control);
 
 /* Find all the objects that contain a given pointer. */
-void debug_find_pointers(P2(const gs_ref_memory_t *mem, const void *target));
+void debug_find_pointers(const gs_ref_memory_t *mem, const void *target);
 
 #endif /* DEBUG */
 
