@@ -1,23 +1,20 @@
 /* Copyright (C) 1989, 1995, 1996, 1997, 1998, 1999 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: gxchar.c,v $ $Revision: 1.2.2.4 $ */
+/* $Id: gxchar.c,v 1.44 2005/01/10 11:32:06 igor Exp $ */
 /* Default implementation of text writing */
 #include "gx.h"
 #include "memory_.h"
@@ -37,17 +34,15 @@
 #include "gxfcache.h"
 #include "gspath.h"
 #include "gzpath.h"
+#include "gxfcid.h"
 
 /* Define whether or not to cache characters rotated by angles other than */
 /* multiples of 90 degrees. */
-private bool CACHE_ROTATED_CHARS = true;
-
-/* Define whether or not to oversample characters at small sizes. */
-private bool OVERSAMPLE = true;
+private const bool CACHE_ROTATED_CHARS = true;
 
 /* Define the maximum size of a full temporary bitmap when rasterizing, */
 /* in bits (not bytes). */
-private uint MAX_TEMP_BITMAP_BITS = 80000;
+private const uint MAX_TEMP_BITMAP_BITS = 80000;
 
 /* Define whether the show operation uses the character outline data, */
 /* as opposed to just needing the width (or nothing). */
@@ -75,13 +70,13 @@ private RELOC_PTRS_WITH(show_enum_reloc_ptrs, gs_show_enum *eptr)
 RELOC_PTRS_END
 
 /* Forward declarations */
-private int continue_kshow(P1(gs_show_enum *));
-private int continue_show(P1(gs_show_enum *));
-private int continue_show_update(P1(gs_show_enum *));
-private void show_set_scale(P1(gs_show_enum *));
-private int show_cache_setup(P1(gs_show_enum *));
-private int show_state_setup(P1(gs_show_enum *));
-private int show_origin_setup(P4(gs_state *, fixed, fixed, gs_char_path_mode));
+private int continue_kshow(gs_show_enum *);
+private int continue_show(gs_show_enum *);
+private int continue_show_update(gs_show_enum *);
+private void show_set_scale(const gs_show_enum *, gs_log2_scale_point *log2_scale);
+private int show_cache_setup(gs_show_enum *);
+private int show_state_setup(gs_show_enum *);
+private int show_origin_setup(gs_state *, fixed, fixed, gs_show_enum * penum);
 
 /* Accessors for current_char and current_glyph. */
 #define CURRENT_CHAR(penum) ((penum)->returned.current_char)
@@ -108,6 +103,8 @@ gs_show_enum_alloc(gs_memory_t * mem, gs_state * pgs, client_name_t cname)
     penum->show_gstate = 0;
     penum->dev_cache = 0;
     penum->dev_cache2 = 0;
+    penum->fapi_log2_scale.x = penum->fapi_log2_scale.y = -1;
+    penum->fapi_glyph_shift.x = penum->fapi_glyph_shift.y = 0;
     penum->dev_null = 0;
     penum->fstack.depth = -1;
     return penum;
@@ -219,13 +216,42 @@ gx_default_text_begin(gx_device * dev, gs_imager_state * pis,
     return 0;
 }
 
+/* An auxiliary functions for pdfwrite to process type 3 fonts. */
+int
+gx_hld_stringwidth_begin(gs_imager_state * pis, gx_path **path)
+{
+    gs_state *pgs = (gs_state *)pis;
+    extern_st(st_gs_state);
+    int code;
+    
+    if (gs_object_type(pis->memory, pis) != &st_gs_state)
+	return_error(gs_error_unregistered);
+    code = gs_gsave(pgs);
+    if (code < 0)
+	return code;
+    gs_newpath(pgs);
+    *path = pgs->path;
+    gx_translate_to_fixed(pgs, fixed_0, fixed_0);
+    return gx_path_add_point(pgs->path, fixed_0, fixed_0);
+}
+
+int
+gx_default_text_restore_state(gs_text_enum_t *pte)
+{
+    gs_show_enum *penum;
+    gs_state *pgs;
+
+    if (SHOW_IS(pte, TEXT_DO_NONE))
+	return 0;
+    penum = (gs_show_enum *)pte;
+    pgs = penum->pgs;
+    return gs_grestore(pgs);
+}
 /* ------ Width/cache setting ------ */
 
 private int
-    set_char_width(P4(gs_show_enum *penum, gs_state *pgs,
-		      floatp wx, floatp wy)),
-    set_cache_device(P6(gs_show_enum *penum, gs_state *pgs,
-			floatp llx, floatp lly, floatp urx, floatp ury));
+    set_cache_device(gs_show_enum *penum, gs_state *pgs,
+		     floatp llx, floatp lly, floatp urx, floatp ury);
 
 /* This is the default implementation of text enumerator set_cache. */
 private int
@@ -243,16 +269,17 @@ gx_show_text_set_cache(gs_text_enum_t *pte, const double *pw,
 
 	if (code < 0)
 	    return code;
+	if (SHOW_IS_ALL_OF(penum, TEXT_DO_NONE | TEXT_INTERVENE)) /* cshow */
+            return code;
 	return set_cache_device(penum, pgs, pw[2], pw[3], pw[4], pw[5]);
     }
     case TEXT_SET_CACHE_DEVICE2: {
 	int code;
-	bool retry = penum->width_status == sws_retry? true: false;
+	bool retry = (penum->width_status == sws_retry);
 
 	if (gs_rootfont(pgs)->WMode) {
 	    float vx = pw[8], vy = pw[9];
 	    gs_fixed_point pvxy, dvxy;
-	    cached_char *cc;
 
 	    gs_fixed_point rewind_pvxy;
 	    int rewind_code;
@@ -263,6 +290,8 @@ gx_show_text_set_cache(gs_text_enum_t *pte, const double *pw,
 		return 0;		/* don't cache */
 	    if ((code = set_char_width(penum, pgs, pw[6], pw[7])) < 0)
 		return code;
+	    if (SHOW_IS_ALL_OF(penum, TEXT_DO_NONE | TEXT_INTERVENE))
+		return code;
 	    /* Adjust the origin by (vx, vy). */
 	    gx_translate_to_fixed(pgs, pvxy.x, pvxy.y);
 	    code = set_cache_device(penum, pgs, pw[2], pw[3], pw[4], pw[5]);
@@ -270,8 +299,8 @@ gx_show_text_set_cache(gs_text_enum_t *pte, const double *pw,
 	        if (retry) {
 		   rewind_code = gs_point_transform2fixed(&pgs->ctm, vx, vy, &rewind_pvxy);
 		   if (rewind_code < 0) {
-		       /* If the control reaches here, something wrong. */
-		       return 0;  
+		       /* If the control passes here, something is wrong. */
+		       return_error(gs_error_unregistered);
 		   }
 		   /* Rewind the origin by (-vx, -vy) if the cache is failed. */
 		   gx_translate_to_fixed(pgs, rewind_pvxy.x, rewind_pvxy.y);
@@ -279,13 +308,14 @@ gx_show_text_set_cache(gs_text_enum_t *pte, const double *pw,
 		return code;
 	    }
 	    /* Adjust the character origin too. */
-	    cc = penum->cc;
-	    cc->offset.x += dvxy.x;
-	    cc->offset.y += dvxy.y;
+	    (penum->cc)->offset.x += dvxy.x;
+	    (penum->cc)->offset.y += dvxy.y;
 	} else {
 	    code = set_char_width(penum, pgs, pw[0], pw[1]);
 	    if (code < 0)
 		return code;
+	    if (SHOW_IS_ALL_OF(penum, TEXT_DO_NONE | TEXT_INTERVENE))
+		return code;
 	    code = set_cache_device(penum, pgs, pw[2], pw[3], pw[4], pw[5]);
 	}
 	return code;
@@ -298,13 +328,32 @@ gx_show_text_set_cache(gs_text_enum_t *pte, const double *pw,
 /* Set the character width. */
 /* Note that this returns 1 if the current show operation is */
 /* non-displaying (stringwidth or cshow). */
-private int
+int
 set_char_width(gs_show_enum *penum, gs_state *pgs, floatp wx, floatp wy)
 {
     int code;
 
     if (penum->width_status != sws_none && penum->width_status != sws_retry)
 	return_error(gs_error_undefined);
+    if (penum->fstack.depth > 0 && 
+	penum->fstack.items[penum->fstack.depth].font->FontType == ft_CID_encrypted) {
+    	/* We must not convert advance width with a CID font leaf's FontMatrix,
+	   because CDevProc is attached to the CID font rather than to its leaf.
+	   But show_state_setup sets CTM with the leaf's matrix.
+	   Compensate it here with inverse FontMatrix of the leaf.
+	   ( We would like to do without an inverse transform, but 
+	     we don't like to extend general gs_state or gs_show_enum
+	     for this particular reason. ) */
+	const gx_font_stack_item_t *pfsi = &penum->fstack.items[penum->fstack.depth];
+	gs_point p;
+
+	code = gs_distance_transform_inverse(wx, wy,
+		&gs_cid0_indexed_font(pfsi->font, pfsi->index)->FontMatrix, &p);
+	if (code < 0)
+	    return code;
+	wx = p.x; 
+	wy = p.y;
+    }
     if ((code = gs_distance_transform2fixed(&pgs->ctm, wx, wy, &penum->wxy)) < 0)
 	return code;
     /* Check whether we're setting the scalable width */
@@ -315,9 +364,99 @@ set_char_width(gs_show_enum *penum, gs_state *pgs, floatp wx, floatp wy)
     } else {
 	penum->width_status = sws_no_cache;
     }
+    if (SHOW_IS_ALL_OF(penum, TEXT_DO_NONE | TEXT_INTERVENE)) /* cshow */
+	gs_nulldevice(pgs);
     return !SHOW_IS_DRAWING(penum);
 }
 
+void
+gx_compute_text_oversampling(const gs_show_enum * penum, const gs_font *pfont, 
+                             int alpha_bits, gs_log2_scale_point *p_log2_scale)
+{
+    gs_log2_scale_point log2_scale;
+
+    if (alpha_bits == 1) 
+	log2_scale.x = log2_scale.y = 0;
+    else if (pfont->PaintType != 0) {
+	/* Don't oversample artificially stroked fonts. */
+	log2_scale.x = log2_scale.y = 0;
+    } else if (!penum->is_pure_color) {
+	/* Don't oversample characters for rendering in non-pure color. */
+	log2_scale.x = log2_scale.y = 0;
+    } else {
+	int excess;
+
+	/* Get maximal scale according to cached bitmap size. */
+	show_set_scale(penum, &log2_scale);
+	/* Reduce the scale to fit into alpha bits. */
+	excess = log2_scale.x + log2_scale.y - alpha_bits;
+	while (excess > 0) {
+	    if (log2_scale.y > 0) {
+		log2_scale.y --; 
+		excess--;
+		if (excess == 0)
+		    break;
+	    }
+	    if (log2_scale.x > 0) {
+		log2_scale.x --; 
+		excess--;
+	    }
+	}
+    }
+    *p_log2_scale = log2_scale;
+}
+
+/* Compute glyph raster parameters */
+private int
+compute_glyph_raster_params(gs_show_enum *penum, bool in_setcachedevice, int *alpha_bits, 
+		    int *depth,
+                    gs_fixed_point *subpix_origin, gs_log2_scale_point *log2_scale)
+{
+    gs_state *pgs = penum->pgs;
+    gx_device *dev = gs_currentdevice_inline(pgs);
+    int code;
+    
+    *alpha_bits = (*dev_proc(dev, get_alpha_bits)) (dev, go_text);
+    if (in_setcachedevice) {
+	/* current point should already be in penum->origin */
+    } else {
+	code = gx_path_current_point_inline(pgs->path, &penum->origin);
+	if (code < 0) {
+	    /* For cshow, having no current point is acceptable. */
+	    if (!SHOW_IS(penum, TEXT_DO_NONE))
+		return code;
+	    penum->origin.x = penum->origin.y = 0;	/* arbitrary */
+	}
+    }
+    if (penum->fapi_log2_scale.x != -1)
+	*log2_scale = penum->fapi_log2_scale;
+    else
+	gx_compute_text_oversampling(penum, penum->current_font, *alpha_bits, log2_scale);
+    /*	We never oversample over the device alpha_bits,
+     * so that we don't need to scale down. Perhaps it may happen 
+     * that we underuse alpha_bits due to a big character raster,
+     * so we must compute log2_depth more accurately :
+     */
+    *depth = (log2_scale->x + log2_scale->y == 0 ?
+        1 : min(log2_scale->x + log2_scale->y, *alpha_bits));
+    if (gs_currentaligntopixels(penum->current_font->dir) == 0) {
+	int scx = -1L << (_fixed_shift - log2_scale->x);
+	int rdx =  1L << (_fixed_shift - 1 - log2_scale->x);
+
+#	if 1 /* Ever align Y to pixels to provide an uniform glyph height. */
+	    subpix_origin->y = 0;
+#	else
+	    int scy = -1L << (_fixed_shift - log2_scale->y);
+	    int rdy =  1L << (_fixed_shift - 1 - log2_scale->y);
+
+	    subpix_origin->y = ((penum->origin.y + rdy) & scy) & (fixed_1 - 1);
+#	endif
+	subpix_origin->x = ((penum->origin.x + rdx) & scx) & (fixed_1 - 1);
+    } else
+	subpix_origin->x = subpix_origin->y = 0;
+    return 0;
+}
+
 /* Set up the cache device if relevant. */
 /* Return 1 if we just set up a cache device. */
 /* Used by setcachedevice and setcachedevice2. */
@@ -352,17 +491,16 @@ set_cache_device(gs_show_enum * penum, gs_state * pgs, floatp llx, floatp lly,
     } {
 	const gs_font *pfont = pgs->font;
 	gs_font_dir *dir = pfont->dir;
-	gx_device *dev = gs_currentdevice_inline(pgs);
-	int alpha_bits =
-	(*dev_proc(dev, get_alpha_bits)) (dev, go_text);
+        int alpha_bits, depth;
 	gs_log2_scale_point log2_scale;
-	static const fixed max_cdim[3] =
-	{
+	gs_fixed_point subpix_origin;
+        static const fixed max_cdim[3] =
+        {
 #define max_cd(n)\
 	    (fixed_1 << (arch_sizeof_short * 8 - n)) - (fixed_1 >> n) * 3
 	    max_cd(0), max_cd(1), max_cd(2)
 #undef max_cd
-	};
+        };
 	ushort iwidth, iheight;
 	cached_char *cc;
 	gs_fixed_rect clip_box;
@@ -400,13 +538,12 @@ set_cache_device(gs_show_enum * penum, gs_state * pgs, floatp llx, floatp lly,
 	if (clr.y < cll.y)
 	    cll.y = clr.y, cur.y = cul.y;
 	/* Now cll and cur are the extrema of the box. */
-	cdim.x = cur.x - cll.x;
-	cdim.y = cur.y - cll.y;
-	show_set_scale(penum);
-	log2_scale.x = penum->log2_suggested_scale.x;
-	log2_scale.y = penum->log2_suggested_scale.y;
+	code = compute_glyph_raster_params(penum, true, &alpha_bits, &depth,
+           &subpix_origin, &log2_scale);
+	if (code < 0)
+	    return code;
 #ifdef DEBUG
-	if (gs_debug_c('k')) {
+        if (gs_debug_c('k')) {
 	    dlprintf6("[k]cbox=[%g %g %g %g] scale=%dx%d\n",
 		      fixed2float(cll.x), fixed2float(cll.y),
 		      fixed2float(cur.x), fixed2float(cur.y),
@@ -414,31 +551,10 @@ set_cache_device(gs_show_enum * penum, gs_state * pgs, floatp llx, floatp lly,
 	    dlprintf6("[p]  ctm=[%g %g %g %g %g %g]\n",
 		      pgs->ctm.xx, pgs->ctm.xy, pgs->ctm.yx, pgs->ctm.yy,
 		      pgs->ctm.tx, pgs->ctm.ty);
-	}
+        }
 #endif
-	/*
-	 * If the device wants anti-aliased text,
-	 * increase the sampling scale to ensure that
-	 * if we want N bits of alpha, we generate
-	 * at least 2^N sampled bits per pixel.
-	 */
-	if (alpha_bits > 1) {
-	    int more_bits =
-	    alpha_bits - (log2_scale.x + log2_scale.y);
-
-	    if (more_bits > 0) {
-		if (log2_scale.x <= log2_scale.y) {
-		    log2_scale.x += (more_bits + 1) >> 1;
-		    log2_scale.y += more_bits >> 1;
-		} else {
-		    log2_scale.x += more_bits >> 1;
-		    log2_scale.y += (more_bits + 1) >> 1;
-		}
-	    }
-	} else if (!OVERSAMPLE || pfont->PaintType != 0) {
-	    /* Don't oversample artificially stroked fonts. */
-	    log2_scale.x = log2_scale.y = 0;
-	}
+	cdim.x = cur.x - cll.x;
+	cdim.y = cur.y - cll.y;
 	if (cdim.x > max_cdim[log2_scale.x] ||
 	    cdim.y > max_cdim[log2_scale.y]
 	    )
@@ -464,7 +580,7 @@ set_cache_device(gs_show_enum * penum, gs_state * pgs, floatp llx, floatp lly,
 				(iwidth > MAX_TEMP_BITMAP_BITS / iheight &&
 				 log2_scale.x + log2_scale.y > alpha_bits ?
 				 penum->dev_cache2 : NULL),
-				iwidth, iheight, &log2_scale, alpha_bits);
+				iwidth, iheight, &log2_scale, depth);
 	if (cc == 0)
 	    return 0;		/* too big for cache */
 	/* The mins handle transposed coordinate systems.... */
@@ -485,14 +601,19 @@ set_cache_device(gs_show_enum * penum, gs_state * pgs, floatp llx, floatp lly,
 	cc->code = glyph;
 	cc->wmode = gs_rootfont(pgs)->WMode;
 	cc->wxy = penum->wxy;
+	cc->subpix_origin = subpix_origin;
+	if (penum->pair != 0)
+	    cc_set_pair(cc, penum->pair);
+	else
+	    cc->pair = 0;
 	/* Install the device */
 	gx_set_device_only(pgs, (gx_device *) penum->dev_cache);
 	pgs->ctm_default_set = false;
 	/* Adjust the transformation in the graphics context */
 	/* so that the character lines up with the cache. */
 	gx_translate_to_fixed(pgs,
-			      cc->offset.x << log2_scale.x,
-			      cc->offset.y << log2_scale.y);
+			      (cc->offset.x + subpix_origin.x) << log2_scale.x,
+			      (cc->offset.y + subpix_origin.y) << log2_scale.y);
 	if ((log2_scale.x | log2_scale.y) != 0)
 	    gx_scale_char_matrix(pgs, 1 << log2_scale.x,
 				 1 << log2_scale.y);
@@ -565,10 +686,10 @@ gx_show_text_process(gs_text_enum_t *pte)
 }
 
 /* Continuation procedures */
-private int show_update(P1(gs_show_enum * penum));
-private int show_move(P1(gs_show_enum * penum));
-private int show_proceed(P1(gs_show_enum * penum));
-private int show_finish(P1(gs_show_enum * penum));
+private int show_update(gs_show_enum * penum);
+private int show_move(gs_show_enum * penum);
+private int show_proceed(gs_show_enum * penum);
+private int show_finish(gs_show_enum * penum);
 private int
 continue_show_update(gs_show_enum * penum)
 {
@@ -634,9 +755,15 @@ show_update(gs_show_enum * penum)
 		case 1:
 		    ;
 	    }
-	    gx_add_cached_char(pgs->font->dir, penum->dev_cache,
-			       cc, gx_lookup_fm_pair(pgs->font, pgs),
-			       &penum->log2_scale);
+	    {   cached_fm_pair *pair;
+
+		code = gx_lookup_fm_pair(pgs->font, &char_tm_only(pgs), 
+			    &penum->log2_scale, penum->charpath_flag != cpm_show, &pair);
+		if (code < 0)
+		    return code;
+		gx_add_cached_char(pgs->font->dir, penum->dev_cache,
+			       cc, pair, &penum->log2_scale);
+	    }
 	    if (!SHOW_USES_OUTLINE(penum) ||
 		penum->charpath_flag != cpm_show
 		)
@@ -669,13 +796,39 @@ show_update(gs_show_enum * penum)
 private int
 show_fast_move(gs_state * pgs, gs_fixed_point * pwxy)
 {
-    int code = gx_path_add_rel_point_inline(pgs->path, pwxy->x, pwxy->y);
-
-    /* If the current position is out of range, don't try to move. */
-    if (code == gs_error_limitcheck && pgs->clamp_coordinates)
-	code = 0;
-    return code;
+    return gs_moveto_aux((gs_imager_state *)pgs, pgs->path,
+			      pgs->current_point.x + fixed2float(pwxy->x), 
+			      pgs->current_point.y + fixed2float(pwxy->y));
 }
+
+/* Get the current character code. */
+int gx_current_char(const gs_text_enum_t * pte)
+{
+    const gs_show_enum *penum = (const gs_show_enum *)pte;
+    gs_char chr = CURRENT_CHAR(penum) & 0xff;
+    int fdepth = penum->fstack.depth;
+
+    if (fdepth > 0) {
+	/* Add in the shifted font number. */
+	uint fidx = penum->fstack.items[fdepth].index;
+
+	switch (((gs_font_type0 *) (penum->fstack.items[fdepth - 1].font))->data.FMapType) {
+	case fmap_1_7:
+	case fmap_9_7:
+	    chr += fidx << 7;
+	    break;
+	case fmap_CMap:
+	    chr = CURRENT_CHAR(penum);  /* the full character */
+	    if (!penum->cmap_code)
+		break;
+	    /* falls through */
+	default:
+	    chr += fidx << 8;
+	}
+    }
+    return chr;
+}
+
 private int
 show_move(gs_show_enum * penum)
 {
@@ -690,27 +843,8 @@ show_move(gs_show_enum * penum)
 	double dx = 0, dy = 0;
 
 	if (SHOW_IS_ADD_TO_SPACE(penum)) {
-	    gs_char chr = CURRENT_CHAR(penum) & 0xff;
-	    int fdepth = penum->fstack.depth;
+	    gs_char chr = gx_current_char((const gs_text_enum_t *)penum);
 
-	    if (fdepth > 0) {
-		/* Add in the shifted font number. */
-		uint fidx = penum->fstack.items[fdepth].index;
-
-		switch (((gs_font_type0 *) (penum->fstack.items[fdepth - 1].font))->data.FMapType) {
-		case fmap_1_7:
-		case fmap_9_7:
-		    chr += fidx << 7;
-		    break;
-		case fmap_CMap:
-		    chr = CURRENT_CHAR(penum);  /* the full character */
-		    if (!penum->cmap_code)
-			break;
-		    /* falls through */
-		default:
-		    chr += fidx << 8;
-		}
-	    }
 	    if (chr == penum->text.space.s_char) {
 		dx = penum->text.delta_space.x;
 		dy = penum->text.delta_space.y;
@@ -765,8 +899,7 @@ show_proceed(gs_show_enum * penum)
     gs_glyph glyph;
     int code;
     cached_char *cc;
-    gx_device *dev = gs_currentdevice_inline(pgs);
-    int alpha_bits = (*dev_proc(dev, get_alpha_bits)) (dev, go_text);
+    gs_log2_scale_point log2_scale;
 
     if (penum->charpath_flag == cpm_show && SHOW_USES_OUTLINE(penum)) {
 	code = gs_state_color_load(pgs);
@@ -795,6 +928,7 @@ show_proceed(gs_show_enum * penum)
 		    pgs->char_tm_valid = false;
 		    show_state_setup(penum);
 		    pair = 0;
+		    penum->pair = 0;
 		    /* falls through */
 		case 0:	/* plain char */
 		    /*
@@ -804,29 +938,59 @@ show_proceed(gs_show_enum * penum)
 		     * do it here.
 		     */
 		    SET_CURRENT_CHAR(penum, chr);
+		    /*
+		     * Store glyph now, because pdfwrite needs it while
+		     * synthezising bitmap fonts (see assign_char_code).
+		     */
 		    if (glyph == gs_no_glyph) {
 			glyph = (*penum->encode_char)(pfont, chr,
 						      GLYPH_SPACE_NAME);
+			SET_CURRENT_GLYPH(penum, glyph);
 			if (glyph == gs_no_glyph) {
 			    cc = 0;
 			    goto no_cache;
 			}
+		    } else
+    			SET_CURRENT_GLYPH(penum, glyph);
+		    penum->is_pure_color = gs_color_writes_pure(penum->pgs); /* Save
+		                 this data for compute_glyph_raster_params to work 
+				 independently on the color change in BuildChar. 
+				 Doing it here because cshow proc may modify 
+				 the graphic state.
+				 */
+		    {
+			int alpha_bits, depth;
+			gs_fixed_point subpix_origin;
+
+			code = compute_glyph_raster_params(penum, false, 
+				    &alpha_bits, &depth, &subpix_origin, &log2_scale);
+			if (code < 0)
+			    return code;
+			if (pair == 0) {
+			    code = gx_lookup_fm_pair(pfont, &char_tm_only(pgs), &log2_scale, 
+				penum->charpath_flag != cpm_show, &pair);
+			    if (code < 0)
+				return code;
+			}
+			penum->pair = pair;
+			cc = gx_lookup_cached_char(pfont, pair, glyph, wmode,
+						   depth, &subpix_origin);
 		    }
-		    if (pair == 0)
-			pair = gx_lookup_fm_pair(pfont, pgs);
-		    cc = gx_lookup_cached_char(pfont, pair, glyph, wmode,
-					       alpha_bits);
 		    if (cc == 0) {
 			/* Character is not in cache. */
 			/* If possible, try for an xfont before */
 			/* rendering from the outline. */
+
+		        /* If antialiasing is in effect, don't use xfont */
+		        if (log2_scale.x + log2_scale.y > 0)
+			    goto no_cache;
 			if (pfont->ExactSize == fbit_use_outlines ||
 			    pfont->PaintType == 2
 			    )
 			    goto no_cache;
 			if (pfont->BitmapWidths) {
 			    cc = gx_lookup_xfont_char(pgs, pair, chr,
-				     glyph, &pfont->procs.callbacks, wmode);
+				     glyph, wmode);
 			    if (cc == 0)
 				goto no_cache;
 			} else {
@@ -838,7 +1002,7 @@ show_proceed(gs_show_enum * penum)
 			    /* We might have an xfont, but we still */
 			    /* want the scalable widths. */
 			    cc = gx_lookup_xfont_char(pgs, pair, chr,
-				     glyph, &pfont->procs.callbacks, wmode);
+				     glyph, wmode);
 			    /* Render up to the point of */
 			    /* setcharwidth or setcachedevice, */
 			    /* just as for stringwidth. */
@@ -901,8 +1065,7 @@ show_proceed(gs_show_enum * penum)
 		    } else
 			code = show_fast_move(pgs, &cc->wxy);
 		    if (code) {
-			/* Might be kshow, so store the state. */
-			SET_CURRENT_GLYPH(penum, glyph);
+			/* Might be kshow, glyph is stored above. */
 			return code;
 		    }
 	    }
@@ -920,13 +1083,29 @@ show_proceed(gs_show_enum * penum)
 		pfont = penum->fstack.items[penum->fstack.depth].font;
 		penum->current_font = pfont;
 		show_state_setup(penum);
+		pair = 0;
 	    case 0:
-		;
+		{   int alpha_bits, depth;
+		    gs_log2_scale_point log2_scale;
+		    gs_fixed_point subpix_origin;
+
+		    code = compute_glyph_raster_params(penum, false, &alpha_bits, &depth, &subpix_origin, &log2_scale);
+		    if (code < 0)
+			return code;
+		    if (pair == 0) {
+			code = gx_lookup_fm_pair(pfont, &char_tm_only(pgs), &log2_scale,
+				penum->charpath_flag != cpm_show, &pair);
+			if (code < 0)
+			    return code;
+		    }
+		    penum->pair = pair;
+		}
 	}
 	SET_CURRENT_CHAR(penum, chr);
 	if (glyph == gs_no_glyph) {
 	    glyph = (*penum->encode_char)(pfont, chr, GLYPH_SPACE_NAME);
 	}
+        SET_CURRENT_GLYPH(penum, glyph);
 	cc = 0;
     }
   no_cache:
@@ -935,10 +1114,9 @@ show_proceed(gs_show_enum * penum)
      * we only do this if the character is not cached (cc = 0);
      * however, we also must do this if we have an xfont but
      * are using scalable widths.  In this case, and only this case,
-     * we get here with cc != 0.  penum->current_char has already
-     * been set, but not penum->current_glyph.
+     * we get here with cc != 0.  penum->current_char and penum->current_glyph
+     * has already been set.
      */
-    SET_CURRENT_GLYPH(penum, glyph);
     if ((code = gs_gsave(pgs)) < 0)
 	return code;
     /* Set the font to the current descendant font. */
@@ -946,8 +1124,6 @@ show_proceed(gs_show_enum * penum)
     /* Reset the in_cachedevice flag, so that a recursive show */
     /* will use the cache properly. */
     pgs->in_cachedevice = CACHE_DEVICE_NONE;
-    /* Reset the sampling scale. */
-    penum->log2_scale.x = penum->log2_scale.y = 0;
     /* Set the charpath data in the graphics context if necessary, */
     /* so that fill and stroke will add to the path */
     /* rather than having their usual effect. */
@@ -997,13 +1173,14 @@ show_proceed(gs_show_enum * penum)
 	    cpt.y = float2fixed(fpy);
 	}
 	gs_newpath(pgs);
-	code = show_origin_setup(pgs, cpt.x, cpt.y,
-				 penum->charpath_flag);
+	code = show_origin_setup(pgs, cpt.x, cpt.y, penum);
 	if (code < 0)
 	    goto rret;
     }
     penum->width_status = sws_none;
     penum->continue_proc = continue_show_update;
+    /* Reset the sampling scale. */
+    penum->log2_scale.x = penum->log2_scale.y = 0;
     /* Try using the build procedure in the font. */
     /* < 0 means error, 0 means success, 1 means failure. */
     penum->cc = cc;		/* set this now for build procedure */
@@ -1060,6 +1237,7 @@ gx_show_text_retry(gs_text_enum_t *pte)
     gs_grestore(penum->pgs);
     penum->width_status = sws_retry;
     penum->log2_scale.x = penum->log2_scale.y = 0;
+    penum->pair = 0;
     return 0;
 }
 
@@ -1150,6 +1328,22 @@ gs_show_current_font(const gs_show_enum * penum)
 
 /* ------ Internal routines ------ */
 
+private inline bool
+is_matrix_good_for_caching(const gs_matrix_fixed *m)
+{
+    /* Skewing or non-rectangular rotation are not supported,
+       but we ignore a small noise skew. */
+    const float axx = any_abs(m->xx), axy = any_abs(m->xy);
+    const float ayx = any_abs(m->yx), ayy = any_abs(m->yy);
+    const float thr = 5000; /* examples/alphabet.ps */
+
+    if (ayx * thr < axx || axy * thr < ayy)
+	return true;
+    if (axx * thr < ayx || ayy * thr < axy)
+	return true;
+    return false;
+}
+
 /* Initialize the gstate-derived parts of a show enumerator. */
 /* We do this both when starting the show operation, */
 /* and when returning from the kshow callout. */
@@ -1173,14 +1367,16 @@ show_state_setup(gs_show_enum * penum)
 	pfont = pfsi->font;
 	gs_matrix_multiply(&pfont->FontMatrix,
 			   &pfsi[-1].font->FontMatrix, &mat);
+	if (pfont->FontType == ft_CID_encrypted) {
+	    /* concatenate the Type9 leaf's matrix */
+	    gs_matrix_multiply(&mat,
+		&(gs_cid0_indexed_font(pfont, pfsi->index)->FontMatrix), &mat);
+	}
 	gs_setcharmatrix(pgs, &mat);
     }
     penum->current_font = pfont;
     /* Skewing or non-rectangular rotation are not supported. */
-    if (!CACHE_ROTATED_CHARS &&
-	(is_fzero2(pgs->char_tm.xy, pgs->char_tm.yx) ||
-	 is_fzero2(pgs->char_tm.xx, pgs->char_tm.yy))
-	)
+    if (!CACHE_ROTATED_CHARS && is_matrix_good_for_caching(&pgs->char_tm))
 	penum->can_cache = 0;
     if (penum->can_cache >= 0 &&
 	gx_effective_clip_path(pgs, &pcpath) >= 0
@@ -1227,7 +1423,7 @@ show_state_setup(gs_show_enum * penum)
 
 /* Set the suggested oversampling scale for character rendering. */
 private void
-show_set_scale(gs_show_enum * penum)
+show_set_scale(const gs_show_enum * penum, gs_log2_scale_point *log2_scale)
 {
     /*
      * Decide whether to oversample.
@@ -1238,12 +1434,11 @@ show_set_scale(gs_show_enum * penum)
     if ((penum->charpath_flag == cpm_show ||
 	 penum->charpath_flag == cpm_charwidth) &&
 	SHOW_USES_OUTLINE(penum) &&
-	gx_path_is_void_inline(pgs->path) &&
+	/* gx_path_is_void_inline(pgs->path) && */
     /* Oversampling rotated characters doesn't work well. */
-	(is_fzero2(pgs->char_tm.xy, pgs->char_tm.yx) ||
-	 is_fzero2(pgs->char_tm.xx, pgs->char_tm.yy))
+	is_matrix_good_for_caching(&pgs->char_tm)
 	) {
-	const gs_font_base *pfont = (gs_font_base *) pgs->font;
+	const gs_font_base *pfont = (const gs_font_base *)penum->current_font;
 	gs_fixed_point extent;
 	int code = gs_distance_transform2fixed(&pgs->char_tm,
 				  pfont->FontBBox.q.x - pfont->FontBBox.p.x,
@@ -1252,14 +1447,12 @@ show_set_scale(gs_show_enum * penum)
 
 	if (code >= 0) {
 	    int sx =
-	    (extent.x == 0 ? 0 :
-	     any_abs(extent.x) < int2fixed(25) ? 2 :
-	     any_abs(extent.x) < int2fixed(60) ? 1 :
+	    (any_abs(extent.x) < int2fixed(60) ? 2 :
+	     any_abs(extent.x) < int2fixed(200) ? 1 :
 	     0);
 	    int sy =
-	    (extent.y == 0 ? 0 :
-	     any_abs(extent.y) < int2fixed(25) ? 2 :
-	     any_abs(extent.y) < int2fixed(60) ? 1 :
+	    (any_abs(extent.y) < int2fixed(60) ? 2 :
+	     any_abs(extent.y) < int2fixed(200) ? 1 :
 	     0);
 
 	    /* If we oversample at all, make sure we do it */
@@ -1268,14 +1461,13 @@ show_set_scale(gs_show_enum * penum)
 		sx = 1;
 	    else if (sy == 0 && sx != 0)
 		sy = 1;
-	    penum->log2_suggested_scale.x = sx;
-	    penum->log2_suggested_scale.y = sy;
+	    log2_scale->x = sx;
+	    log2_scale->y = sy;
 	    return;
 	}
     }
     /* By default, don't scale. */
-    penum->log2_suggested_scale.x =
-	penum->log2_suggested_scale.y = 0;
+    log2_scale->x = log2_scale->y = 0;
 }
 
 /* Set up the cache device and related information. */
@@ -1318,14 +1510,23 @@ show_cache_setup(gs_show_enum * penum)
 /* Used before rendering characters, and for moving the origin */
 /* in setcachedevice2 when WMode=1. */
 private int
-show_origin_setup(gs_state * pgs, fixed cpt_x, fixed cpt_y,
-		  gs_char_path_mode charpath_flag)
+show_origin_setup(gs_state * pgs, fixed cpt_x, fixed cpt_y, gs_show_enum * penum)
 {
-    if (charpath_flag == cpm_show) {
+    if (penum->charpath_flag == cpm_show) {
 	/* Round the translation in the graphics state. */
 	/* This helps prevent rounding artifacts later. */
-	cpt_x = fixed_rounded(cpt_x);
-	cpt_y = fixed_rounded(cpt_y);
+	if (gs_currentaligntopixels(penum->current_font->dir) == 0) {
+	    int scx = -1L << (_fixed_shift - penum->log2_scale.x);
+	    int scy = -1L << (_fixed_shift - penum->log2_scale.y);
+	    int rdx =  1L << (_fixed_shift - 1 - penum->log2_scale.x);
+	    int rdy =  1L << (_fixed_shift - 1 - penum->log2_scale.y);
+
+	    cpt_x = (cpt_x + rdx) & scx;
+	    cpt_y = (cpt_y + rdy) & scy;
+	} else {
+	    cpt_x = fixed_rounded(cpt_x);
+	    cpt_y = fixed_rounded(cpt_y);
+	}
     }
     /*
      * BuildChar procedures expect the current point to be undefined,
