@@ -1,23 +1,20 @@
 /* Copyright (C) 1989, 1995, 1996, 1997, 1998, 1999 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: gxstroke.c,v $ $Revision: 1.2.2.1 $ */
+/* $Id: gxstroke.c,v 1.13 2004/10/18 15:23:20 igor Exp $ */
 /* Path stroking procedures for Ghostscript library */
 #include "math_.h"
 #include "gx.h"
@@ -36,6 +33,7 @@
 #include "gzpath.h"
 #include "gzcpath.h"
 #include "gxpaint.h"
+#include "vdtrace.h"
 
 /*
  * We don't really know whether it's a good idea to take fill adjustment
@@ -87,7 +85,7 @@
  * check and covers many common cases.  Clients that care always have the
  * option of using strokepath to get an exact result.
  */
-private float join_expansion_factor(P2(const gs_imager_state *, gs_line_join));
+private float join_expansion_factor(const gs_imager_state *, gs_line_join);
 int
 gx_stroke_path_expansion(const gs_imager_state * pis, const gx_path * ppath,
 			 gs_fixed_point * ppt)
@@ -211,18 +209,18 @@ typedef partial_line *pl_ptr;
   ((pp)->x = (p).x, (pp)->y = (p).y)
 
 /* Other forward declarations */
-private bool width_is_thin(P1(pl_ptr));
-private void adjust_stroke(P3(pl_ptr, const gs_imager_state *, bool));
-private int line_join_points(P6(const gx_line_params * pgs_lp,
-				pl_ptr plp, pl_ptr nplp,
-				gs_fixed_point * join_points,
-				const gs_matrix * pmat, gs_line_join join));
-private void compute_caps(P1(pl_ptr));
-private int add_points(P4(gx_path *, const gs_fixed_point *,
-			  int, bool));
-private int add_round_cap(P2(gx_path *, const_ep_ptr));
-private int cap_points(P3(gs_line_cap, const_ep_ptr,
-			  gs_fixed_point * /*[3] */ ));
+private bool width_is_thin(pl_ptr);
+private void adjust_stroke(pl_ptr, const gs_imager_state *, bool);
+private int line_join_points(const gx_line_params * pgs_lp,
+			     pl_ptr plp, pl_ptr nplp,
+			     gs_fixed_point * join_points,
+			     const gs_matrix * pmat, gs_line_join join);
+private void compute_caps(pl_ptr);
+private int add_points(gx_path *, const gs_fixed_point *,
+		       int, bool);
+private int add_round_cap(gx_path *, const_ep_ptr);
+private int cap_points(gs_line_cap, const_ep_ptr,
+		       gs_fixed_point * /*[3] */ );
 
 /* Define the default implementation of the device stroke_path procedure. */
 int
@@ -231,8 +229,22 @@ gx_default_stroke_path(gx_device * dev, const gs_imager_state * pis,
 		       const gx_drawing_color * pdcolor,
 		       const gx_clip_path * pcpath)
 {
-    return gx_stroke_path_only(ppath, (gx_path *) 0, dev, pis, params,
+    int code;
+
+    if (vd_allowed('S')) {
+	vd_get_dc('S');
+	if (vd_enabled) {
+	    vd_set_shift(0, 100);
+	    vd_set_scale(0.03);
+	    vd_set_origin(0, 0);
+	    vd_erase(RGB(192, 192, 192));
+	}
+    }
+    code = gx_stroke_path_only(ppath, (gx_path *) 0, dev, pis, params,
 			       pdcolor, pcpath);
+    if (vd_allowed('S'))
+	vd_release_dc;
+    return code;
 }
 
 /* Fill a partial stroked path.  Free variables: */
@@ -256,10 +268,10 @@ gx_default_stroke_path(gx_device * dev, const gs_imager_state * pis,
  * with an appropriate cap.
  */
 #define stroke_line_proc(proc)\
-  int proc(P11(gx_path *, int, pl_ptr, pl_ptr, const gx_device_color *,\
-	       gx_device *, const gs_imager_state *,\
-	       const gx_stroke_params *, const gs_fixed_rect *, int,\
-	       gs_line_join))
+  int proc(gx_path *, int, pl_ptr, pl_ptr, const gx_device_color *,\
+	   gx_device *, const gs_imager_state *,\
+	   const gx_stroke_params *, const gs_fixed_rect *, int,\
+	   gs_line_join)
 typedef stroke_line_proc((*stroke_line_proc_t));
 
 private stroke_line_proc(stroke_add);
@@ -341,7 +353,8 @@ gx_stroke_path_only(gx_path * ppath, gx_path * to_path, gx_device * pdev,
 	 pgs_lp->join == gs_join_none ? gs_join_bevel : pgs_lp->join);
     float line_width = pgs_lp->half_width;	/* (*half* the line width) */
     bool always_thin;
-    double line_width_and_scale, device_line_width_scale;
+    double line_width_and_scale;
+    double device_line_width_scale = 0; /* Quiet compiler. */
     double device_dot_length = pgs_lp->dot_length * fixed_1;
     const subpath *psub;
 
@@ -559,14 +572,18 @@ gx_stroke_path_only(gx_path * ppath, gx_path * to_path, gx_device * pdev,
 		}
 		/*
 		 * The entire subpath is degenerate, but it includes
-		 * more than one point.  If we are using round caps or
-		 * the dot length is non-zero, draw the caps, otherwise
-		 * do nothing.
+		 * more than one point.  If the dot length is non-zero,
+		 * draw the caps, otherwise do nothing.
 		 */
-		if (!(pgs_lp->cap == gs_cap_round ||
-		      pgs_lp->dot_length != 0)
-		    )
+		if (pgs_lp->dot_length != 0)
 		    break;
+		if (pgs_lp->cap != gs_cap_round) {
+		    /* From PLRM, stroke operator :
+		       If a subpath is degenerate (consists of a single-point closed path 
+		       or of two or more points at the same coordinates), 
+		       stroke paints it only if round line caps have been specified */
+		    break;
+		}
 		/*
 		 * Orient the dot according to the previous segment if
 		 * any, or else the next segment if any, or else
@@ -631,8 +648,8 @@ gx_stroke_path_only(gx_path * ppath, gx_path * to_path, gx_device * pdev,
 		if (uniform != 0) {
 		    /* We can save a lot of work in this case. */
 		    /* We know orient != orient_other. */
-		    float dpx = udx, dpy = udy;
-		    float wl = device_line_width_scale /
+		    double dpx = udx, dpy = udy;
+		    double wl = device_line_width_scale /
 		    hypot(dpx, dpy);
 
 		    pl.e.cdelta.x = (fixed) (dpx * wl);
@@ -851,11 +868,11 @@ line_intersect(
 		  p_ptr pi)
 {				/* return intersection here */
     /* We don't have to do any scaling, the factors all work out right. */
-    float u1 = pd1->x, v1 = pd1->y;
-    float u2 = pd2->x, v2 = pd2->y;
+    double u1 = pd1->x, v1 = pd1->y;
+    double u2 = pd2->x, v2 = pd2->y;
     double denom = u1 * v2 - u2 * v1;
-    float xdiff = pp2->x - pp1->x;
-    float ydiff = pp2->y - pp1->y;
+    double xdiff = pp2->x - pp1->x;
+    double ydiff = pp2->y - pp1->y;
     double f1;
     double max_result = any_abs(denom) * (double)max_fixed;
 
@@ -1066,6 +1083,7 @@ stroke_add(gx_path * ppath, int first, pl_ptr plp, pl_ptr nplp,
   done:
     if (code < 0)
 	return code;
+    vd_closepath;
     return gx_path_close_subpath(ppath);
 }
 
@@ -1074,14 +1092,20 @@ private int
 add_points(gx_path * ppath, const gs_fixed_point * points, int npoints,
 	   bool moveto_first)
 {
+    vd_setcolor(0);
+    vd_setlinewidth(0);
     if (moveto_first) {
 	int code = gx_path_add_point(ppath, points[0].x, points[0].y);
 
+	vd_moveto(points[0].x, points[0].y);
 	if (code < 0)
 	    return code;
+	vd_lineto_multi(points + 1, npoints - 1);
 	return gx_path_add_lines(ppath, points + 1, npoints - 1);
-    } else
+    } else {
+	vd_lineto_multi(points, npoints);
 	return gx_path_add_lines(ppath, points, npoints);
+    }
 }
 
 /* ---------------- Join computation ---------------- */
@@ -1198,13 +1222,18 @@ line_join_points(const gx_line_params * pgs_lp, pl_ptr plp, pl_ptr nplp,
 	double u1 = plp->e.cdelta.y, v1 = plp->e.cdelta.x;
 	double u2 = nplp->o.cdelta.y, v2 = nplp->o.cdelta.x;
 	double num, denom;
+	int code;
 
 	if (pmat) {
 	    gs_point pt;
 
-	    gs_distance_transform_inverse(v1, u1, pmat, &pt);
+	    code = gs_distance_transform_inverse(v1, u1, pmat, &pt);
+	    if (code < 0)
+		return code;
 	    v1 = pt.x, u1 = pt.y;
-	    gs_distance_transform_inverse(v2, u2, pmat, &pt);
+	    code = gs_distance_transform_inverse(v2, u2, pmat, &pt);
+	    if (code < 0)
+		return code;
 	    v2 = pt.x, u2 = pt.y;
 	    /*
 	     * We need to recompute ccw according to the
