@@ -1,4 +1,4 @@
-/* Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1995-2000, 2004 artofcode LLC.  All rights reserved.
   
   This software is provided AS-IS with no warranty, either express or
   implied.
@@ -14,21 +14,21 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: ilocate.c,v 1.4.2.1 2002/02/22 19:45:59 ray Exp $ */
+/* $Id: ilocate.c,v 1.12 2004/09/28 05:34:29 giles Exp $ */
 /* Object locating and validating for Ghostscript memory manager */
 #include "ghost.h"
 #include "memory_.h"
-#include "errors.h"
+#include "ierrors.h"
 #include "gsexit.h"
 #include "gsstruct.h"
 #include "iastate.h"
 #include "idict.h"
-#include "igc.h"		/* for gc_state_t */
-#include "igcstr.h"		/* for prototype */
+#include "igc.h"	/* for gc_state_t and gcst_get_memory_ptr() */
+#include "igcstr.h"	/* for prototype */
 #include "iname.h"
 #include "ipacked.h"
 #include "isstate.h"
-#include "iutil.h"		/* for packed_get */
+#include "iutil.h"	/* for packed_get */
 #include "ivmspace.h"
 #include "store.h"
 
@@ -285,8 +285,8 @@ object_size_valid(const obj_header_t * pre, uint size, const chunk_t * cp)
 }
 
 /* Validate all the objects in a chunk. */
-private void ialloc_validate_ref(P2(const ref *, gc_state_t *));
-private void ialloc_validate_ref_packed(P2(const ref_packed *, gc_state_t *));
+private void ialloc_validate_ref(const ref *, gc_state_t *);
+private void ialloc_validate_ref_packed(const ref_packed *, gc_state_t *);
 void
 ialloc_validate_chunk(const chunk_t * cp, gc_state_t * gcst)
 {
@@ -317,13 +317,16 @@ ialloc_validate_chunk(const chunk_t * cp, gc_state_t * gcst)
 	gs_ptr_type_t ptype;
 
 	if (proc != gs_no_struct_enum_ptrs)
-	    for (; (ptype = (*proc) (pre + 1, size, index, &eptr, pre->o_type, NULL)) != 0; ++index)
+	    for (; (ptype = (*proc) (gcst_get_memory_ptr(gcst), 
+				     pre + 1, size, index, &eptr, 
+				     pre->o_type, gcst)) != 0; ++index) {
 		if (eptr.ptr == 0)
 		    DO_NOTHING;
 		else if (ptype == ptr_struct_type)
 		    ialloc_validate_object(eptr.ptr, NULL, gcst);
 		else if (ptype == ptr_ref_type)
 		    ialloc_validate_ref_packed(eptr.ptr, gcst);
+	    }
     }
     END_OBJECTS_SCAN
 }
@@ -331,10 +334,12 @@ ialloc_validate_chunk(const chunk_t * cp, gc_state_t * gcst)
 private void
 ialloc_validate_ref_packed(const ref_packed * rp, gc_state_t * gcst)
 {
+    const gs_memory_t *cmem = gcst->spaces.memories.named.system->stable_memory;
+
     if (r_is_packed(rp)) {
 	ref unpacked;
 
-	packed_get(rp, &unpacked);
+	packed_get(cmem, rp, &unpacked);
 	ialloc_validate_ref(&unpacked, gcst);
     } else {
 	ialloc_validate_ref((const ref *)rp, gcst);
@@ -347,6 +352,7 @@ ialloc_validate_ref(const ref * pref, gc_state_t * gcst)
     const ref *rptr;
     const char *tname;
     uint size;
+    const gs_memory_t *cmem = gcst->spaces.memories.named.system->stable_memory;
 
     if (!gs_debug_c('?'))
 	return;			/* no check */
@@ -367,15 +373,15 @@ cks:	    if (optr != 0)
 		ialloc_validate_object(optr, NULL, gcst);
 	    break;
 	case t_name:
-	    if (name_index_ptr(name_index(pref)) != pref->value.pname) {
+	    if (name_index_ptr(cmem, name_index(cmem, pref)) != pref->value.pname) {
 		lprintf3("At 0x%lx, bad name %u, pname = 0x%lx\n",
-			 (ulong) pref, (uint)name_index(pref),
+			 (ulong) pref, (uint)name_index(cmem, pref),
 			 (ulong) pref->value.pname);
 		break;
 	    } {
 		ref sref;
 
-		name_string_ref(pref, &sref);
+		name_string_ref(cmem, pref, &sref);
 		if (r_space(&sref) != avm_foreign &&
 		    !gc_locate(sref.value.const_bytes, gcst)
 		    ) {
@@ -467,7 +473,7 @@ ialloc_validate_object(const obj_header_t * ptr, const chunk_t * cp,
     if (otype == &st_free) {
 	lprintf3("Reference to free object 0x%lx(%lu), in chunk 0x%lx!\n",
 		 (ulong) ptr, (ulong) size, (ulong) cp);
-	gs_abort();
+	gs_abort(gcst->heap);
     }
     if ((cp != 0 && !object_size_valid(pre, size, cp)) ||
 	otype->ssize == 0 ||
@@ -479,7 +485,7 @@ ialloc_validate_object(const obj_header_t * ptr, const chunk_t * cp,
 		 (ulong) ptr, (ulong) size);
 	dprintf2(" ssize = %u, in chunk 0x%lx!\n",
 		 otype->ssize, (ulong) cp);
-	gs_abort();
+	gs_abort(gcst->heap);
     }
 }
 
