@@ -1,28 +1,25 @@
 /* Copyright (C) 1993, 2000 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: iparam.c,v $ $Revision: 1.5.2.1 $ */
+/* $Id: iparam.c,v 1.10 2004/08/04 19:36:13 stefan Exp $ */
 /* Interpreter implementations of parameter dictionaries */
 #include "memory_.h"
 #include "string_.h"
 #include "ghost.h"
-#include "errors.h"
+#include "ierrors.h"
 #include "oper.h"		/* for check_type */
 #include "opcheck.h"
 #include "ialloc.h"
@@ -49,7 +46,7 @@ ref_param_key(const iparam_list * plist, gs_param_name pkey, ref * pkref)
 	make_int(pkref, key);
 	return 0;
     } else
-	return name_ref((const byte *)pkey, strlen(pkey), pkref, 0);
+        return name_ref(plist->memory, (const byte *)pkey, strlen(pkey), pkref, 0);
 }
 
 /* Fill in a gs_param_key_t from a name or int ref. */
@@ -59,7 +56,7 @@ ref_to_key(const ref * pref, gs_param_key_t * key, iparam_list *plist)
     if (r_has_type(pref, t_name)) {
 	ref nref;
 
-	name_string_ref(pref, &nref);
+	name_string_ref(plist->memory, pref, &nref);
 	key->data = nref.value.const_bytes;
 	key->size = r_size(&nref);
 	key->persistent = false; /* names may be freed */
@@ -85,9 +82,9 @@ ref_to_key(const ref * pref, gs_param_key_t * key, iparam_list *plist)
 /* ================ Writing parameters to refs ================ */
 
 /* Forward references */
-private int array_new_indexed_plist_write(P4(dict_param_list *plist,
-					     ref *parray, const ref *pwanted,
-					     gs_ref_memory_t *imem));
+private int array_new_indexed_plist_write(dict_param_list *plist,
+					  ref *parray, const ref *pwanted,
+					  gs_ref_memory_t *imem);
 
 /* ---------------- Generic writing procedures ---------------- */
 
@@ -105,12 +102,12 @@ private const gs_param_list_procs ref_write_procs =
     NULL,			/* request */
     ref_param_requested
 };
-private int ref_array_param_requested(P5(const iparam_list *, gs_param_name,
-					 ref *, uint, client_name_t));
-private int ref_param_write(P3(iparam_list *, gs_param_name, const ref *));
-private int ref_param_write_string_value(P3(ref *, const gs_param_string *,
-					    gs_ref_memory_t *));
-private int ref_param_write_name_value(P2(ref *, const gs_param_string *));
+private int ref_array_param_requested(const iparam_list *, gs_param_name,
+				      ref *, uint, client_name_t);
+private int ref_param_write(iparam_list *, gs_param_name, const ref *);
+private int ref_param_write_string_value(ref *, const gs_param_string *,
+					 gs_ref_memory_t *);
+private int ref_param_write_name_value(const gs_memory_t *mem, ref *, const gs_param_string *);
 private int
 ref_param_make_int(ref *pe, const void *pvalue, uint i, gs_ref_memory_t *imem)
 {
@@ -135,14 +132,14 @@ ref_param_make_string(ref *pe, const void *pvalue, uint i, gs_ref_memory_t *imem
 private int
 ref_param_make_name(ref * pe, const void *pvalue, uint i, gs_ref_memory_t *imem)
 {
-    return ref_param_write_name_value(pe,
+    return ref_param_write_name_value((const gs_memory_t *)imem, pe,
 			 &((const gs_param_string_array *)pvalue)->data[i]);
 }
 private int
 ref_param_write_typed_array(gs_param_list * plist, gs_param_name pkey,
 			    void *pvalue, uint count,
-			    int (*make)(P4(ref *, const void *, uint,
-					   gs_ref_memory_t *)))
+			    int (*make)(ref *, const void *, uint,
+					gs_ref_memory_t *))
 {
     iparam_list *const iplist = (iparam_list *) plist;
     ref value;
@@ -238,7 +235,7 @@ ref_param_write_typed(gs_param_list * plist, gs_param_name pkey,
 	case gs_param_type_name:
 	    if (!ref_param_requested(plist, pkey))
 		return 0;
-	    code = ref_param_write_name_value(&value, &pvalue->value.n);
+	    code = ref_param_write_name_value(iplist->memory, &value, &pvalue->value.n);
 	    break;
 	case gs_param_type_int_array:
 	    return ref_param_write_typed_array(plist, pkey, &pvalue->value.ia,
@@ -325,9 +322,9 @@ ref_param_write_string_value(ref * pref, const gs_param_string * pvalue,
 
 /* Prepare to write a name value. */
 private int
-ref_param_write_name_value(ref * pref, const gs_param_string * pvalue)
+ref_param_write_name_value(const gs_memory_t *mem, ref * pref, const gs_param_string * pvalue)
 {
-    return name_ref(pvalue->data, pvalue->size, pref,
+    return name_ref(mem, pvalue->data, pvalue->size, pref,
 		    (pvalue->persistent ? 0 : 1));
 }
 
@@ -521,12 +518,13 @@ private const gs_param_list_procs ref_read_procs =
     ref_param_read_signal_error,
     ref_param_read_commit
 };
-private int ref_param_read(P4(iparam_list *, gs_param_name,
-			      iparam_loc *, int));
-private int ref_param_read_string_value(P2(const iparam_loc *,
-					   gs_param_string *));
-private int ref_param_read_array(P3(iparam_list *, gs_param_name,
-				    iparam_loc *));
+private int ref_param_read(iparam_list *, gs_param_name,
+			   iparam_loc *, int);
+private int ref_param_read_string_value(const gs_memory_t *mem,
+					const iparam_loc *,
+					gs_param_string *);
+private int ref_param_read_array(iparam_list *, gs_param_name,
+				 iparam_loc *);
 
 #define iparam_note_error(loc, code)\
   gs_note_error(*(loc).presult = code)
@@ -559,7 +557,7 @@ ref_param_read_int_array(gs_param_list * plist, gs_param_name pkey,
     for (i = 0; i < size; i++) {
 	ref elt;
 
-	array_get(loc.pvalue, i, &elt);
+	array_get(plist->memory, loc.pvalue, i, &elt);
 	if (!r_has_type(&elt, t_integer)) {
 	    code = gs_note_error(e_typecheck);
 	    break;
@@ -604,7 +602,7 @@ ref_param_read_float_array(gs_param_list * plist, gs_param_name pkey,
     aref = *loc.pvalue;
     loc.pvalue = &elt;
     for (i = 0; code >= 0 && i < size; i++) {
-	array_get(&aref, i, &elt);
+        array_get(plist->memory, &aref, i, &elt);
 	code = float_param(&elt, pfv + i);
     }
     if (code < 0) {
@@ -640,15 +638,15 @@ ref_param_read_string_array(gs_param_list * plist, gs_param_name pkey,
     if (r_has_type(&aref, t_array)) {
 	for (i = 0; code >= 0 && i < size; i++) {
 	    loc.pvalue = aref.value.refs + i;
-	    code = ref_param_read_string_value(&loc, psv + i);
+	    code = ref_param_read_string_value(plist->memory, &loc, psv + i);
 	}
     } else {
 	ref elt;
 
 	loc.pvalue = &elt;
 	for (i = 0; code >= 0 && i < size; i++) {
-	    array_get(&aref, i, &elt);
-	    code = ref_param_read_string_value(&loc, psv + i);
+	    array_get(plist->memory, &aref, i, &elt);
+	    code = ref_param_read_string_value(plist->memory, &loc, psv + i);
 	}
     }
     if (code < 0) {
@@ -735,7 +733,7 @@ ref_param_read_typed(gs_param_list * plist, gs_param_name pkey,
 	     * on the type of the first element of the array.  If that
 	     * fails, we try again with more general types.
 	     */
-	    array_get(loc.pvalue, 0, &elt);
+	    array_get(plist->memory, loc.pvalue, 0, &elt);
 	    switch (r_type(&elt)) {
 		case t_integer:
 		    pvalue->type = gs_param_type_int_array;
@@ -793,7 +791,7 @@ ref_param_read_typed(gs_param_list * plist, gs_param_name pkey,
 	    return 0;
 	case t_name:
 	    pvalue->type = gs_param_type_name;
-	    return ref_param_read_string_value(&loc, &pvalue->value.n);
+	    return ref_param_read_string_value(plist->memory, &loc, &pvalue->value.n);
 	case t_null:
 	    pvalue->type = gs_param_type_null;
 	    return 0;
@@ -803,7 +801,7 @@ ref_param_read_typed(gs_param_list * plist, gs_param_name pkey,
 	    return 0;
 	case t_string:
 	    pvalue->type = gs_param_type_string;
-	    return ref_param_read_string_value(&loc, &pvalue->value.s);
+	    return ref_param_read_string_value(plist->memory, &loc, &pvalue->value.s);
 	default:
 	    break;
     }
@@ -869,7 +867,7 @@ ref_param_get_next_key(gs_param_list * plist, gs_param_enumerator_t * penum,
 
 /* Read a string value. */
 private int
-ref_param_read_string_value(const iparam_loc * ploc, gs_param_string * pvalue)
+ref_param_read_string_value(const gs_memory_t *mem, const iparam_loc * ploc, gs_param_string * pvalue)
 {
     const ref *pref = ploc->pvalue;
 
@@ -877,7 +875,7 @@ ref_param_read_string_value(const iparam_loc * ploc, gs_param_string * pvalue)
 	case t_name: {
 	    ref nref;
 
-	    name_string_ref(pref, &nref);
+	    name_string_ref(mem, pref, &nref);
 	    pvalue->data = nref.value.const_bytes;
 	    pvalue->size = r_size(&nref);
 	    pvalue->persistent = true;
