@@ -14,7 +14,7 @@
   San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$Id: zdevice2.c,v 1.4.2.2 2002/02/22 19:46:00 ray Exp $ */
+/* $Id: zdevice2.c,v 1.9 2004/08/04 19:36:13 stefan Exp $ */
 /* Level 2 device operators */
 #include "math_.h"
 #include "memory_.h"
@@ -32,8 +32,8 @@
 #include "gsstate.h"
 
 /* Forward references */
-private int z2copy_gstate(P1(i_ctx_t *));
-private int push_callout(P2(i_ctx_t *, const char *));
+private int z2copy_gstate(i_ctx_t *);
+private int push_callout(i_ctx_t *, const char *);
 
 /* Extend the `copy' operator to deal with gstates. */
 /* This is done with a hack -- we know that gstates are the only */
@@ -254,7 +254,7 @@ restore_page_device(const gs_state * pgs_old, const gs_state * pgs_new)
     gx_device *dev_new;
     gx_device *dev_t1;
     gx_device *dev_t2;
-    bool samepagedevice = obj_eq(&gs_int_gstate(pgs_old)->pagedevice,
+    bool samepagedevice = obj_eq(dev_old->memory, &gs_int_gstate(pgs_old)->pagedevice,
     	&gs_int_gstate(pgs_new)->pagedevice);
 
     if ((dev_t1 = (*dev_proc(dev_old, get_page_device)) (dev_old)) == 0)
@@ -367,7 +367,7 @@ push_callout(i_ctx_t *i_ctx_p, const char *callout_name)
     int code;
 
     check_estack(1);
-    code = name_enter_string(callout_name, esp + 1);
+    code = name_enter_string(imemory, callout_name, esp + 1);
     if (code < 0)
 	return code;
     ++esp;
