@@ -1,23 +1,20 @@
 /* Copyright (C) 1997, 1998, 1999, 2000 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: zfcmap.c,v $ $Revision: 1.8.2.2 $ */
+/* $Id: zfcmap.c,v 1.16 2004/09/20 06:13:03 ray Exp $ */
 /* CMap creation operator */
 #include "memory_.h"
 #include "ghost.h"
@@ -25,7 +22,7 @@
 #include "gsmatrix.h"		/* for gxfont.h */
 #include "gsstruct.h"
 #include "gsutil.h"		/* for bytes_compare */
-#include "gxfcmap.h"
+#include "gxfcmap1.h"
 #include "gxfont.h"
 #include "ialloc.h"
 #include "icid.h"
@@ -52,7 +49,7 @@ free_code_map(gx_code_map_t * pcmap, gs_memory_t * mem)
 	int i;
 
 	for (i = 0; i < pcmap->num_lookup; ++i) {
-	    gx_code_lookup_range_t *pclr = &pcmap->lookup[i];
+	    gx_cmap_lookup_range_t *pclr = &pcmap->lookup[i];
 
 	    if (pclr->value_type == CODE_VALUE_GLYPH)
 		gs_free_string(mem, pclr->values.data, pclr->values.size,
@@ -64,16 +61,28 @@ free_code_map(gx_code_map_t * pcmap, gs_memory_t * mem)
 
 /* Convert code ranges to internal form. */
 private int
-acquire_code_ranges(gs_cmap_t *cmap, const ref * pref, gs_memory_t * mem)
+acquire_code_ranges(gs_cmap_adobe1_t *cmap, const ref *pref, gs_memory_t *mem)
 {
-    uint num_ranges;
+    uint num_ranges = 0;
     gx_code_space_range_t *ranges;
-    uint i;
+    uint i, j, elem_sz;
+    ref elem;
 
-    if (!r_is_array(pref) || (num_ranges = r_size(pref)) == 0 ||
-	num_ranges & 1)
+    if (!r_is_array(pref))
+	return_error(e_rangecheck);
+    for (i=0; i < r_size(pref); i++) {
+        int code = array_get(mem, pref, i, &elem);
+        if (code < 0)
+            return code;
+        elem_sz = r_size(&elem);
+        if (elem_sz & 1)
+	    return_error(e_rangecheck);
+        num_ranges += elem_sz;
+    }
+    if (num_ranges == 0)
 	return_error(e_rangecheck);
     num_ranges >>= 1;
+
     ranges = (gx_code_space_range_t *)
 	gs_alloc_byte_array(mem, num_ranges, sizeof(gx_code_space_range_t),
 			    "acquire_code_ranges");
@@ -81,12 +90,16 @@ acquire_code_ranges(gs_cmap_t *cmap, const ref * pref, gs_memory_t * mem)
 	return_error(e_VMerror);
     cmap->code_space.ranges = ranges;
     cmap->code_space.num_ranges = num_ranges;
-    for (i = 0; i < num_ranges; ++i, ++ranges) {
+
+    for (i = 0; i < r_size(pref); i++) {
+        array_get(mem, pref, i, &elem);
+        elem_sz = r_size(&elem);
+        for (j = 0; j < elem_sz; j += 2) {
 	ref rfirst, rlast;
 	int size;
 
-	array_get(pref, i * 2L, &rfirst);
-	array_get(pref, i * 2L + 1, &rlast);
+	    array_get(mem, &elem, j, &rfirst);
+	    array_get(mem, &elem, j + 1, &rlast);
 	if (!r_has_type(&rfirst, t_string) ||
 	    !r_has_type(&rlast, t_string) ||
 	    (size = r_size(&rfirst)) == 0 || size > MAX_CMAP_CODE_SIZE ||
@@ -96,38 +109,57 @@ acquire_code_ranges(gs_cmap_t *cmap, const ref * pref, gs_memory_t * mem)
 	memcpy(ranges->first, rfirst.value.bytes, size);
 	memcpy(ranges->last, rlast.value.bytes, size);
 	ranges->size = size;
+            ++ranges;
+        }
     }
     return 0;
 }
 
 /* Convert a code map to internal form. */
 private int
-acquire_code_map(gx_code_map_t * pcmap, const ref * pref, gs_cmap_t * root,
-		 gs_memory_t * mem)
+acquire_code_map(gx_code_map_t *pcmap, const ref *pref, gs_cmap_adobe1_t *root,
+		 gs_memory_t *mem)
 {
-    uint num_lookup;
-    gx_code_lookup_range_t *pclr;
+    uint num_lookup = 0;
+    gx_cmap_lookup_range_t *pclr;
     long i;
+    ref elem;
+    uint elem_sz;
 
-    if (!r_is_array(pref) || (num_lookup = r_size(pref)) % 5 != 0)
+    if (!r_is_array(pref))
 	return_error(e_rangecheck);
+    for (i=0; i < r_size(pref); i++) {
+        int code = array_get(mem, pref, i, &elem);
+        if (code < 0)
+            return code;
+        elem_sz = r_size(&elem);
+        if (elem_sz % 5 != 0)
+	return_error(e_rangecheck);
+        num_lookup += elem_sz;
+    }
     num_lookup /= 5;
-    pclr = gs_alloc_struct_array(mem, num_lookup, gx_code_lookup_range_t,
-				 &st_code_lookup_range_element,
+    pclr = gs_alloc_struct_array(mem, num_lookup, gx_cmap_lookup_range_t,
+				 &st_cmap_lookup_range_element,
 				 "acquire_code_map(lookup ranges)");
     if (pclr == 0)
 	return_error(e_VMerror);
     memset(pclr, 0, sizeof(*pclr) * num_lookup);
     pcmap->lookup = pclr;
     pcmap->num_lookup = num_lookup;
-    for (i = 0; i < num_lookup * 5; i += 5, ++pclr) {
+    
+   
+    for (i = 0; i < r_size(pref); i++) {
+        uint j;
+        array_get(mem, pref, i, &elem);
+        elem_sz = r_size(&elem);
+        for (j = 0; j < elem_sz; j += 5) {
 	ref rprefix, rmisc, rkeys, rvalues, rfxs;
 
-	array_get(pref, i, &rprefix);
-	array_get(pref, i + 1, &rmisc);
-	array_get(pref, i + 2, &rkeys);
-	array_get(pref, i + 3, &rvalues);
-	array_get(pref, i + 4, &rfxs);
+	    array_get(mem, &elem, j, &rprefix);
+	    array_get(mem, &elem, j + 1, &rmisc);
+	    array_get(mem, &elem, j + 2, &rkeys);
+	    array_get(mem, &elem, j + 3, &rvalues);
+	    array_get(mem, &elem, j + 4, &rfxs);
 
 	if (!r_has_type(&rprefix, t_string) ||
 	    !r_has_type(&rmisc, t_string) ||
@@ -151,13 +183,13 @@ acquire_code_map(gx_code_map_t * pcmap, const ref * pref, gs_cmap_t * root,
 	    /* This is a single entry consisting only of the prefix. */
 	    if (r_size(&rkeys) != 0)
 		return_error(e_rangecheck);
-	    pclr->num_keys = 1;
+	    pclr->num_entries = 1;
 	} else {
 	    int step = pclr->key_size * (pclr->key_is_range ? 2 : 1);
 
 	    if (r_size(&rkeys) % step != 0)
 		return_error(e_rangecheck);
-	    pclr->num_keys = r_size(&rkeys) / step;
+	    pclr->num_entries = r_size(&rkeys) / step;
 	}
 	pclr->keys.data = rkeys.value.bytes,
 	    pclr->keys.size = r_size(&rkeys);
@@ -166,18 +198,18 @@ acquire_code_map(gx_code_map_t * pcmap, const ref * pref, gs_cmap_t * root,
 	if (r_has_type(&rvalues, t_string)) {
 	    if (pclr->value_type == CODE_VALUE_GLYPH)
 		return_error(e_rangecheck);
-	    if (r_size(&rvalues) % pclr->num_keys != 0 ||
-		r_size(&rvalues) / pclr->num_keys != pclr->value_size)
+	    if (r_size(&rvalues) % pclr->num_entries != 0 ||
+		r_size(&rvalues) / pclr->num_entries != pclr->value_size)
 		return_error(e_rangecheck);
 	    pclr->values.data = rvalues.value.bytes,
 		pclr->values.size = r_size(&rvalues);
 	} else {
-	    uint values_size = pclr->num_keys * pclr->value_size;
+	    uint values_size = pclr->num_entries * pclr->value_size;
 	    long k;
 	    byte *pvalue;
 
 	    if (pclr->value_type != CODE_VALUE_GLYPH ||
-		r_size(&rvalues) != pclr->num_keys ||
+		r_size(&rvalues) != pclr->num_entries ||
 		pclr->value_size > sizeof(gs_glyph))
 		return_error(e_rangecheck);
 	    pclr->values.data = gs_alloc_string(mem, values_size,
@@ -186,15 +218,15 @@ acquire_code_map(gx_code_map_t * pcmap, const ref * pref, gs_cmap_t * root,
 		return_error(e_VMerror);
 	    pclr->values.size = values_size;
 	    pvalue = pclr->values.data;
-	    for (k = 0; k < pclr->num_keys; ++k) {
+	    for (k = 0; k < pclr->num_entries; ++k) {
 		ref rvalue;
 		gs_glyph value;
 		int i;
 
-		array_get(&rvalues, k, &rvalue);
+		array_get(mem, &rvalues, k, &rvalue);
 		if (!r_has_type(&rvalue, t_name))
 		    return_error(e_rangecheck);
-		value = name_index(&rvalue);
+		value = name_index(mem, &rvalue);
 		/*
 		 * We need a special check here because some CPUs cannot
 		 * shift by the full size of an int or long.
@@ -209,6 +241,8 @@ acquire_code_map(gx_code_map_t * pcmap, const ref * pref, gs_cmap_t * root,
 	}
 	check_int_leu_only(rfxs, 0xff);
 	pclr->font_index = (int)rfxs.value.intval;
+            ++pclr;
+        }
     }
     return 0;
 }
@@ -241,10 +275,10 @@ acquire_cid_system_info(ref *psia, const ref *op)
  * null, return 1.
  */
 private int
-get_cid_system_info(gs_cid_system_info_t *pcidsi, const ref *psia, uint index)
+get_cid_system_info(const gs_memory_t *mem, gs_cid_system_info_t *pcidsi, const ref *psia, uint index)
 {
     ref rcidsi;
-    int code = array_get(psia, (long)index, &rcidsi);
+    int code = array_get(mem, psia, (long)index, &rcidsi);
 
     if (code < 0 || r_has_type(&rcidsi, t_null)) {
 	cid_system_info_set_null(pcidsi);
@@ -287,10 +321,16 @@ ztype0_get_cmap(const gs_cmap_t **ppcmap, const ref *pfdepvector,
     uint num_fonts;
     uint i;
 
+    /*
+     * We have no way of checking whether the CodeMap is a concrete
+     * subclass of gs_cmap_t, so we just check that it is in fact a
+     * t_struct and is large enough.
+     */
     if (dict_find_string(op, "CMap", &prcmap) <= 0 ||
 	!r_has_type(prcmap, t_dictionary) ||
 	dict_find_string(prcmap, "CodeMap", &pcodemap) <= 0 ||
-	!r_has_stype(pcodemap, imem, st_cmap)
+	!r_is_struct(pcodemap) ||
+	gs_object_size(imem, r_ptr(pcodemap, gs_cmap_t)) < sizeof(gs_cmap_t)
 	)
 	return_error(e_invalidfont);
     pcmap = r_ptr(pcodemap, gs_cmap_t);
@@ -298,7 +338,7 @@ ztype0_get_cmap(const gs_cmap_t **ppcmap, const ref *pfdepvector,
     for (i = 0; i < num_fonts; ++i) {
 	ref rfdep, rfsi;
 
-	array_get(pfdepvector, (long)i, &rfdep);
+	array_get(imem, pfdepvector, (long)i, &rfdep);
 	code = acquire_cid_system_info(&rfsi, &rfdep);
 	if (code < 0)
 	    return code;
@@ -335,15 +375,16 @@ ztype0_get_cmap(const gs_cmap_t **ppcmap, const ref *pfdepvector,
  * For details, see lib/gs_cmap.ps and the Adobe documentation.
  */
 private int
-zfcmap_glyph_name(gs_glyph glyph, gs_const_string *pstr, void *proc_data)
+zfcmap_glyph_name(const gs_memory_t *mem, 
+		  gs_glyph glyph, gs_const_string *pstr, void *proc_data)
 {
     ref nref, nsref;
     int code = 0;
 
-    /*code = */name_index_ref((uint)glyph, &nref);
+    /*code = */name_index_ref(mem, (uint)glyph, &nref);
     if (code < 0)
 	return code;
-    name_string_ref(&nref, &nsref);
+    name_string_ref(mem, &nref, &nsref);
     pstr->data = nsref.value.const_bytes;
     pstr->size = r_size(&nsref);
     return 0;
@@ -358,25 +399,12 @@ zbuildcmap(i_ctx_t *i_ctx_p)
     ref *pcodemapdata;
     ref *pcodemap;
     ref rname, rcidsi, rcoderanges, rdefs, rnotdefs;
-    gs_cmap_t *pcmap = 0;
-    gs_cid_system_info_t *pcidsi = 0;
+    gs_cmap_adobe1_t *pcmap = 0;
     ref rcmap;
     uint i;
 
     check_type(*op, t_dictionary);
     check_dict_write(*op);
-    pcmap = ialloc_struct(gs_cmap_t, &st_cmap, "zbuildcmap(cmap)");
-    if (pcmap == 0) {
-	code = gs_note_error(e_VMerror);
-	goto fail;
-    }
-    gs_cmap_init(pcmap);
-    if ((code = dict_int_param(op, "CMapType", 0, 1, -1, &pcmap->CMapType)) < 0 ||
-	(code = dict_float_param(op, "CMapVersion", 0.0, &pcmap->CMapVersion)) < 0 ||
-	(code = dict_uid_param(op, &pcmap->uid, 0, imemory, i_ctx_p)) < 0 ||
-	(code = dict_int_param(op, "WMode", 0, 1, 0, &pcmap->WMode)) < 0
-	)
-	goto fail;
     if ((code = dict_find_string(op, "CMapName", &pcmapname)) <= 0) {
 	code = gs_note_error(e_rangecheck);
 	goto fail;
@@ -385,16 +413,7 @@ zbuildcmap(i_ctx_t *i_ctx_p)
 	code = gs_note_error(e_typecheck);
 	goto fail;
     }
-    name_string_ref(pcmapname, &rname);
-    pcmap->CMapName.data = rname.value.const_bytes;
-    pcmap->CMapName.size = r_size(&rname);
-    if (dict_find_string(op, "UIDOffset", &puidoffset) > 0) {
-	if (!r_has_type(puidoffset, t_integer)) {
-	    code = gs_note_error(e_typecheck);
-	    goto fail;
-	}
-	pcmap->UIDOffset = puidoffset->value.intval; /* long, not int */
-    }
+    name_string_ref(imemory, pcmapname, &rname);
     if (dict_find_string(op, ".CodeMapData", &pcodemapdata) <= 0 ||
 	!r_has_type(pcodemapdata, t_array) ||
 	r_size(pcodemapdata) != 3 ||
@@ -406,23 +425,31 @@ zbuildcmap(i_ctx_t *i_ctx_p)
     }
     if ((code = acquire_cid_system_info(&rcidsi, op)) < 0)
 	goto fail;
-    pcidsi = ialloc_struct_array(r_size(&rcidsi), gs_cid_system_info_t,
-				 &st_cid_system_info_element,
-				 "zbuildcmap(CIDSystemInfo)");
-    if (pcidsi == 0) {
-	code = gs_note_error(e_VMerror);
+    if ((code = gs_cmap_adobe1_alloc(&pcmap, 0, rname.value.const_bytes,
+				     r_size(&rname), r_size(&rcidsi),
+				     0, 0, 0, 0, 0, imemory)) < 0)
 	goto fail;
+    if ((code = dict_int_param(op, "CMapType", 0, 1, 0, &pcmap->CMapType)) < 0 ||
+	(code = dict_float_param(op, "CMapVersion", 0.0, &pcmap->CMapVersion)) < 0 ||
+	(code = dict_uid_param(op, &pcmap->uid, 0, imemory, i_ctx_p)) < 0 ||
+	(code = dict_int_param(op, "WMode", 0, 1, 0, &pcmap->WMode)) < 0
+	)
+	goto fail;
+    if (dict_find_string(op, "UIDOffset", &puidoffset) > 0) {
+	if (!r_has_type(puidoffset, t_integer)) {
+	    code = gs_note_error(e_typecheck);
+	    goto fail;
+	}
+	pcmap->UIDOffset = puidoffset->value.intval; /* long, not int */
     }
-    pcmap->CIDSystemInfo = pcidsi;
-    pcmap->num_fonts = r_size(&rcidsi);
     for (i = 0; i < r_size(&rcidsi); ++i) {
-	code = get_cid_system_info(pcidsi + i, &rcidsi, i);
+        code = get_cid_system_info(imemory, pcmap->CIDSystemInfo + i, &rcidsi, i);
 	if (code < 0)
 	    goto fail;
     }
-    array_get(pcodemapdata, 0L, &rcoderanges);
-    array_get(pcodemapdata, 1L, &rdefs);
-    array_get(pcodemapdata, 2L, &rnotdefs);
+    array_get(imemory, pcodemapdata, 0L, &rcoderanges);
+    array_get(imemory, pcodemapdata, 1L, &rdefs);
+    array_get(imemory, pcodemapdata, 2L, &rnotdefs);
     if ((code = acquire_code_ranges(pcmap, &rcoderanges, imemory)) < 0)
 	goto fail;
     if ((code = acquire_code_map(&pcmap->def, &rdefs, pcmap, imemory)) < 0)
@@ -439,59 +466,19 @@ zbuildcmap(i_ctx_t *i_ctx_p)
 	goto fail;
     return zreadonly(i_ctx_p);
 fail:
-    free_code_map(&pcmap->notdef, imemory);
-    free_code_map(&pcmap->def, imemory);
-    ifree_object(pcmap, "zbuildcmap(cmap)");
-    ifree_object(pcidsi, "zbuildcmap(CIDSystemInfo)");
+    if (pcmap) {
+	free_code_map(&pcmap->notdef, imemory);
+	free_code_map(&pcmap->def, imemory);
+	ifree_object(pcmap->CIDSystemInfo, "zbuildcmap(CIDSystemInfo)");
+	ifree_object(pcmap, "zbuildcmap(cmap)");
+    }
     return code;
 }
 
-#if defined(DEBUG) || defined(PROFILE) || defined(TEST)
-
-#include "stream.h"
-#include "spprint.h"
-#include "files.h"
-#include "gdevpsf.h"
-
-/* <file> <cmap> .writecmap - */
-private int
-zfcmap_put_name_default(stream *s, const byte *str, uint size)
-{
-    stream_putc(s, '/');
-    stream_write(s, str, size);
-    return 0;
-}
-private int
-zwritecmap(i_ctx_t *i_ctx_p)
-{
-    os_ptr op = osp;
-    ref *pcodemap;
-    gs_cmap_t *pcmap;
-    int code;
-    stream *s;
-
-    check_type(*op, t_dictionary);
-    if (dict_find_string(op, "CodeMap", &pcodemap) <= 0 ||
-	!r_is_struct(pcodemap)
-	)
-	return_error(e_typecheck);
-    check_write_file(s, op - 1);
-    pcmap = r_ptr(pcodemap, gs_cmap_t);
-    code = psf_write_cmap(s, pcmap, zfcmap_put_name_default, NULL);
-    if (code >= 0)
-	pop(2);
-    return code;
-}
-
-#endif
-
 /* ------ Initialization procedure ------ */
 
 const op_def zfcmap_op_defs[] =
 {
     {"1.buildcmap", zbuildcmap},
-#if defined(DEBUG) || defined(PROFILE) || defined(TEST)
-    {"2.writecmap", zwritecmap},
-#endif
     op_def_end(0)
 };
