@@ -1,30 +1,29 @@
-/* Copyright (C) 1989, 2000 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 2000-2004, artofcode LLC.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: zfile.c,v $ $Revision: 1.5.2.4 $ */
+/* $Id: zfile.c,v 1.41 2004/08/04 19:36:13 stefan Exp $ */
 /* Non-I/O file operators */
 #include "memory_.h"
 #include "string_.h"
+#include "unistd_.h"
 #include "ghost.h"
 #include "gscdefs.h"		/* for gx_io_device_table */
 #include "gsutil.h"		/* for bytes_compare */
 #include "gp.h"
+#include "gpmisc.h"
 #include "gsfname.h"
 #include "gsstruct.h"		/* for registering root */
 #include "gxalloc.h"		/* for streams */
@@ -54,18 +53,22 @@ extern_gx_io_device_table();
 extern const char iodev_dtype_stdio[];
 
 /* Forward references: file name parsing. */
-private int parse_file_name(P2(const ref * op, gs_parsed_file_name_t * pfn));
-private int parse_real_file_name(P4(const ref * op,
-				    gs_parsed_file_name_t * pfn,
-				    gs_memory_t *mem, client_name_t cname));
+private int parse_file_name(const ref * op, gs_parsed_file_name_t * pfn, bool safemode);
+private int parse_real_file_name(const ref * op,
+				 gs_parsed_file_name_t * pfn,
+				 gs_memory_t *mem, client_name_t cname);
+private int parse_file_access_string(const ref *op, char file_access[4]);
 
 /* Forward references: other. */
-private int zopen_file(P5(i_ctx_t *, const gs_parsed_file_name_t *pfn,
-			  const char *file_access, stream **ps,
-			  gs_memory_t *mem));
+private int execfile_finish(i_ctx_t *);
+private int execfile_cleanup(i_ctx_t *);
+private int zopen_file(i_ctx_t *, const gs_parsed_file_name_t *pfn,
+		       const char *file_access, stream **ps,
+		       gs_memory_t *mem);
 private iodev_proc_open_file(iodev_os_open_file);
-private int execfile_finish(P1(i_ctx_t *));
-private int execfile_cleanup(P1(i_ctx_t *));
+private void file_init_stream(stream *s, FILE *file, const char *fmode,
+			      byte *buffer, uint buffer_size);
+stream_proc_report_error(filter_report_error);
 
 /*
  * Since there can be many file objects referring to the same file/stream,
@@ -140,89 +143,98 @@ make_invalid_file(ref * fp)
 }
 
 /* Check a file name for permission by stringmatch on one of the */
-/* strings of the permitgroup array */
+/* strings of the permitgroup array. */
 private int
-check_file_permissions(i_ctx_t *i_ctx_p, const char *fname, int len,
+check_file_permissions_reduced(i_ctx_t *i_ctx_p, const char *fname, int len,
 			const char *permitgroup)
 {
     long i;
     ref *permitlist = NULL;
     /* an empty string (first character == 0) if '\' character is */
     /* recognized as a file name separator as on DOS & Windows	  */
-    const char *filenamesep = gp_file_name_concat_string("\\", 1);
+    const char *win_sep2 = "\\";
+    bool use_windows_pathsep = (gs_file_name_check_separator(win_sep2, 1, win_sep2) == 1);
+    uint plen = gp_file_name_parents(fname, len);
+
+    /* Assuming a reduced file name. */
 
-    /*
-     * We can't know where we will get to if we reference the parent
-     * directory, so don't allow access if LockFilePermissions is true
-     * Also check here for the %pipe device which is illegal when
-     * LockFilePermissions is true. In the future we might want to allow
-     * the %pipe device to be included on the PermitFile... paths, but
-     * for now it is simply disallowed.
-     */
-    if (i_ctx_p->LockFilePermissions &&
-	    (gp_file_name_references_parent(fname, len) ||
-               string_match(fname, len, "%pipe*", 5, NULL))
-       ) {
-	return e_invalidfileaccess;
-    }
     if (dict_find_string(&(i_ctx_p->userparams), permitgroup, &permitlist) <= 0)
         return 0;	/* if Permissions not found, just allow access */
+
     for (i=0; i<r_size(permitlist); i++) {
         ref permitstring;
 	const string_match_params win_filename_params = {
 		'*', '?', '\\', true, true	/* ignore case & '/' == '\\' */
 	};
+	const byte *permstr;
+	uint permlen;
+	int cwd_len = 0;
 
-	if (array_get(permitlist, i, &permitstring) < 0 ||
-	    r_type(&permitstring) != t_string
+	if (array_get(imemory, permitlist, i, &permitstring) < 0 ||
+	    r_type(&permitstring) != t_string 
 	   )    
 	    break;	/* any problem, just fail */
-        if (string_match(fname, len, permitstring.value.bytes,
-		r_size(&permitstring), 
-		filenamesep[0] == 0 ? &win_filename_params : NULL)
-	   )
+	permstr = permitstring.value.bytes;
+	permlen = r_size(&permitstring);
+	/* 
+	 * Check if any file name is permitted with "*".
+	 */
+	if (permlen == 1 && permstr[0] == '*')
+	    return 0;		/* success */
+	/* 
+	 * If the filename starts with parent references, 
+	 * the permission element must start with same number of parent references.
+	 */
+	if (plen != 0 && plen != gp_file_name_parents((const char *)permstr, permlen))
+	    continue;
+	cwd_len = gp_file_name_cwds((const char *)permstr, permlen);
+	/*
+	 * If the permission starts with "./", absolute paths
+	 * are not permitted.
+	 */
+	if (cwd_len > 0 && gp_file_name_is_absolute(fname, len))
+	    continue;
+	/*
+	 * If the permission starts with "./", relative paths
+	 * with no "./" are allowed as well as with "./".
+	 * 'fname' has no "./" because it is reduced.
+	 */
+        if (string_match( (const unsigned char*) fname, len,
+			  permstr + cwd_len, permlen - cwd_len, 
+		use_windows_pathsep ? &win_filename_params : NULL))
 	    return 0;		/* success */
     }
     /* not found */
     return e_invalidfileaccess;
 }
 
+/* Check a file name for permission by stringmatch on one of the */
+/* strings of the permitgroup array */
+private int
+check_file_permissions(i_ctx_t *i_ctx_p, const char *fname, int len,
+			const char *permitgroup)
+{
+    char fname_reduced[gp_file_name_sizeof];
+    uint rlen = sizeof(fname_reduced);
+
+    if (gp_file_name_reduce(fname, len, fname_reduced, &rlen) != gp_combine_success)
+	return e_invalidaccess;		/* fail if we couldn't reduce */
+    return check_file_permissions_reduced(i_ctx_p, fname_reduced, rlen, permitgroup);
+}
+
 /* <name_string> <access_string> file <file> */
 private int
 zfile(i_ctx_t *i_ctx_p)
 {
     os_ptr op = osp;
-    char file_access[3];
+    char file_access[4];
     gs_parsed_file_name_t pname;
-    const byte *astr;
-    int code;
+    int code = parse_file_access_string(op, file_access);
     stream *s;
 
-    check_read_type(*op, t_string);
-    astr = op->value.const_bytes;
-    switch (r_size(op)) {
-	case 2:
-	    if (astr[1] != '+')
-		return_error(e_invalidfileaccess);
-	    file_access[1] = '+';
-	    file_access[2] = 0;
-	    break;
-	case 1:
-	    file_access[1] = 0;
-	    break;
-	default:
-	    return_error(e_invalidfileaccess);
-    }
-    switch (astr[0]) {
-	case 'r':
-	case 'w':
-	case 'a':
-	    break;
-	default:
-	    return_error(e_invalidfileaccess);
-    }
-    file_access[0] = astr[0];
-    code = parse_file_name(op - 1, &pname);
+    if (code < 0)
+	return code;
+    code = parse_file_name(op - 1, &pname, i_ctx_p->LockFilePermissions);
     if (code < 0)
 	return code;
 	/*
@@ -231,8 +243,29 @@ zfile(i_ctx_t *i_ctx_p)
 	 * more information.
 	 */
     if (pname.iodev && pname.iodev->dtype == iodev_dtype_stdio) {
+	bool statement = (strcmp(pname.iodev->dname, "%statementedit%") == 0);
+	bool lineedit = (strcmp(pname.iodev->dname, "%lineedit%") == 0);
 	if (pname.fname)
 	    return_error(e_invalidfileaccess);
+	if (statement || lineedit) {
+	    /* These need special code to support callouts */
+	    gx_io_device *indev = gs_findiodevice((const byte *)"%stdin", 6);
+	    stream *ins;
+	    if (strcmp(file_access, "r"))
+		return_error(e_invalidfileaccess);
+	    indev->state = i_ctx_p;
+	    code = (indev->procs.open_device)(indev, file_access, &ins, imemory);
+	    indev->state = 0;
+	    if (code < 0)
+		return code;
+	    check_ostack(2);
+	    push(2);
+	    make_stream_file(op - 3, ins, file_access);
+	    make_bool(op-2, statement);
+	    make_int(op-1, 0);
+	    make_string(op, icurrent_space, 0, NULL);
+	    return zfilelineedit(i_ctx_p);
+	}
 	pname.iodev->state = i_ctx_p;
 	code = (*pname.iodev->procs.open_device)(pname.iodev,
 						 file_access, &s, imemory);
@@ -254,6 +287,27 @@ zfile(i_ctx_t *i_ctx_p)
     return code;
 }
 
+/*
+ * Files created with .tempfile permit some operations even if the 
+ * temp directory is not explicitly named on the PermitFile... path 
+ * The names 'SAFETY' and 'tempfiles' are defined by gs_init.ps
+*/
+private bool
+file_is_tempfile(i_ctx_t *i_ctx_p, const ref *op)
+{
+    ref *SAFETY;
+    ref *tempfiles;
+    ref kname;
+
+    if (dict_find_string(systemdict, "SAFETY", &SAFETY) <= 0 ||
+	    dict_find_string(SAFETY, "tempfiles", &tempfiles) <= 0)
+	return false;
+    if (name_ref(imemory, op->value.bytes, r_size(op), &kname, -1) < 0 ||
+	    dict_find(tempfiles, &kname, &SAFETY) <= 0)
+	return false;
+    return true;
+}
+
 /* ------ Level 2 extensions ------ */
 
 /* <string> deletefile - */
@@ -268,7 +322,8 @@ zdeletefile(i_ctx_t *i_ctx_p)
 	return code;
     if (pname.iodev == iodev_default) {
 	if ((code = check_file_permissions(i_ctx_p, pname.fname, pname.len,
-		"PermitFileControl")) < 0 ) {
+		"PermitFileControl")) < 0 &&
+		 !file_is_tempfile(i_ctx_p, op)) {
 	    return code;
 	}
     }
@@ -281,27 +336,43 @@ zdeletefile(i_ctx_t *i_ctx_p)
 }
 
 /* <template> <proc> <scratch> filenameforall - */
-/****** NOT CONVERTED FOR IODEVICES YET ******/
-private int file_continue(P1(i_ctx_t *));
-private int file_cleanup(P1(i_ctx_t *));
+private int file_continue(i_ctx_t *);
+private int file_cleanup(i_ctx_t *);
 private int
 zfilenameforall(i_ctx_t *i_ctx_p)
 {
     os_ptr op = osp;
     file_enum *pfen;
-    int code;
+    gx_io_device *iodev = NULL;
+    gs_parsed_file_name_t pname;
+    int code = 0;
 
     check_write_type(*op, t_string);
     check_proc(op[-1]);
     check_read_type(op[-2], t_string);
-    /* Push a mark, the pattern, the scratch string, the enumerator, */
+    /* Push a mark, the iodev, devicenamelen, the scratch string, the enumerator, */
     /* and the procedure, and invoke the continuation. */
     check_estack(7);
-    pfen = gp_enumerate_files_init((char *)op[-2].value.bytes, r_size(op - 2), imemory);
+    /* Get the iodevice */
+    code = parse_file_name(op - 2, &pname, i_ctx_p->LockFilePermissions);
+    if (code < 0)
+	return code;
+    iodev = (pname.iodev == NULL) ? iodev_default : pname.iodev;
+
+    /* Check for several conditions that just cause us to return success */
+    if (pname.len == 0 || iodev->procs.enumerate_files == iodev_no_enumerate_files) {
+        pop(3);
+        return 0;	/* no pattern, or device not found -- just return */
+    }
+    pfen = iodev->procs.enumerate_files(iodev, (const char *)pname.fname,
+    		pname.len, imemory);
     if (pfen == 0)
 	return_error(e_VMerror);
     push_mark_estack(es_for, file_cleanup);
-    *++esp = op[-2];
+    ++esp;
+    make_istruct(esp, 0, iodev);
+    ++esp;
+    make_int(esp, r_size(op-2) - pname.len);
     *++esp = *op;
     ++esp;
     make_istruct(esp, 0, pfen);
@@ -317,19 +388,25 @@ file_continue(i_ctx_t *i_ctx_p)
     os_ptr op = osp;
     es_ptr pscratch = esp - 2;
     file_enum *pfen = r_ptr(esp - 1, file_enum);
+    long devlen = esp[-3].value.intval;
+    gx_io_device *iodev = r_ptr(esp - 4, gx_io_device);
     uint len = r_size(pscratch);
-    uint code =
-    gp_enumerate_files_next(pfen, (char *)pscratch->value.bytes, len);
+    uint code;
 
+    if (len < devlen)
+	return_error(e_rangecheck);	/* not even room for device len */
+    memcpy((char *)pscratch->value.bytes, iodev->dname, devlen);
+    code = iodev->procs.enumerate_next(pfen, (char *)pscratch->value.bytes + devlen,
+    		len - devlen);
     if (code == ~(uint) 0) {	/* all done */
-	esp -= 4;		/* pop proc, pfen, scratch, mark */
+	esp -= 5;		/* pop proc, pfen, devlen, iodev , mark */
 	return o_pop_estack;
     } else if (code > len)	/* overran string */
 	return_error(e_rangecheck);
     else {
 	push(1);
 	ref_assign(op, pscratch);
-	r_set_size(op, code);
+	r_set_size(op, code + devlen);
 	push_op_estack(file_continue);	/* come again */
 	*++esp = pscratch[2];	/* proc */
 	return o_push_estack;
@@ -339,7 +416,9 @@ file_continue(i_ctx_t *i_ctx_p)
 private int
 file_cleanup(i_ctx_t *i_ctx_p)
 {
-    gp_enumerate_files_close(r_ptr(esp + 4, file_enum));
+    gx_io_device *iodev = r_ptr(esp + 2, gx_io_device);
+
+    iodev->procs.enumerate_close(r_ptr(esp + 5, file_enum));
     return 0;
 }
 
@@ -357,13 +436,26 @@ zrenamefile(i_ctx_t *i_ctx_p)
     pname2.fname = 0;
     code = parse_real_file_name(op, &pname2, imemory, "renamefile(to)");
     if (code >= 0) {
-        if (pname1.iodev != pname2.iodev ||
-	      check_file_permissions(i_ctx_p, pname1.fname, pname1.len,
-	      				"PermitFileControl") < 0  ||
-	      check_file_permissions(i_ctx_p, pname2.fname, pname2.len,
+	if (pname1.iodev != pname2.iodev ) {
+	    if (pname1.iodev == iodev_default)
+		pname1.iodev = pname2.iodev;
+	    if (pname2.iodev == iodev_default)
+		pname2.iodev = pname1.iodev;
+	}
+	if (pname1.iodev != pname2.iodev ||
+	    (pname1.iodev == iodev_default &&
+		/*
+		 * We require FileControl permissions on the source path
+		 * unless it is a temporary file. Also, we require FileControl
+		 * and FileWriting permissions to the destination file/path.
+		 */
+	      ((check_file_permissions(i_ctx_p, pname1.fname, pname1.len,
+	      				"PermitFileControl") < 0 &&
+	          !file_is_tempfile(i_ctx_p, op - 1)) ||
+	      (check_file_permissions(i_ctx_p, pname2.fname, pname2.len,
 	      				"PermitFileControl") < 0 ||
 	      check_file_permissions(i_ctx_p, pname2.fname, pname2.len,
-	      				"PermitFileWriting") < 0 ) {
+	      				"PermitFileWriting") < 0 )))) {
 	    code = gs_note_error(e_invalidfileaccess);
 	} else {
 	    code = (*pname1.iodev->procs.rename_file)(pname1.iodev,
@@ -398,7 +490,7 @@ zstatus(i_ctx_t *i_ctx_p)
 	    {
 		gs_parsed_file_name_t pname;
 		struct stat fstat;
-		int code = parse_file_name(op, &pname);
+		int code = parse_file_name(op, &pname, i_ctx_p->LockFilePermissions);
 
 		if (code < 0)
 		    return code;
@@ -481,22 +573,6 @@ execfile_cleanup(i_ctx_t *i_ctx_p)
     return zclosefile(i_ctx_p);
 }
 
-/* <dir> .filenamedirseparator <string> */
-private int
-zfilenamedirseparator(i_ctx_t *i_ctx_p)
-{
-    os_ptr op = osp;
-    const char *sepr;
-
-    check_read_type(*op, t_string);
-    sepr =
-	gp_file_name_concat_string((const char *)op->value.const_bytes,
-				   r_size(op));
-    make_const_string(op, avm_foreign | a_readonly,
-		      strlen(sepr), (const byte *)sepr);
-    return 0;
-}
-
 /* - .filenamelistseparator <string> */
 private int
 zfilenamelistseparator(i_ctx_t *i_ctx_p)
@@ -534,7 +610,7 @@ zlibfile(i_ctx_t *i_ctx_p)
     stream *s;
 
     check_ostack(2);
-    code = parse_file_name(op, &pname);
+    code = parse_file_name(op, &pname, i_ctx_p->LockFilePermissions);
     if (code < 0)
 	return code;
     if (pname.iodev == NULL)
@@ -557,24 +633,7 @@ zlibfile(i_ctx_t *i_ctx_p)
     } else {
 	ref fref;
 
-	/* Skip checks if this file name came from the command line. */
-	if (i_ctx_p->LockFilePermissions &&
-	    (i_ctx_p->filearg == NULL ||
-	    bytes_compare(op->value.bytes, r_size(op),
-	    (const byte *)i_ctx_p->filearg, strlen(i_ctx_p->filearg)) != 0)
-	   ) {
-	    /* Check to see if this file is allowed */
-	    /* Possibly we should allow access to parent directories if */
-	    /* PermitFileReading includes (*), but this is unlikely if  */
-	    /* we are locked. */
-	    if (gp_file_name_references_parent(pname.fname, pname.len) ||
-		(gp_file_name_is_absolute(pname.fname, pname.len) &&
-		check_file_permissions(i_ctx_p, pname.fname, pname.len,
-					"PermitFileReading") < 0)
-	       ) 
-		    return_error(e_invalidfileaccess);
-	}
-	code = lib_file_open(pname.fname, pname.len, cname, MAX_CNAME,
+	code = lib_file_open(i_ctx_p->lib_path, i_ctx_p, pname.fname, pname.len, cname, MAX_CNAME,
 			     &clen, &fref, imemory);
 	if (code >= 0) {
 	    s = fptr(&fref);
@@ -585,7 +644,7 @@ zlibfile(i_ctx_t *i_ctx_p)
 	    }
 	}
 	if (code < 0) {
-	    if (code == e_VMerror)
+	    if (code == e_VMerror || code == e_invalidfileaccess)
 		return code;
 	    push(1);
 	    make_false(op);
@@ -598,6 +657,89 @@ zlibfile(i_ctx_t *i_ctx_p)
     return 0;
 }
 
+/* A "simple" prefix is defined as a (possibly empty) string of
+   alphanumeric, underscore, and hyphen characters. */
+private bool
+prefix_is_simple(const char *pstr)
+{
+    int i;
+    char c;
+
+    for (i = 0; (c = pstr[i]) != 0; i++) {
+	if (!(c == '-' || c == '_' || (c >= '0' && c <= '9') ||
+	      (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z')))
+	    return false;
+    }
+    return true;
+}
+
+/* <prefix|null> <access_string> .tempfile <name_string> <file> */
+private int
+ztempfile(i_ctx_t *i_ctx_p)
+{
+    os_ptr op = osp;
+    const char *pstr;
+    char fmode[4];
+    int code = parse_file_access_string(op, fmode);
+    char prefix[gp_file_name_sizeof];
+    char fname[gp_file_name_sizeof];
+    uint fnlen;
+    FILE *sfile;
+    stream *s;
+    byte *buf;
+
+    if (code < 0)
+	return code;
+    strcat(fmode, gp_fmode_binary_suffix);
+    if (r_has_type(op - 1, t_null))
+	pstr = gp_scratch_file_name_prefix;
+    else {
+	uint psize;
+
+	check_read_type(op[-1], t_string);
+	psize = r_size(op - 1);
+	if (psize >= gp_file_name_sizeof)
+	    return_error(e_rangecheck);
+	memcpy(prefix, op[-1].value.const_bytes, psize);
+	prefix[psize] = 0;
+	pstr = prefix;
+    }
+
+    if (gp_file_name_is_absolute(pstr, strlen(pstr))) {
+	if (check_file_permissions(i_ctx_p, pstr, strlen(pstr),
+				   "PermitFileWriting") < 0) {
+	    return_error(e_invalidfileaccess);
+	}
+    } else if (!prefix_is_simple(pstr)) {
+	return_error(e_invalidfileaccess);
+    }
+
+    s = file_alloc_stream(imemory, "ztempfile(stream)");
+    if (s == 0)
+	return_error(e_VMerror);
+    buf = gs_alloc_bytes(imemory, file_default_buffer_size,
+			 "ztempfile(buffer)");
+    if (buf == 0)
+	return_error(e_VMerror);
+    sfile = gp_open_scratch_file(pstr, fname, fmode);
+    if (sfile == 0) {
+	gs_free_object(imemory, buf, "ztempfile(buffer)");
+	return_error(e_invalidfileaccess);
+    }
+    fnlen = strlen(fname);
+    file_init_stream(s, sfile, fmode, buf, file_default_buffer_size);
+    code = ssetfilename(s, (const unsigned char*) fname, fnlen);
+    if (code < 0) {
+	sclose(s);
+	iodev_default->procs.delete_file(iodev_default, fname);
+	return_error(e_VMerror);
+    }
+    make_const_string(op - 1, a_readonly | icurrent_space, fnlen,
+		      s->file_name.data);
+    make_stream_file(op, s, fmode);
+    return code;
+}
+
 /* ------ Initialization procedure ------ */
 
 const op_def zfile_op_defs[] =
@@ -606,12 +748,12 @@ const op_def zfile_op_defs[] =
     {"1.execfile", zexecfile},
     {"2file", zfile},
     {"3filenameforall", zfilenameforall},
-    {"1.filenamedirseparator", zfilenamedirseparator},
     {"0.filenamelistseparator", zfilenamelistseparator},
     {"1.filenamesplit", zfilenamesplit},
     {"1.libfile", zlibfile},
     {"2renamefile", zrenamefile},
     {"1status", zstatus},
+    {"2.tempfile", ztempfile},
 		/* Internal operators */
     {"0%file_continue", file_continue},
     {"0%execfile_finish", execfile_finish},
@@ -623,11 +765,24 @@ const op_def zfile_op_defs[] =
 /* Parse a file name into device and individual name. */
 /* See gsfname.c for details. */
 private int
-parse_file_name(const ref * op, gs_parsed_file_name_t * pfn)
+parse_file_name(const ref * op, gs_parsed_file_name_t * pfn, bool safemode)
 {
+    int code;
+
     check_read_type(*op, t_string);
-    return gs_parse_file_name(pfn, (const char *)op->value.const_bytes,
+    code = gs_parse_file_name(pfn, (const char *)op->value.const_bytes,
 			      r_size(op));
+    if (code < 0)
+	return code;
+    /*
+     * Check here for the %pipe device which is illegal when
+     * LockFilePermissions is true. In the future we might want to allow
+     * the %pipe device to be included on the PermitFile... paths, but
+     * for now it is simply disallowed.
+     */
+    if (pfn->iodev && safemode && strcmp(pfn->iodev->dname, "%pipe%") == 0)
+	return e_invalidfileaccess;
+    return code;
 }
 
 /* Parse a real (non-device) file name and convert to a C string. */
@@ -641,6 +796,40 @@ parse_real_file_name(const ref *op, gs_parsed_file_name_t *pfn,
 				   r_size(op), mem, cname);
 }
 
+/* Parse the access string for opening a file. */
+/* [4] is for r/w, +, b, \0. */
+private int
+parse_file_access_string(const ref *op, char file_access[4])
+{
+    const byte *astr;
+
+    check_read_type(*op, t_string);
+    astr = op->value.const_bytes;
+    switch (r_size(op)) {
+	case 2:
+	    if (astr[1] != '+')
+		return_error(e_invalidfileaccess);
+	    file_access[1] = '+';
+	    file_access[2] = 0;
+	    break;
+	case 1:
+	    file_access[1] = 0;
+	    break;
+	default:
+	    return_error(e_invalidfileaccess);
+    }
+    switch (astr[0]) {
+	case 'r':
+	case 'w':
+	case 'a':
+	    break;
+	default:
+	    return_error(e_invalidfileaccess);
+    }
+    file_access[0] = astr[0];
+    return 0;
+}
+
 /* ------ Stream opening ------ */
 
 /*
@@ -682,7 +871,7 @@ iodev_os_open_file(gx_io_device * iodev, const char *fname, uint len,
 {
     return file_open_stream(fname, len, file_access,
 			    file_default_buffer_size, ps,
-			    iodev->procs.fopen, mem);
+			    iodev, iodev->procs.fopen, mem);
 }
 
 /* Make a t_file reference to a stream. */
@@ -702,96 +891,193 @@ make_stream_file(ref * pfile, stream * s, const char *access)
     }
 }
 
-/* Open an OS-level file (like fopen), using the search paths if necessary. */
-/* Note that it does not automatically look in the current */
-/* directory first (or at all): this is like Unix, and unlike MS-DOS. */
-private int
-lib_file_fopen(gx_io_device * iodev, const char *bname,
-	       const char *ignore_access, FILE ** pfile,
-	       char *rfname, uint rnamelen)
+private gp_file_name_combine_result 
+gp_file_name_combine_patch(const char *prefix, uint plen, const char *fname, uint flen, 
+			    bool no_sibling, char *buffer, uint *blen)
 {
-    char fmode[3];		/* r, [b], null */
-    int len = strlen(bname);
-    const gs_file_path *pfpath = &gs_lib_path;
-    uint pi;
+    return gp_file_name_combine(prefix, plen, fname, flen, no_sibling, buffer, blen);
+}
 
-    strcpy(fmode, "r");
+/* Prepare a stream with a file name. */
+/* Return 0 if successful, error code if not. */
+/* On a successful return, the C file name is in the stream buffer. */
+/* If fname==0, set up stream, and buffer. */
+private int
+file_prepare_stream(const char *fname, uint len, const char *file_access, 
+		 uint buffer_size, stream ** ps, char fmode[4], 
+		 gx_io_device *iodev, gs_memory_t *mem)
+{
+    byte *buffer;
+    register stream *s;
+
+    /* Open the file, always in binary mode. */
+    strcpy(fmode, file_access);
     strcat(fmode, gp_fmode_binary_suffix);
-    if (gp_file_name_is_absolute(bname, len))
-	return (*iodev->procs.fopen)(iodev, bname, fmode, pfile,
-				     rfname, rnamelen);
-    /* Go through the list of search paths */
-    for (pi = 0; pi < r_size(&pfpath->list); ++pi) {
-	const ref *prdir = pfpath->list.value.refs + pi;
-	const char *pstr = (const char *)prdir->value.const_bytes;
-	uint plen = r_size(prdir);
-	const char *cstr = gp_file_name_concat_string(pstr, plen);
-	int up, i;
-	int code;
+    if (buffer_size == 0)
+	buffer_size = file_default_buffer_size;
+    if (len >= buffer_size)    /* we copy the file name into the buffer */
+	return_error(e_limitcheck);
+    /* Allocate the stream first, since it persists */
+    /* even after the file has been closed. */
+    s = file_alloc_stream(mem, "file_prepare_stream");
+    if (s == 0)
+	return_error(e_VMerror);
+    /* Allocate the buffer. */
+    buffer = gs_alloc_bytes(mem, buffer_size, "file_prepare_stream(buffer)");
+    if (buffer == 0)
+	return_error(e_VMerror);
+    if (fname != 0) {
+	memcpy(buffer, fname, len);
+	buffer[len] = 0;	/* terminate string */
+    } else
+	buffer[0] = 0;	/* safety */
+    s->cbuf = buffer;
+    s->bsize = s->cbsize = buffer_size;
+    *ps = s;
+    return 0;
+}
 
-	/* Concatenate the prefix, combiner, and file name. */
-	/* Do this carefully in case rfname is the same */
-	/* as fname.  (We don't worry about the case */
-	/* where rfname only overlaps fname.) */
-	up = plen + strlen(cstr);
-	if (up + len + 1 > rnamelen)
-	    return_error(e_limitcheck);
-	for (i = len + 1; --i >= 0;)
-	    rfname[i + up] = bname[i];
-	memcpy(rfname, pstr, plen);
-	memcpy(rfname + plen, cstr, strlen(cstr));
-	code = (*iodev->procs.fopen)(iodev, rfname, fmode,
-				     pfile, rfname, rnamelen);
-	if (code >= 0)
-	    return code;
-	/* strcpy isn't guaranteed to work for overlapping */
-	/* source and destination, so: */
-	if (rfname == bname)
-	    for (i = 0; (rfname[i] = rfname[i + up]) != 0; i++);
+private int
+check_file_permissions_aux(i_ctx_t *i_ctx_p, char *fname, uint flen)
+{   /* i_ctx_p is NULL running init files. */
+    /* fname must be reduced. */
+    if (i_ctx_p == NULL)
+	return 0;
+    if (check_file_permissions_reduced(i_ctx_p, fname, flen, "PermitFileReading") < 0)
+	return_error(e_invalidfileaccess);
+    return 0;
+}
+
+
+/* The startup code calls this to open @-files. */
+private int
+lib_fopen_with_libpath(gs_file_path_ptr  lib_path,
+		       const gs_memory_t *mem,
+		       i_ctx_t *i_ctx_p,      
+		       gx_io_device *iodev, 
+		       const char *fname, uint flen, char fmode[4], char *buffer, int blen,
+		       FILE **file)
+{   /* i_ctx_p is NULL running init files. 
+     * lib_path and mem are never NULL 
+     */
+    bool starting_arg_file = false;
+    bool search_with_no_combine = false;
+    bool search_with_combine = false;
+
+    if (i_ctx_p != NULL) {
+	starting_arg_file = i_ctx_p->starting_arg_file;
+	i_ctx_p->starting_arg_file = false;
+    } else
+	starting_arg_file = true;
+    if (gp_file_name_is_absolute(fname, flen)) {
+       search_with_no_combine = true;
+       search_with_combine = false;
+    } else {
+       search_with_no_combine = starting_arg_file;
+       search_with_combine = true;
+    }
+    if (search_with_no_combine) {
+	uint blen1 = blen;
+
+	if (gp_file_name_reduce(fname, flen, buffer, &blen1) != gp_combine_success)
+	    goto skip;
+	if (iodev->procs.fopen(iodev, buffer, fmode, file,
+				 buffer, blen) == 0) {
+	    if (starting_arg_file ||
+		check_file_permissions_aux(i_ctx_p, buffer, blen1) >= 0)
+		    return 0;
+	    iodev->procs.fclose(iodev, *file);
+	    *file = NULL;
+	    return_error(e_invalidfileaccess);
+	} else
+	    *file = NULL;
+	skip:;
+    } 
+    if (search_with_combine) {
+	const gs_file_path *pfpath = lib_path;
+	uint pi;
+
+	for (pi = 0; pi < r_size(&pfpath->list); ++pi) {
+	    const ref *prdir = pfpath->list.value.refs + pi;
+	    const char *pstr = (const char *)prdir->value.const_bytes;
+	    uint plen = r_size(prdir), blen1 = blen;
+
+	    gp_file_name_combine_result r = gp_file_name_combine_patch(pstr, plen, 
+		    fname, flen, false, buffer, &blen1);
+	    if (r != gp_combine_success)
+		continue;
+	    if (iodev->procs.fopen(iodev, buffer, fmode, file,
+					 buffer, blen) == 0) {
+		if (starting_arg_file ||
+		    check_file_permissions_aux(i_ctx_p, buffer, blen1) >= 0)
+		    return 0;
+		iodev->procs.fclose(iodev, *file);
+		*file = NULL;
+		return_error(e_invalidfileaccess);
+	    }
+	    *file = NULL; /* safety */
+	}
     }
     return_error(e_undefinedfilename);
 }
+
 /* The startup code calls this to open @-files. */
 FILE *
-lib_fopen(const char *bname)
+lib_fopen(const gs_file_path_ptr pfpath, const gs_memory_t *mem, const char *fname)
 {
-    FILE *file = NULL;
     /* We need a buffer to hold the expanded file name. */
     char buffer[gp_file_name_sizeof];
     /* We can't count on the IODevice table to have been initialized yet. */
     /* Allocate a copy of the default IODevice. */
-    gx_io_device iodev_default_copy;
-    int code;
+    gx_io_device iodev_default_copy = *gx_io_device_table[0];
+    char fmode[3] = "r";
+    FILE *file = NULL;
 
-    iodev_default_copy = *gx_io_device_table[0];
-    code = lib_file_fopen(&iodev_default_copy, bname, "r", &file,
-			  buffer, gp_file_name_sizeof);
-    return (code < 0 ? NULL : file);
+    strcat(fmode, gp_fmode_binary_suffix);
+    lib_fopen_with_libpath(pfpath, mem, NULL, &iodev_default_copy, fname, strlen(fname), 
+			    fmode, buffer, sizeof(buffer), &file);
+    return file;
 }
 
 /* Open a file stream on an OS file and create a file object, */
 /* using the search paths. */
 /* The startup code calls this to open the initialization file gs_init.ps. */
 int
-lib_file_open(const char *fname, uint len, byte * cname, uint max_clen,
+lib_file_open(const gs_file_path_ptr pfpath, 
+	      i_ctx_t *i_ctx_p, const char *fname, uint len, byte * cname, uint max_clen,
 	      uint * pclen, ref * pfile, gs_memory_t *mem)
-{
+{   /* i_ctx_p is NULL running init files. */
     stream *s;
-    int code = file_open_stream(fname, len, "r", file_default_buffer_size,
-				&s, lib_file_fopen, mem);
-    char *bname;
+    int code;
+    char fmode[4];  /* r/w/a, [+], [b], null */
+    char *buffer;
     uint blen;
+    gx_io_device *iodev = iodev_default;
+    FILE *file;
 
+    code = file_prepare_stream(fname, len, "r", file_default_buffer_size, 
+			    &s, fmode, iodev, mem);
     if (code < 0)
 	return code;
+    if (fname == 0)
+	return 0;
+    buffer = (char *)s->cbuf;
+    code = lib_fopen_with_libpath(pfpath, mem, i_ctx_p, 
+				  iodev, fname, len, fmode, buffer, s->bsize, &file);
+    if (code < 0) {
+	s->cbuf = NULL;
+	s->bsize = s->cbsize = 0;
+	gs_free_object(mem, buffer, "lib_file_open");
+        return code;
+    }
+    file_init_stream(s, file, fmode, (byte *)buffer, s->bsize);
     /* Get the name from the stream buffer. */
-    bname = (char *)s->cbuf;
-    blen = strlen(bname);
+    blen = strlen(buffer);
     if (blen > max_clen) {
 	sclose(s);
-	return_error(e_limitcheck);
+        return_error(e_limitcheck);
     }
-    memcpy(cname, bname, blen);
+    memcpy(cname, buffer, blen);
     *pclen = blen;
     make_stream_file(pfile, s, "r");
     return 0;
@@ -816,6 +1102,37 @@ file_read_string(const byte *str, uint len, ref *pfile, gs_ref_memory_t *imem)
     return 0;
 }
 
+/*
+ * Set up a file stream on an OS file.  The caller has allocated the
+ * stream and buffer.
+ */
+private void
+file_init_stream(stream *s, FILE *file, const char *fmode, byte *buffer,
+		 uint buffer_size)
+{
+    switch (fmode[0]) {
+    case 'a':
+	sappend_file(s, file, buffer, buffer_size);
+	break;
+    case 'r':
+	/* Defeat buffering for terminals. */
+	{
+	    struct stat rstat;
+
+	    fstat(fileno(file), &rstat);
+	    sread_file(s, file, buffer,
+		       (S_ISCHR(rstat.st_mode) ? 1 : buffer_size));
+	}
+	break;
+    case 'w':
+	swrite_file(s, file, buffer, buffer_size);
+    }
+    if (fmode[1] == '+')
+	s->file_modes |= s_mode_read | s_mode_write;
+    s->save_close = s->procs.close;
+    s->procs.close = file_close_file;
+}
+
 /* Open a file stream, optionally on an OS file. */
 /* Return 0 if successful, error code if not. */
 /* On a successful return, the C file name is in the stream buffer. */
@@ -823,71 +1140,26 @@ file_read_string(const byte *str, uint len, ref *pfile, gs_ref_memory_t *imem)
 /* but don't open an OS file or initialize the stream. */
 int
 file_open_stream(const char *fname, uint len, const char *file_access,
-		 uint buffer_size, stream ** ps, iodev_proc_fopen_t fopen_proc,
-		 gs_memory_t *mem)
+		 uint buffer_size, stream ** ps, gx_io_device *iodev,
+		 iodev_proc_fopen_t fopen_proc, gs_memory_t *mem)
 {
-    byte *buffer;
-    register stream *s;
+    int code;
+    FILE *file;
+    char fmode[4];  /* r/w/a, [+], [b], null */
 
-    if (buffer_size == 0)
-	buffer_size = file_default_buffer_size;
-    if (len >= buffer_size)    /* we copy the file name into the buffer */
-	return_error(e_limitcheck);
-    /* Allocate the stream first, since it persists */
-    /* even after the file has been closed. */
-    s = file_alloc_stream(mem, "file_open_stream");
-    if (s == 0)
-	return_error(e_VMerror);
-    /* Allocate the buffer. */
-    buffer = gs_alloc_bytes(mem, buffer_size, "file_open_stream(buffer)");
-    if (buffer == 0)
-	return_error(e_VMerror);
-    if (fname != 0) {
-	/* Copy the name (so we can terminate it with a zero byte.) */
-	char *file_name = (char *)buffer;
-	char fmode[4];		/* r/w/a, [+], [b], null */
-	FILE *file;
-	int code;
-
-	memcpy(file_name, fname, len);
-	file_name[len] = 0;	/* terminate string */
-	/* Open the file, always in binary mode. */
-	strcpy(fmode, file_access);
-	strcat(fmode, gp_fmode_binary_suffix);
-	/****** iodev_default IS QUESTIONABLE ******/
-	code = (*fopen_proc)(iodev_default, file_name, fmode, &file,
-			     (char *)buffer, buffer_size);
-	if (code < 0) {
-	    gs_free_object(mem, buffer, "file_open_stream(buffer)");
-	    return code;
-	}
-	/* Set up the stream. */
-	switch (fmode[0]) {
-	    case 'a':
-		sappend_file(s, file, buffer, buffer_size);
-		break;
-	    case 'r':
-		/* Defeat buffering for terminals. */
-		{
-		    struct stat rstat;
-
-		    fstat(fileno(file), &rstat);
-		    sread_file(s, file, buffer,
-			       (S_ISCHR(rstat.st_mode) ? 1 : buffer_size));
-		}
-		break;
-	    case 'w':
-		swrite_file(s, file, buffer, buffer_size);
-	}
-	if (fmode[1] == '+')
-	    s->file_modes |= s_mode_read | s_mode_write;
-	s->save_close = s->procs.close;
-	s->procs.close = file_close_file;
-    } else {			/* save the buffer and size */
-	s->cbuf = buffer;
-	s->bsize = s->cbsize = buffer_size;
-    }
-    *ps = s;
+    if (!iodev)
+	iodev = iodev_default;
+    code = file_prepare_stream(fname, len, file_access, buffer_size, ps, fmode, 
+			    (!iodev ? iodev_default : iodev), mem);
+    if (code < 0)
+	return code;
+    if (fname == 0)
+	return 0;
+    code = (*fopen_proc)(iodev, (char *)(*ps)->cbuf, fmode, &file,
+			 (char *)(*ps)->cbuf, (*ps)->bsize);
+    if (code < 0)
+	return code;
+    file_init_stream(*ps, file, fmode, (*ps)->cbuf, (*ps)->bsize);
     return 0;
 }
 
@@ -918,8 +1190,8 @@ filter_open(const char *file_access, uint buffer_size, ref * pfile,
 	if (sst == 0)
 	    return_error(e_VMerror);
     }
-    code = file_open_stream((char *)0, 0, file_access,
-			    buffer_size, &s, (iodev_proc_fopen_t)0, mem);
+    code = file_open_stream((char *)0, 0, file_access, buffer_size, &s,
+    				(gx_io_device *)0, (iodev_proc_fopen_t)0, mem);
     if (code < 0) {
 	gs_free_object(mem, sst, "filter_open(stream_state)");
 	return code;
