@@ -1,23 +1,20 @@
 /* Copyright (C) 1989, 2000, 2001 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This software is provided AS-IS with no warranty, either express or
+  implied.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
-  to anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer
-  to the GNU General Public License for full details.
+  This software is distributed under license and may not be copied,
+  modified or distributed except as expressly authorized under the terms
+  of the license contained in the file LICENSE in this distribution.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given
-  to you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
+  For more information about licensing, please refer to
+  http://www.ghostscript.com/licensing/. For information on
+  commercial licensing, go to http://www.artifex.com/licensing/ or
+  contact Artifex Software, Inc., 101 Lucas Valley Road #110,
+  San Rafael, CA  94903, U.S.A., +1(415)492-9861.
 */
 
-/*$RCSfile: zfileio.c,v $ $Revision: 1.9.2.4 $ */
+/* $Id: zfileio.c,v 1.17 2004/08/04 19:36:13 stefan Exp $ */
 /* File I/O operators */
 #include "memory_.h"
 #include "ghost.h"
@@ -35,11 +32,11 @@
 #include "estack.h"
 
 /* Forward references */
-private int write_string(P2(ref *, stream *));
-private int handle_read_status(P5(i_ctx_t *, int, const ref *, const uint *,
-				  op_proc_t));
-private int handle_write_status(P5(i_ctx_t *, int, const ref *, const uint *,
-				   op_proc_t));
+private int write_string(ref *, stream *);
+private int handle_read_status(i_ctx_t *, int, const ref *, const uint *,
+			       op_proc_t);
+private int handle_write_status(i_ctx_t *, int, const ref *, const uint *,
+				op_proc_t);
 
 /* ------ Operators ------ */
 
@@ -77,15 +74,27 @@ zread(i_ctx_t *i_ctx_p)
     int ch;
 
     check_read_file(s, op);
+    /* We 'push' first in case of ostack block overflow and the */
+    /* usual case is we will need to push anyway. If we get EOF */
+    /* we will need to 'pop' and decrement the 'op' pointer.    */
+    /* This is required since the 'push' macro might return with*/
+    /* stackoverflow which will result in another stack block   */
+    /* added on, then the operator being retried. We can't read */
+    /* (sgetc) prior to having a place on the ostack to return  */
+    /* the character.						*/
+    push(1);
     ch = sgetc(s);
     if (ch >= 0) {
-	push(1);
 	make_int(op - 1, ch);
 	make_bool(op, 1);
-    } else if (ch == EOFC)
+    } else {
+	pop(1);		/* Adjust ostack back from preparatory 'pop' */
+	op--;
+	if (ch == EOFC) 
 	make_bool(op, 0);
     else
 	return handle_read_status(i_ctx_p, ch, op, NULL, zread);
+    }
     return 0;
 }
 
@@ -110,7 +119,7 @@ zwrite(i_ctx_t *i_ctx_p)
 }
 
 /* <file> <string> readhexstring <substring> <filled_bool> */
-private int zreadhexstring_continue(P1(i_ctx_t *));
+private int zreadhexstring_continue(i_ctx_t *);
 
 /* We keep track of the odd digit in the next byte of the string */
 /* beyond the bytes already used.  (This is just for convenience; */
@@ -196,7 +205,7 @@ zreadhexstring_continue(i_ctx_t *i_ctx_p)
 }
 
 /* <file> <string> writehexstring - */
-private int zwritehexstring_continue(P1(i_ctx_t *));
+private int zwritehexstring_continue(i_ctx_t *);
 private int
 zwritehexstring_at(i_ctx_t *i_ctx_p, os_ptr op, uint odd)
 {
@@ -275,7 +284,7 @@ zwritehexstring_continue(i_ctx_t *i_ctx_p)
 }
 
 /* <file> <string> readstring <substring> <filled_bool> */
-private int zreadstring_continue(P1(i_ctx_t *));
+private int zreadstring_continue(i_ctx_t *);
 private int
 zreadstring_at(i_ctx_t *i_ctx_p, os_ptr op, uint start)
 {
@@ -353,8 +362,8 @@ zwritestring(i_ctx_t *i_ctx_p)
 }
 
 /* <file> <string> readline <substring> <bool> */
-private int zreadline(P1(i_ctx_t *));
-private int zreadline_continue(P1(i_ctx_t *));
+private int zreadline(i_ctx_t *);
+private int zreadline_continue(i_ctx_t *);
 
 /*
  * We could handle readline the same way as readstring,
@@ -652,8 +661,8 @@ zfilename(i_ctx_t *i_ctx_p)
 	return_error(e_VMerror);
     memcpy(str, fname.data, fname.size);
     push(1);			/* can't fail */
-    make_const_string(op - 1, 
-		      a_all | imemory_space(imemory), 
+    make_const_string( op - 1 , 
+		      a_all | imemory_space((const struct gs_ref_memory_s*) imemory), 
 		      fname.size, 
 		      str);
     make_true(op);
@@ -737,7 +746,7 @@ zunread(i_ctx_t *i_ctx_p)
 }
 
 /* <file> <obj> <==flag> .writecvp - */
-private int zwritecvp_continue(P1(i_ctx_t *));
+private int zwritecvp_continue(i_ctx_t *);
 private int
 zwritecvp_at(i_ctx_t *i_ctx_p, os_ptr op, uint start, bool first)
 {
@@ -753,7 +762,7 @@ zwritecvp_at(i_ctx_t *i_ctx_p, os_ptr op, uint start, bool first)
     code = obj_cvp(op - 1, str, sizeof(str), &len, (int)op->value.intval,
 		   start, imemory);
     if (code == e_rangecheck) {
-	code = obj_string_data(op - 1, &data, &len);
+        code = obj_string_data(imemory, op - 1, &data, &len);
 	if (len < start)
 	    return_error(e_rangecheck);
 	data += start;
@@ -810,6 +819,32 @@ zwritecvp_continue(i_ctx_t *i_ctx_p)
     return zwritecvp_at(i_ctx_p, op - 1, (uint) op->value.intval, false);
 }
 
+/* Callout for stdin */
+/* - .needstdin - */
+int
+zneedstdin(i_ctx_t *i_ctx_p)
+{
+    return e_NeedStdin;		/* Interpreter will exit to caller. */
+}
+
+/* Callout for stdout */
+/* - .needstdout - */
+int
+zneedstdout(i_ctx_t *i_ctx_p)
+{
+    return e_NeedStdout;	/* Interpreter will exit to caller. */
+}
+
+/* Callout for stderr */
+/* - .needstderr - */
+int
+zneedstderr(i_ctx_t *i_ctx_p)
+{
+    return e_NeedStderr;	/* Interpreter will exit to caller. */
+}
+
+
+
 /* ------ Initialization procedure ------ */
 
 /* We need to split the table because of the 16-element limit. */
@@ -846,6 +881,9 @@ const op_def zfileio2_op_defs[] = {
     {"3%zreadstring_continue", zreadstring_continue},
     {"4%zwritecvp_continue", zwritecvp_continue},
     {"3%zwritehexstring_continue", zwritehexstring_continue},
+    {"0.needstdin", zneedstdin},
+    {"0.needstdout", zneedstdout},
+    {"0.needstderr", zneedstderr},
     op_def_end(0)
 };
 
@@ -927,21 +965,10 @@ copy_error_string(i_ctx_t *i_ctx_p, const ref *fop)
     return_error(e_ioerror);
 }
 
-/**
- * handle_read_status: Handle an exceptional status return from a read stream.
- * @i_ctx_p: The interpreter context.
- * @ch: The status return code.
- * @fop: A pointer to the ref for the stream.
- * @pindex: Pointer to additional integer to push on estack.
- * @cont: Continuation procedure.
- *
- * This routine handles exceptional status returns from read streams. If
- * @pindex is non-NULL, the integer it points to is pushed on the estack
- * for the continuation to consume.
- *
- * Return value: 1 on EOF, 0 on other success, o_push_estack if estack
- * needs enlargement, or error.
- **/
+/* Handle an exceptional status return from a read stream. */
+/* fop points to the ref for the stream. */
+/* ch may be any stream exceptional value. */
+/* Return 0, 1 (EOF), o_push_estack, or an error. */
 private int
 handle_read_status(i_ctx_t *i_ctx_p, int ch, const ref * fop,
 		   const uint * pindex, op_proc_t cont)
@@ -965,21 +992,10 @@ handle_read_status(i_ctx_t *i_ctx_p, int ch, const ref * fop,
     }
 }
 
-/**
- * handle_write_status: Handle exceptional status return from a write stream.
- * @i_ctx_p: The interpreter context.
- * @ch: The status return code.
- * @fop: A pointer to the ref for the stream.
- * @pindex: Pointer to additional integer to push on estack.
- * @cont: Continuation procedure.
- *
- * This routine handles exceptional status returns from write streams. If
- * @pindex is non-NULL, the integer it points to is pushed on the estack
- * for the continuation to consume.
- *
- * Return value: 1 on EOF, 0 on other success, o_push_estack if estack
- * needs enlargement, or error.
- **/
+/* Handle an exceptional status return from a write stream. */
+/* fop points to the ref for the stream. */
+/* ch may be any stream exceptional value. */
+/* Return 0, 1 (EOF), o_push_estack, or an error. */
 private int
 handle_write_status(i_ctx_t *i_ctx_p, int ch, const ref * fop,
 		    const uint * pindex, op_proc_t cont)
